(* cloud_relations.v

   Define logical relation on cloud using Setoids
   This is similar to state_relations butusing 
   a classical conjunction.
   This duplication of code is due to the necessity
   to have two different carrers if one want to use 
   two ring structures on the same objetc.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import state_relations.

Implicit Arguments nullPfunction [].
Implicit Arguments join_disj_comm [A B].

Definition Acloud := nat.
Definition Bcloud := nat.

Record cloudRel : Type := mkCloudRel {
  cloudrel :> (Pfunction Acloud Bcloud) -> (Pfunction Acloud Bcloud)-> 
              program -> program -> nat -> Prop;
  cloudRel_cond : forall s1 s2 s1' s2' p1 p2 k1 k2, 
    cloudrel s1 s2 p1 p2 k1 -> 
    subPfunction s1 s1' -> 
    subPfunction s2 s2' -> 
    (k2 <= k1) -> 
    cloudrel s1' s2' p1 p2 k2}.

Implicit Arguments cloudRel_cond [].

Definition cloudRelEq (P Q: cloudRel) := forall s s' p p' k , P s s' p p' k <-> Q s s' p p' k.

Lemma cloudRelEq_refl : forall (P : cloudRel), cloudRelEq P P.
Proof.
  firstorder.
Qed.

Debug Off.

Lemma cloudRelEq_comm : forall (P Q : cloudRel), cloudRelEq P Q -> cloudRelEq Q P.
Proof.
  unfold cloudRelEq; intros.
  hreplace H with (H s s' p p' k).
  intuition.
Qed.

Lemma cloudRelEq_trans : forall (P Q R : cloudRel), cloudRelEq P Q -> cloudRelEq Q R -> cloudRelEq P R.
Proof.
  unfold cloudRelEq; intros; simpl.
  hreplace H0 with (H0 s s' p p' k); hreplace H with (H s s' p p' k).
  firstorder.
Qed.

Hint Resolve cloudRelEq_refl cloudRelEq_trans.

Add Relation cloudRel cloudRelEq
   reflexivity proved by cloudRelEq_refl
   symmetry proved by cloudRelEq_comm
   transitivity proved by cloudRelEq_trans
 as cloudRelEqRel.

Add Morphism cloudrel with signature cloudRelEq ==> (@eq (Pfunction Acloud Bcloud)) ==>
  (@eq (Pfunction Acloud Bcloud)) ==> (@eq program) ==> (@eq program) ==> (@eq nat) ==> iff 
as cloudrelcompat.
   intros ; subst; apply H.
Qed.

Definition stateToCloud (P:stateRel) : cloudRel.
Proof.
  intros P.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun k => P s s' p p' k) _).
  intuition.
  eapply (stateRel_cond P);try eassumption.
Defined.

Definition cloudToState (P:cloudRel) : stateRel.
Proof.
  intros P.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => P s s' p p' k) _).
  intuition.
  eapply (cloudRel_cond P);try eassumption.
Defined.

Lemma bijection_state_pred1 : forall P, stateRelEq P (cloudToState(stateToCloud P)).
Proof.
  unfold stateRelEq; simpl; intuition.
Qed.

Lemma bijection_state_pred2 : forall P, cloudRelEq P (stateToCloud(cloudToState P)).
Proof.
  unfold cloudRelEq; simpl; intuition.
Qed.

Definition cloudRelLeq (P Q : cloudRel) := forall s s' p p' k , P s s' p p' k -> Q s s' p p' k.

Lemma cloudRelLeq_refl : forall P, cloudRelLeq P P.
Proof.
  firstorder.
Qed.

Lemma cloudRelLeq_trans : forall P Q R,  cloudRelLeq P Q -> cloudRelLeq Q R -> cloudRelLeq P R.
Proof.
  firstorder.
Qed.

Hint Resolve cloudRelLeq_refl cloudRelLeq_trans.

Add Relation cloudRel cloudRelLeq
   reflexivity proved by cloudRelLeq_refl
   transitivity proved by cloudRelLeq_trans
 as cloudRelLeqRel.

Add Morphism cloudRelLeq with signature cloudRelEq ==> cloudRelEq ==> iff
as cloudpredleqcompat.
Proof.
  unfold cloudRelEq, cloudRelLeq; intuition.
  assert (foo := H s s' p p' k ); assert (foo2 := H0 s s' p p' k).
  firstorder.
  assert (foo := H s s' p p' k ); assert (foo2 := H1 s s' p p' k); assert (foo3 := H0 s s' p p' k).
  firstorder.
Qed.

Lemma cloudRelEq_Leq: forall P Q, cloudRelEq P Q -> cloudRelLeq P Q.
Proof.
  unfold cloudRelLeq,cloudRelEq; intros.
  rewrite <- (H s s' p p' k); assumption.
Qed.

Add Morphism cloudToState with signature 
  cloudRelEq ==> stateRelEq
as cloudtostatecompat.
Proof.
  intuition.
Qed.

Add Morphism stateToCloud with signature 
  stateRelEq ==> cloudRelEq
as statetocloudcompat.
Proof.
  intuition.
Qed.

Add Morphism cloudToState with signature 
  cloudRelLeq ++> stateRelLeq
as cloudtostatecompat2.
Proof.
  intuition.
Qed.

Add Morphism stateToCloud with signature 
  stateRelLeq ++> cloudRelLeq
as statetocloudcompat2.
Proof.
  intuition.
Qed.


(* Lifting pure predicates, true and false *)

Definition cloudRel_lift (P: Prop) : cloudRel.
Proof.
  intro P.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun  k => P) _);
  (intros; assumption).
Defined.

Add Morphism cloudRel_lift with signature iff ==> cloudRelEq
as cloudpredliftcompat.
Proof.
  unfold cloudRelEq, cloudRel_lift; simpl; firstorder. 
Qed.

Definition cloudPtrue := cloudRel_lift (True).

Definition cloudPfalse := cloudRel_lift (False).

Lemma cloudRellift_Leq : forall (P Q : Prop), (P -> Q) -> cloudRelLeq (cloudRel_lift P) (cloudRel_lift Q).
Proof.
  unfold cloudRel_lift, cloudRelLeq; intuition.
Qed.

Lemma cloudRelLeq_false_min : forall P , cloudRelLeq cloudPfalse P.
Proof.
  unfold cloudRelLeq, cloudPfalse, cloudRel_lift; simpl; intros.
  inversion H.
Qed.

Lemma cloudToStatePtrue : stateRelEq (cloudToState cloudPtrue) statePtrue.
Proof.
  unfold stateRelEq, cloudToState; simpl; intuition.
Qed.

Lemma cloudToStatePfalse : stateRelEq (cloudToState cloudPfalse) statePfalse.
Proof.
  unfold stateRelEq, cloudToState; simpl; intuition.
Qed.

Lemma stateToCloudPtrue : cloudRelEq (stateToCloud statePtrue) cloudPtrue.
Proof.
  unfold cloudRelEq, stateToCloud; simpl; intuition.
Qed.

Lemma stateToCloudPfalse : cloudRelEq (stateToCloud statePfalse) cloudPfalse.
Proof.
  unfold cloudRelEq, stateToCloud; simpl; intuition.
Qed.
  
(*-------------------------*)
(* Disj *)
(*-------------------------*)

Definition cloudRelDisj (P Q: cloudRel) : cloudRel.
Proof. 
  intros.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun k =>
    P s s' p p' k \/ Q s s' p p' k)
  _).
  intros.
  destruct H.
  left; apply (cloudRel_cond P s1 s2 s1' s2' p1 p2 k1 k2);assumption.
  right; apply (cloudRel_cond Q s1 s2 s1' s2' p1 p2 k1 k2); assumption.
Defined.


Add Morphism cloudRelDisj with signature cloudRelEq ==> cloudRelEq ==> cloudRelEq
as cloudpreddisjcompat.
Proof.
  unfold cloudRelDisj in *; unfold cloudRelEq in *; simpl; intros.
  hreplace H with (H s s' p p' k); hreplace H0 with (H0 s s' p p' k).
  firstorder.
Qed.

Add Morphism cloudRelDisj with signature cloudRelLeq ++> cloudRelLeq ++> cloudRelLeq
as cloudpreddisjcompat2.
Proof.
  unfold cloudRelDisj; unfold cloudRelLeq; simpl; intros.
  firstorder.
Qed.


(******************)
(*     Notation    *)
(*******************)

Notation "P1 @@ P2" := (cloudRelDisj P1 P2) (at level 75, right associativity).


(*----------------------------*)
(* Disj properties *)
(*----------------------------*)

Lemma cloudRelDisj_unit : forall P: cloudRel, cloudRelEq (cloudPfalse @@ P) P.
Proof.
  unfold cloudRelEq, cloudRelDisj; simpl;intros.
  intuition.
Qed.

Lemma cloudRelDisj_comm : forall P Q, cloudRelEq (P @@ Q) (Q @@ P).
Proof.
  unfold cloudRelEq, cloudRelDisj; simpl; intros.
  tauto.
Qed.

Lemma cloudRelDisj_assoc : forall P Q R, cloudRelEq (P @@ (Q @@ R)) ((P @@ Q) @@ R).
Proof.
  unfold cloudRelEq, cloudRelDisj; simpl; intros.
  tauto.
Qed.

(*-------------------------*)
(* Conj*)
(*-------------------------*)

Definition cloudRelConj (P Q: cloudRel) : cloudRel. 
  intros.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun k => P s s' p p' k /\ Q s s' p p' k)
                _).
  intros.
  deconstruct H as (Hp, Hq).
  split.
  apply (cloudRel_cond P s1 s2 s1' s2' p1 p2 k1 k2); assumption.
  apply (cloudRel_cond Q s1 s2 s1' s2' p1 p2 k1 k2); assumption.
Defined.

Add Morphism cloudRelConj with signature cloudRelEq ==> cloudRelEq ==> cloudRelEq
as ppredconjcompat.
Proof.
  unfold cloudRelEq; simpl; intros.
  hreplace H0 with (H0 s s' p p' k); hreplace H with (H s s' p p' k);
  firstorder.
Qed.

Add Morphism cloudRelConj with signature cloudRelLeq ++> cloudRelLeq ++> cloudRelLeq
as ppredconjcompat2.
Proof.
  unfold cloudRelLeq; simpl; firstorder.
Qed.

Lemma stateCloudDisj : forall P Q, cloudRelEq (stateToCloud P @@ stateToCloud Q) (stateToCloud (stateRelDisj P Q)).
Proof.
  unfold cloudRelEq, stateRelDisj, cloudRelDisj; simpl;intuition.
Qed.
  
(******************)
(*    Notation     *)
(*******************)

Notation "P1 //\\ P2" := (cloudRelConj P1 P2) (at level 75, right associativity).

(*----------------------------*)
(* Conj properties *)
(*----------------------------*)

Lemma cloudConjContraction : forall P Q, cloudRelLeq P Q -> cloudRelEq P (P//\\Q).
  unfold cloudRelEq, cloudRelLeq; simpl; intros.
  intuition.
Qed.

Lemma cloudRelConj_proj : forall P1 P2 Q, cloudRelEq P1 P2 -> 
	           cloudRelEq (P1 //\\ Q) (P2//\\Q).
Proof.
  unfold cloudRelEq; simpl;  intros.
  hreplace H with (H s s' p p' k).
  intuition.
Qed.

Lemma cloudRelConj_unit : forall P, cloudRelEq (cloudPtrue //\\ P) P.
Proof.
  unfold cloudRelEq; simpl; intros.
  intuition.
Qed.

Lemma cloudRelConj_zero : forall P, cloudRelEq (cloudPfalse //\\ P) cloudPfalse.
Proof.
  unfold cloudRelEq; simpl; intros.
  intuition.
Qed.

Lemma cloudRelConj_comm : forall P Q, cloudRelEq (P //\\ Q) (Q //\\ P).
Proof.
  unfold cloudRelEq; simpl; intros.
  intuition.
Qed.

Lemma cloudRelConj_assoc : forall P Q R, cloudRelEq (P //\\ (Q //\\ R)) ((P //\\ Q) //\\ R).
Proof.
  unfold cloudRelEq ; simpl; intros.
  intuition.
Qed.

Lemma cloudRelConj_weakening : forall P Q, cloudRelLeq (P //\\ Q) P.
Proof.
  unfold cloudRelLeq; simpl;  intros.
  intuition. 
Qed.

Lemma Distributivity : forall A B C, 
  stateRelLeq (cloudToState (stateToCloud A //\\ stateToCloud B) ## C)   (cloudToState ( stateToCloud (A ## C) //\\ stateToCloud (B ## C))).
Proof.
  unfold stateRelLeq; simpl; intuition;
  destruct H as [x [y [z [t H]]]];
  exists x; exists y; exists z; exists t;
  intuition.
Qed.

Lemma linearDistributivity : forall A B C, 
  stateRelLeq (cloudToState ( A //\\ stateToCloud B) ## C)   (cloudToState ( A //\\ stateToCloud (B ## C))).
Proof.
  unfold stateRelLeq; simpl; intuition;
  destruct H as [x [y [z [t H]]]].
  intuition.
  eapply cloudRel_cond; try eassumption; try trivial;
  eapply join_extends; eassumption.
  exists x; exists y; exists z; exists t;
  intuition.
Qed.

(*----------------------------*)
(* Conj vs Disj properties *)
(*----------------------------*)

Lemma cloudRelDistributivity : forall P Q R, cloudRelEq ( ( P @@ Q) //\\ R) ((P //\\ R) @@ (Q //\\ R)).
Proof.
  unfold cloudRelEq, cloudRelDisj, cloudRelConj; simpl; intuition.
Qed.

(*----------------------------*)
(* Definition of a ring on cloudRel *)
(*----------------------------*)


Definition ppred_ring2 : semi_ring_theory cloudPfalse cloudPtrue cloudRelDisj cloudRelConj cloudRelEq.
Proof.
apply mk_srt;
[ apply cloudRelDisj_unit | apply cloudRelDisj_comm | apply cloudRelDisj_assoc |
apply cloudRelConj_unit | apply cloudRelConj_zero | apply cloudRelConj_comm | 
apply cloudRelConj_assoc | apply cloudRelDistributivity].
Qed.

Add Ring cloudRel_ring : ppred_ring2.


(************)
(* Extensionality to rewrite under Exists and Forall  *)
(************)


Definition cloudRelExtensionalityLeq (X:Type) (P Q : X -> cloudRel) := forall n, cloudRelLeq (P n) (Q n).

Add Relation (nat -> cloudRel) (@cloudRelExtensionalityLeq nat) as cloudRelExt2.

Add Relation ((list nat) -> cloudRel) (@cloudRelExtensionalityLeq (list nat)) as cloudRelExtList2.


Definition cloudRelExtensionality (X:Type) (P Q : X -> cloudRel) := forall n, cloudRelEq (P n) (Q n).

Add Relation (nat -> cloudRel) (@cloudRelExtensionality nat) as cloudRelExt.

Add Relation ((list nat) -> cloudRel) (@cloudRelExtensionality (list nat)) as cloudRelExtList.

(************)
(* ExistP  *)
(************)

Definition cloudExists (X : Type) (h : X -> cloudRel ) : cloudRel.
  intros.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun k => exists a: X, h a s s' p p' k) _).
  intros.
  witness a in H.
  exists a.
  apply (cloudRel_cond (h a) s1 s2 s1' s2' p1 p2 k1 k2); assumption.
Defined.

Implicit Arguments cloudExists [X].

Notation "'Exists_c' h" := (cloudExists  h ) (at level 70).

Notation "'Ex_c' x , P" := (Exists_c (fun x=> P) ) (at level 70, x at next level).


Add Morphism (@stateExists nat) with signature (@stateRelExtensionality nat) ==> stateRelEq
as existsstatenat.
Proof.
    unfold stateRelEq, stateRelExtensionality,stateExists; simpl; firstorder.
    exists x0; rewrite <- H; intuition.
    exists x0; rewrite H; intuition.
Qed.

Add Morphism (@cloudExists (list nat)) with signature (@cloudRelExtensionality (list nat)) ==> cloudRelEq
as existscloudllistnat.
Proof.
  unfold cloudRelEq, cloudRelExtensionality,cloudExists; simpl; firstorder. 
   exists x0; rewrite <- H; intuition.
   exists x0; rewrite H; intuition.
Qed.


Add Morphism (@cloudExists nat) with signature (@cloudRelExtensionalityLeq nat) ++> cloudRelLeq
as existscloudnat2.
Proof.
    unfold cloudRelExtensionalityLeq,cloudExists,cloudRelLeq; simpl; firstorder.
Qed.

Add Morphism (@cloudExists (list nat)) with signature (@cloudRelExtensionalityLeq (list nat)) ==> cloudRelLeq
as existscloudllistnat2.
Proof.
    unfold cloudRelExtensionalityLeq,cloudExists,cloudRelLeq; simpl; firstorder.
Qed.


(* compatibility with tensor *)

Lemma TensorConj_strength : forall P Q,
	stateRelLeq (P ## cloudToState Q) (cloudToState((stateToCloud P //\\ Q))).
Proof.
  unfold stateRelTensor, stateRelLeq; simpl;intros. 
  destruct H as [s1 [s2 [s1' [s2' (Hdisj12,(Hjoins12,(Hdisj12',(Hjoins12',(Hp1,Hp2)))))]]]].
  split.  
  assert (H:subPfunction s1 s).
  rewrite Hjoins12; apply Pfunction_ext; assumption.
  assert (H':subPfunction s1' s').
  rewrite Hjoins12'; apply Pfunction_ext; assumption.
  eapply stateRel_cond; [apply Hp1 | apply H | apply H' |trivial].
  
  assert (H:subPfunction s2 s).
  rewrite Hjoins12;  rewrite join_disj_comm; intuition.
  apply Pfunction_ext; rewrite (disjointPfunction_comm s2 s1); assumption.
  assert (H':subPfunction s2' s').
  rewrite Hjoins12';  rewrite join_disj_comm; intuition.
  apply Pfunction_ext; rewrite (disjointPfunction_comm s2' s1'); assumption.
  eapply cloudRel_cond ; [apply Hp2 | apply H | apply H' | trivial].
Qed.

(*****************************)
(* some lemmas about ExistsP *)
(*****************************)

Lemma cloudExP_mono : forall (X: Type) (P Q : X -> cloudRel), 
  (forall x, cloudRelLeq (P x) (Q x)) -> (cloudRelLeq (Exists_c P) (Exists_c Q)).
Proof.
  unfold cloudRelLeq; simpl;  intros.
  witness a in H0.
  exists a.
  firstorder.
Qed.

Lemma cloudExP_rewrite : forall (X:Type) (P Q: X->cloudRel), 
  (forall x, cloudRelEq (P x) (Q x)) ->  (cloudRelEq (Exists_c P) (Exists_c Q)).
Proof.
  unfold cloudRelEq; intros; simpl.
  firstorder; exists x; assert (foo:= H x s s' p p' k);  firstorder.
Qed.

Lemma cloudExP_conj : forall (X: Type) (h:X -> cloudRel) P,
  cloudRelEq (Ex_c x, (h x //\\ P)) (Ex_c x, h x //\\ P).
Proof.
  split; unfold cloudRelLeq; simpl; intros;  firstorder.
Qed.


Lemma cloudExists_conjsplit : forall (X: Type) (h g:X -> cloudRel),
  cloudRelLeq (Ex_c x, (h x //\\ g x)) (Ex_c x, Ex_c y , (h x //\\ g y)).
Proof.
  unfold cloudRelLeq; simpl; intros.
  firstorder.
Qed.

Lemma cloudExP_comm : forall (X Y: Type) (h: X -> Y -> cloudRel),
  cloudRelEq (Ex_c x,Ex_c y, (h x y)) (Ex_c y, Ex_c x, (h x y)).
Proof.
  split; simpl; firstorder.
Qed.

Lemma cloudExistsLeq : forall (X:Type) (h:X->cloudRel) P Q,
 (exists x : X, cloudRelLeq Q ((h x) //\\ P))
 -> cloudRelLeq Q ((Ex_c x, h x) //\\ P).
Proof.
  unfold cloudRelLeq, cloudExists; simpl; firstorder.
  exists x.
  hreplace H with  (H s s' p p' k H0).
  destruct H; assumption.
  hreplace H with  (H s s' p p' k H0).
  destruct H; assumption.
Qed.

Lemma cloudForallLeq : forall (X:Type) (h:X->cloudRel) P Q,
 (forall x : X, cloudRelLeq ((h x)//\\P) Q)
 -> cloudRelLeq ((Ex_c x, h x)//\\P) Q .
Proof. 
  unfold cloudRelLeq, cloudExists; simpl;  firstorder.
Qed.

Lemma cloudExistsLeq_single : forall (X:Type) (h:X->cloudRel) Q,
 (exists x : X, cloudRelLeq Q (h x))
 -> cloudRelLeq Q ((Ex_c x, h x)).
Proof.
  unfold cloudRelLeq, cloudExists;simpl; firstorder.
Qed.

Lemma cloudForallLeq_single : forall (X:Type) (h:X->cloudRel) Q,
 (forall x : X, cloudRelLeq (h x) Q)
 -> cloudRelLeq ((Ex_c x, h x)) Q .
Proof. 
  unfold cloudRelLeq, cloudExists; simpl; firstorder.
Qed.

Lemma cloudExP_extract : forall (X:Type) (h:X -> cloudRel),
  stateRelEq (Ex x, cloudToState (h x)) 
                         (cloudToState (Ex_c x, (h x))).
Proof.
  split; unfold stateRelLeq; simpl; intros;  firstorder.
Qed.


(************)
(* ForAllP  *)
(************)

Definition cloudForall (X : Type) (h : X -> cloudRel ) : cloudRel.
  intros.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun k => forall a: X, h a s s' p p' k) _).
  intros.
  hreplace H with (H a).
  apply (cloudRel_cond (h a) s1 s2 s1' s2' p1 p2 k1 k2);  assumption.
Defined.

Implicit Arguments cloudForall [X].

Notation "'ForAll_c' h" := (cloudForall  h ) (at level 70).

Notation "'Forall_c' x , P" := (ForAll_c (fun x=> P) ) (at level 70, x at next level).

Add Morphism (@cloudForall nat) with signature (@cloudRelExtensionality nat) ==> cloudRelEq
as forallPcloudnat.
Proof.
    unfold cloudRelEq, cloudRelExtensionality,cloudForall; simpl; intuition.
    rewrite <- H; intuition.
    rewrite H; intuition.
Qed.

Add Morphism (@cloudForall (list nat)) with signature (@cloudRelExtensionality (list nat)) ==> cloudRelEq
as foralcloudllistnat.
Proof.
  unfold cloudRelEq, cloudRelExtensionality,cloudForall; simpl; intuition. 
  rewrite <- H; intuition.
  rewrite H; intuition.
Qed.


Add Morphism (@cloudForall nat) with signature (@cloudRelExtensionalityLeq nat) ++> cloudRelLeq
as forallPcloudnat2.
Proof.
    unfold cloudRelExtensionalityLeq,cloudForall,cloudRelLeq; simpl; intuition.
Qed.

Add Morphism (@cloudForall (list nat)) with signature (@cloudRelExtensionalityLeq (list nat)) ==> cloudRelLeq
as foralcloudllistnat2.
Proof.
    unfold cloudRelExtensionalityLeq,cloudForall,cloudRelLeq; simpl; intuition.
Qed.

Lemma cloudForall_ForallLeq : forall (X:Type) (h:X->cloudRel) P Q,
 cloudRelLeq Q ((Forall_c x, h x) //\\ P)
 -> (forall x : X, cloudRelLeq Q ((h x) //\\ P)).
Proof.
  unfold cloudRelLeq, cloudForall; simpl; firstorder.
  hreplace H with  (H s s' p p' k H0).
  firstorder.
  hreplace H with  (H s s' p p' k H0).
  intuition.
Qed.

Lemma cloudForall_ExistsLeq : forall (X:Type) (h:X->cloudRel) P Q,
 (exists x : X, cloudRelLeq ((h x)//\\P) Q)
 -> cloudRelLeq ((Forall_c x, h x)//\\P) Q .
Proof. 
  unfold cloudRelLeq, cloudExists; simpl;  firstorder.
Qed.

Lemma cloudForall_ForallLeq_single : forall (X:Type) (h:X->cloudRel) Q,
 (forall x : X, cloudRelLeq Q (h x)) <-> cloudRelLeq Q ((Forall_c x, h x)).
Proof.
  unfold cloudRelLeq, cloudExists;simpl; firstorder.
Qed.

Lemma cloudForall_ExistsLeq_single : forall (X:Type) (h:X->cloudRel) Q,
 (exists x : X, cloudRelLeq (h x) Q) -> cloudRelLeq ((Forall_c x, h x)) Q .
Proof. 
  unfold cloudRelLeq, cloudExists; simpl; firstorder.
Qed.

Lemma cloudForall_extract : forall (X:Type) (h:X -> cloudRel),
  stateRelEq (Forall x, cloudToState (h x)) 
                         (cloudToState (Forall_c x, (h x))).
Proof.
  split; unfold stateRelLeq; simpl; intros;  firstorder.
Qed.

Lemma cloudForall_conj : forall (X: Type) (h:X -> cloudRel) P,
  X -> (cloudRelEq ((Forall_c x, (h x)) //\\ P) (Forall_c x, (h x) //\\ P)).
Proof.
  unfold cloudRelEq; simpl; intros;  firstorder.
Qed.

Lemma cloudForall_mono : forall (X: Type) (P Q : X -> cloudRel), 
  (forall x, cloudRelLeq (P x) (Q x)) -> (cloudRelLeq (ForAll_c P) (ForAll_c Q)).
Proof.
  unfold cloudRelLeq; simpl;  intros.
  apply (H a); apply (H0 a).
Qed.

Lemma cloudForall_rewrite : forall (X:Type) (P Q: X->cloudRel), 
  (forall x, cloudRelEq (P x) (Q x)) ->  (cloudRelEq (ForAll_c P) (ForAll_c Q)).
Proof.
  unfold cloudRelEq; simpl; firstorder;  hreplace H with (H a s s' p p' k); firstorder.
Qed.



(* *******************  Modality ********************* *)

Definition Later_c (P:cloudRel) : cloudRel.
  intros P.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun k => 
    forall j, j < k -> P s s' p p' j) _).
  intuition.
  specialize (H j).
  eapply cloudRel_cond; intuition.
Defined.

Definition Necessity_c (P:cloudRel) : cloudRel.
  intros P.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun k => 
    forall j, j <= k -> P s s' p p' j) _).
  intuition.
  specialize (H j).
  eapply cloudRel_cond; intuition.
Defined.

