(* compiler.v

   Basic lambda calculus compiler to the machine code defined in programs.v

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import programs.
Require Export allocator_code.
Require Export lambda.

(* --------------- 
   Basic functions 
   --------------- *)

Fixpoint iterate (A : Type) (af: A -> A) (a0 : A) (n:nat) {struct n} : A :=
  match n with 0 => a0 | (S m) => af (iterate A af a0 m) end.

Fixpoint numlist (A:Set) (l:list A) (n:nat) {struct l} : (list (nat * A)) :=
  match l with
    | nil => nil
    | a::l1 => (n,a):: (numlist A l1 (n+1))
  end.


(* --------------
   Generated code
   -------------- *)

Inductive is_application : Set := 
         | no_app : is_application
         | app    : is_application.

Definition make_opt_code (f : nat -> nat -> nat -> list instruction) (stack_size label :nat) (is_app : is_application ):= 
                        Some (fun closure_entry => fun alloc => fun l => 
                                            ((f l alloc closure_entry), stack_size , l + label), is_app).

(*** Closures and applications ***)

Definition closure_make (label alloc closure_entry: nat)  := 
  [(i_move (d_imm argreg) (s_cst 2)) ;
   (i_move (d_imm retreg) (s_cst (label + 3))) ;
   (i_branch (s_cst alloc)) ;
   (i_op plus (d_imm spreg) (s_imm spreg) (s_cst 1)) ;
   (i_move (d_ind spreg) (s_imm retreg)) ;
   (i_move (d_ind retreg) (s_cst closure_entry)) ;
   (i_move (d_indo 1 retreg) (s_imm envreg)) ].

Definition lambda_function := make_opt_code closure_make 1 7 app.

 (* [arg] <- 2                 allocate heap
    [ret] <- label + 3
    jmp alloc
    [sp] <- [sp] + 1        shift the stack pointer
    [[sp]] <- [ret]            push heap address onto stack
    [[ret]] <- closure_entry   set code pointer
    [[ret]+1] <- [env]         set pointer to rest of environment
 *)

Definition closure_rec_make (label alloc closure_entry: nat) :=
 [(i_move (d_imm argreg) (s_cst 2)) ;
  (i_move (d_imm retreg) (s_cst (label + 3))) ;
  (i_branch (s_cst alloc)) ;
  (i_move (d_indo 1 retreg) (s_imm envreg));
  (i_move (d_imm envreg) (s_imm retreg))].

Definition fixpoint_function := make_opt_code closure_rec_make 0 5 no_app.

 (* [arg] <- 2                 allocate heap
    [ret] <- label + 3
    jmp alloc
    [[ret]+1] <- [env]         set pointer to rest of environment
    [env] <- [ret]             set pointer to the environment with a hole
 *)

Definition closure_rec_make2 (label alloc closure_entry: nat) :=
 [(i_move (d_ind envreg) (s_ind spreg));
  (i_move (d_imm envreg) (s_indo 1 envreg))].

Definition fixpoint_function2 := make_opt_code closure_rec_make2 0 2 no_app.


 (* [[env]] <- [[sp]]          push the ptr to the function  into the hole 
    [env] <- [[env]+1]         restore the former environment
 *)

Definition application_code  (label alloc closure_entry: nat) :=
 [(i_move (d_indo 4 envreg) (s_cst (label + 3))) ;
  (i_move (d_imm workreg) (s_ind spreg)) ;
   (i_branch (s_ind workreg))].

Definition application_function := make_opt_code application_code 0 3 no_app.

 (* [[env]+4] <- label + 3  save application return label
    [wk] <- [[sp]]
    jmp [[wk]]              jump to function code *)

Definition closure_prolog_size := 11.
Definition closure_prolog label alloc stack_size :=
 [(i_move (d_imm argreg) (s_cst (stack_size + 5))) ;
  (i_move (d_imm retreg) (s_cst (label + 3))) ;
  (i_branch (s_cst alloc)) ;
  (i_move (d_imm workreg) (s_ind spreg)) ;
  (i_move (d_indo 1 retreg) (s_indo 1 workreg)) ;
  (i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)) ;
  (i_move (d_ind retreg) (s_ind spreg)) ;
  (i_move (d_indo 2 retreg) (s_imm envreg)) ;
  (i_move (d_indo 3 retreg) (s_imm spreg)) ;
  (i_move (d_imm envreg) (s_imm retreg)) ;
  (i_op plus (d_imm spreg) (s_imm retreg) (s_cst 4))].

 (* [arg] <- stack_size + 5  allocate stack frame
    [ret] <- label + 3       store the return address
    jmp alloc                call the allocator
    [wk] <- [[sp]]           take function pointer
    [[ret]+1] <- [[wk]+1]    store environment pointer
    [sp] <- [sp]-1           focus on argument
    [[ret]] <- [[sp]]        store argument's value in envrionment
    [[ret]+2] <- [env]       store calling env
    [[ret]+3] <- [sp]        store calling sp
    [env] <- [ret]           switch to new environment
    [sp] <- [ret] + 4        switch to new stack
 *)

Definition closure_epilog_size := 9.
Definition closure_epilog label stack_size dealloc :=
 [(i_move (d_imm workreg) (s_ind spreg)) ;
  (i_move (d_imm spreg) (s_indo 3 envreg)) ;
  (i_move (d_ind spreg) (s_imm workreg)) ;
  (i_op plus (d_imm workreg) (s_imm envreg) (s_cst 2)) ;
  (i_move (d_imm envreg) (s_indo 2 envreg)) ;
  (i_move (d_imm argreg) (s_cst (stack_size + 3))) ;
  (i_move (d_imm retreg) (s_cst (label + 8))) ;
  (i_branch (s_cst dealloc)) ;
  (i_branch (s_indo 4 envreg))].



 (* [wk] <- [[sp]]         save return value
    [sp] <- [[env]+3]      restore old sp
    [[sp]] <- [wk]         push return value
    [wk] <- [env] +2       save current frame
    [env] <- [[env]+2]     restore old frame
    [arg] <- stack_size+3  deallocate frame
    [ret] <- label + 8
    jmp dealloc
    jmp [[env]+4]          return to application code *)

(*** Other compiled constructs ***)

Definition var_code (n label alloc closure_entry: nat) :=
  (i_op plus (d_imm spreg) (s_imm spreg) (s_cst 1)) ::
  (i_move (d_imm workreg) (s_imm envreg)) ::
  (iterate (list instruction)
    (fun l => (i_move (d_imm workreg) (s_indo 1 workreg)) :: l)
    ((i_move (d_ind spreg) (s_ind workreg)) :: nil)
    n).
Definition var_function n := make_opt_code (var_code n) 1 (n+3) no_app.

 (*[sp] <- [sp] + 1 
    [wrk] <- [env]         lookup x in environment
    [wrk] <- [[wrk] + 1]
     ..<n copies>..
    [wrk] <- [[wrk] + 1]
    [[sp]] <- [[wrk]]      push x
 *)

Definition int_code (n label alloc closure_entry: nat):=
 [(i_op plus (d_imm spreg) (s_imm spreg) (s_cst 1)) ;
  (i_move (d_ind spreg) (s_cst n))].
Definition int_function n := make_opt_code (int_code n) 1 2 no_app.
 (* [sp] <- [sp] + 1
     [[sp]] <- n        push n 
 *)

Definition bool_code (b:bool) (label alloc closure_entry: nat):=
 [(i_op plus (d_imm spreg) (s_imm spreg) (s_cst 1));
 (i_move (d_ind spreg) (s_cst (if b then 1 else 0)))].
Definition bool_function b := make_opt_code (bool_code b) 1 2 no_app.

 (* [sp] <- [sp]+1
     [[sp]] <- 0/1      push b 
 *)

Definition op_code op (label alloc closure_entry: nat) :=
 [(i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)) ;
  (i_op op (d_ind spreg) (s_indo 1 spreg) (s_ind spreg))].
Definition op_function op := make_opt_code (op_code op) 0 2 no_app.

 (* [sp] <- [sp] - 1            pop 
    [[sp]] <- op [[sp]+1] [[sp]] compute op
 *)

Definition gt_code (label alloc closure_entry: nat) :=
 [(i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)) ;
  (i_op minus (d_imm workreg) (s_indo 1 spreg) (s_ind spreg)) ;
  (i_brz (s_imm workreg) (s_cst (label + 4))) ;
  (i_move (d_imm workreg) (s_cst 1)) ;
  (i_move (d_ind spreg) (s_imm workreg))].
Definition gt_function := make_opt_code gt_code 0 5 no_app.

(* [sp] <- [sp] - 1            pop
   [wrk] <- [[sp]] - [[sp]+1]
   i_brz [wrk] label + 4       skip if zero      
   [wrk] <- 1
   [[sp]] <- [wrk]
 *)

Definition pair_code (label alloc closure_entry: nat) :=
  [(i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)) ;
  (i_move (d_imm argreg) (s_cst 2)) ;
  (i_move (d_imm retreg) (s_cst (label + 4))) ;
  (i_branch (s_cst alloc)) ;
  (i_move (d_ind retreg) (s_ind spreg)) ;
  (i_move (d_indo 1 retreg) (s_indo 1 spreg)) ;
  (i_move (d_ind spreg) (s_imm retreg))].

Definition pair_function := make_opt_code pair_code 0 7 no_app.

 (* [sp] <- [sp] - 1         pop
    [arg] <- 2              allocate space
    [ret] <- label + 4
    jmp alloc
    [[ret]] <- [[sp]]        store expressions
    [[ret]+1] <- [[sp]+1]
    [[sp]] <- [ret]
 *)


Definition first_code (label alloc closure_entry: nat) :=
 [(i_move (d_imm workreg) (s_ind spreg)) ;
  (i_move (d_ind spreg) (s_indo 1 workreg))].
Definition first_function := make_opt_code first_code 0 2 no_app.

 (* [wk] <- [[sp]]        grap the pointer of the pair
    [[sp]] <- [[wk]+1]    push first element (which is actually the second :-)
 *)

Definition second_code (label alloc closure_entry: nat) :=
 [(i_move (d_imm workreg) (s_ind spreg)) ;
  (i_move (d_ind spreg) (s_ind workreg))].

Definition second_function := make_opt_code second_code 0 2 no_app.

 (*[wk] <- [[sp]]        grap the pointer of the pair
    [[sp]] <- [[wk]]     push second element
  *)

(*** Putting it all together ***)

Definition intro_code_size := 7.

Definition intro_code label stack_size init alloc :=
  [(i_move (d_imm retreg) (s_cst (label + 2))) ;
   (i_branch (s_cst init)) ;
   (i_move (d_imm argreg) (s_cst (stack_size + 5))) ;
   (i_move (d_imm retreg) (s_cst (label + 5))) ;
   (i_branch (s_cst alloc)) ;
   (i_move (d_imm envreg) (s_imm retreg)) ;
   (i_op plus (d_imm spreg) (s_imm retreg) (s_cst 4))].
  
 (* [ret] <- label + 2
    jmp init
    [arg] <- stack_size + 5
    [ret] <- label + 5
    jmp alloc
    [env] <- [ret]
    [sp] <- [ret]+5 *)

Definition retval_code_size := 1.
Definition retval_code := (* should really jump back to calling function *)
  [(i_move (d_imm retreg) (s_ind spreg))].

Definition if_1_code (label alloc closure_entry: nat) :=
  [ (i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)) ;
    (i_brz (s_indo 1 spreg) (s_cst (label + 1)))].
(* [sp] <- [sp] - 1         pop
    branch on the value of the boolean [[sp]+1] *)

Definition if_1_function := make_opt_code if_1_code 0 1 no_app.
Definition if_1 label := if_1_code label 0 0.

Definition if_2_code (label alloc closure_entry: nat) := 
  [(i_branch (s_cst label))].
Definition if_2_function := make_opt_code if_2_code 0 1 no_app.
Definition if_2 label := if_2_code label 0 0.

(* --------
   Compiler
   -------- *)

Definition append_code_table (code_table code_table' : nat -> list instruction * nat * nat) (label:nat) := 
  match (code_table label) with (code, stack_used, next_free) =>
  match (code_table' next_free) with (code', stack_used', next_free') =>
    (code ++ code', max stack_used (stack_used' + (min stack_used 1)), next_free')
  end end.

Definition code_table_nil  : nat -> list instruction * nat * nat := fun l => (nil , 0 , l).

Definition default_code_table (code:list instruction) (label:nat) := (code_table_nil , code , label).

Definition type_optional_function_code := option
         ((nat -> nat -> nat -> list instruction * nat * nat) * is_application).

Fixpoint compile (e : Exp) (aux_code: list instruction) (aux_next : nat) (alloc dealloc:nat) {struct e} : ((nat -> list instruction * nat * nat) * list instruction * nat) := 
  let monad_expr (optional_function_code : type_optional_function_code) 
    e alloc dealloc previous_code:= 
    match previous_code with (code',aux_code',aux_next') =>
    match optional_function_code with 
      | None => match e with  
            | Some e => 
              match (compile e aux_code' aux_next' alloc dealloc) with 
                (code'',aux_code'',aux_next'') => 
                (append_code_table code' code'' , aux_code'', aux_next'')
              end
                  | None => previous_code
                end
      | Some (f, no_app) =>   (append_code_table code' (f aux_next' alloc), aux_code', aux_next')
      | Some (f, app) => 
        match code' (aux_next' + closure_prolog_size) with (closure_code, closure_stack, closure_next) =>
          (append_code_table code_table_nil (f aux_next' alloc) ,
            aux_code' ++ (closure_prolog aux_next' alloc closure_stack)
            ++ closure_code ++ (closure_epilog closure_next closure_stack dealloc),
            closure_next + closure_epilog_size)
        end
    end 
    end
  in
  match e with
  | LAMBDA e => monad_expr lambda_function None alloc dealloc
                      (monad_expr None (Some e) alloc dealloc (default_code_table aux_code aux_next))
		
  | FIX e => monad_expr fixpoint_function2 None alloc dealloc
                    (monad_expr None (Some e) alloc dealloc
                    (monad_expr fixpoint_function None alloc dealloc (default_code_table aux_code aux_next)))

  | APP e1 e2 => monad_expr application_function None alloc dealloc 
                       (monad_expr None (Some e1) alloc dealloc  
                       (monad_expr None (Some e2) alloc dealloc (default_code_table aux_code aux_next)))

  | VAR m => monad_expr (var_function m)  None alloc dealloc (default_code_table aux_code aux_next)

  | INT n => monad_expr (int_function n)  None alloc dealloc (default_code_table aux_code aux_next)
 
  | BOOL b => monad_expr (bool_function b) None alloc dealloc (default_code_table aux_code aux_next)

  | OP op e1 e2 => monad_expr (op_function op) None alloc dealloc
                     (monad_expr None (Some e1) alloc dealloc  
                     (monad_expr None (Some e2) alloc dealloc (default_code_table aux_code aux_next)))

  | GT e1 e2 => monad_expr gt_function None alloc dealloc
                     (monad_expr None (Some e1) alloc dealloc  
                     (monad_expr None (Some e2) alloc dealloc (default_code_table aux_code aux_next)))

  | P e1 e2 => monad_expr pair_function None alloc dealloc
                     (monad_expr None (Some e1) alloc dealloc  
                     (monad_expr None (Some e2) alloc dealloc (default_code_table aux_code aux_next)))

  | FST e => monad_expr first_function None alloc dealloc
                   (monad_expr None (Some e) alloc dealloc (default_code_table aux_code aux_next))

  | SND e => monad_expr second_function None alloc dealloc
                   (monad_expr None (Some e) alloc dealloc (default_code_table aux_code aux_next))

  | IF_THEN_ELSE e1 e2 e3 => 
    match compile e1 aux_code aux_next alloc dealloc with (code',aux_code',aux_next') =>
    match compile e2 aux_code' aux_next' alloc dealloc with (code'',aux_code'',aux_next'') =>
    match compile e3 aux_code'' aux_next'' alloc dealloc with (code''',aux_code''',aux_next''') =>
      ((fun l => match code' l with (code_code, code_stack, code_next) =>
		 match code'' (code_next + 2) with (code_code', code_stack', code_next') =>
		 match code''' (code_next' + 1) with (code_code'', code_stack'', code_next'') =>
		   (code_code ++ if_1 code_next' ++ code_code' ++ if_2 code_next'' ++ code_code'',
		     code_stack + (max  code_stack' code_stack'') , (* XXX *)
		     code_next'')
		 end
		 end
		 end),
	  aux_code''',aux_next''')
    end
    end
    end
    
  end.


Definition compile' e start init alloc dealloc :=
  match compile e nil start alloc dealloc with (main, aux_code, aux_next) =>
    match main (aux_next + intro_code_size) with (global_code, main_stack, main_next) =>
      (aux_next, aux_code ++ (intro_code aux_next main_stack init alloc) ++ global_code ++ retval_code)
    end
  end.
  
Definition compiler e :=
  let init := 0 in
  let alloc := 10 in
  let dealloc := 30 in
  let start := 100 in
  let (entry, code) := compile' e start init alloc dealloc in
    (entry, halting_extension 
      ( fragfromprog (progfromlist (ntinitinstructions init)) init ;;
	fragfromprog (progfromlist (ntallocinstructions alloc)) alloc ;;
	fragfromprog (progfromlist (ntdeallocinstructions dealloc)) dealloc ;;
        fragfromprog (progfromlist code) start)).

Definition test_code e :=
  let (entry,code) := compile' e 100 0 10 30 in
    (entry, numlist instruction code 100).

Definition test_state e n steps :=
  let (pc, p) := compiler e in
  let (s',pc') := (execute_bounded steps p (fun n => 0) pc)
  in (dumpstateupto n s', pc').

Definition test_trace e steps :=
  let (pc,p) := compiler e in
  numlist nat (program_trace p (fun n => 0) pc steps) 1.

Definition test e steps :=
  match head (fst (test_state e 1 steps)) with
    | Some (0, n) => Some n
    | _ => None
  end.


Unset Implicit Arguments.

(*** Examples ***)

Definition eg1 := (IF_THEN_ELSE (BOOL true)(INT 71) (INT 93)). 

Definition eg2 := FST (P (eg1) (INT 53) ).

Eval compute in test eg2 2000.

Definition identity := LAMBDA (VAR 0).

Definition const_32 := LAMBDA (INT 32).

Definition eg3 := FST (P (INT 94) (BOOL true)).

Definition succ := LAMBDA (OP plus (VAR 0) (INT 1)).

Definition double := LAMBDA (OP plus (VAR 0) (VAR 0)).

(* compute min { 2^n * x | 2^n * x > 100 }*)

Definition redouble := FIX (LAMBDA (IF_THEN_ELSE (GT (VAR 0) (INT 100))
                    (VAR 0) (APP (VAR 1) (APP double (VAR 0))))).

Definition Fibonacci := FIX (LAMBDA (IF_THEN_ELSE (GT (FST (VAR 0)) (INT 25))
                   (FST(VAR 0))  
                   (APP (VAR 1) (P (OP plus (FST(VAR 0)) (SND(VAR 0))) (FST(VAR 0)))))).

Definition Sum := LAMBDA (LAMBDA (OP plus (VAR 0) (VAR 1))).

Definition eg4 := APP redouble (INT 1).

Definition eg5 := APP Fibonacci (P (INT 1) (INT 1)).

Definition Go_wrong := APP (APP Sum (INT 3)) (INT 5).

Eval compute in test Go_wrong 2000.

(*Eval compute in test_code eg5.

Eval compute in test_state eg5 40 75.
*)
Eval compute in test eg5 2000.

Eval compute in test eg4 2000.


Tactic Notation (at level 0) "surjective_pairing_one_arg"  constr(e) constr(aux_code) 
	constr(start) constr(alloc) constr(dealloc) "in" "goal":=   
  let code1_gen := fresh "code1_gen" in 
    set (code1_gen := compile e aux_code (start+length aux_code) alloc dealloc);
  rewrite (surjective_pairing code1_gen);
  rewrite (surjective_pairing (fst code1_gen));
  unfold append_code_table ; simpl.

Tactic Notation (at level 0) "surjective_pairing_one_arg"  constr(e) constr(aux_code) 
	constr(start) constr(alloc) constr(dealloc) "in" constr(H):=   
  let code1_gen := fresh "code1_gen" in 
    set (code1_gen := compile e aux_code (start+length aux_code) alloc dealloc) in H;
  rewrite (surjective_pairing code1_gen) in H;
  rewrite (surjective_pairing (fst code1_gen)) in H;
  unfold append_code_table in H; simpl in H.

Tactic Notation (at level 0) "surjective_pairing_one_arg"  constr(e) constr(code_gen) constr(label) "in" "goal":=   
  let code1 := fresh "code1" in 
    set (code1 :=  fst (fst code_gen) label);
  rewrite (surjective_pairing code1);
  rewrite (surjective_pairing (fst code1)); simpl.

Tactic Notation (at level 0) "surjective_pairing_one_arg"  constr(e) constr(code_gen) constr(label) "in" constr(H):=   
  let code1 := fresh "code1" in 
    set (code1 :=  fst (fst code_gen) label) in H;
  rewrite (surjective_pairing code1) in H;
  rewrite (surjective_pairing (fst code1)) in H; simpl in H.

Tactic Notation (at level 0) "surjective_pairing_function" constr(e) constr(aux_code) 
	constr(start) constr(alloc) constr(dealloc) "in" "goal"  :=  
  let code1_gen := fresh "code1_gen" in 
    set (code1_gen := compile e aux_code (start+length aux_code) alloc dealloc);
  rewrite (surjective_pairing code1_gen);
  rewrite (surjective_pairing (fst code1_gen));
  unfold append_code_table ; simpl.

Tactic Notation (at level 0) "surjective_pairing_function" constr(e) constr(aux_code) 
	constr(start) constr(alloc) constr(dealloc) "in" constr(H)  :=  
  let code1_gen := fresh "code1_gen" in 
    set (code1_gen := compile e aux_code (start+length aux_code) alloc dealloc) in H;
  rewrite (surjective_pairing code1_gen) in H;
  rewrite (surjective_pairing (fst code1_gen)) in H;
  unfold append_code_table in H; simpl in H.

Tactic Notation (at level 0) "surjective_pairing_function" constr(e) constr(code_gen) "in" "goal"  :=  
  let code1 := fresh "code1" in 
    set (code1 :=  fst (fst code_gen) (snd code_gen + closure_prolog_size));
  rewrite (surjective_pairing code1);
  rewrite (surjective_pairing (fst code1)); simpl.

Tactic Notation (at level 0) "surjective_pairing_function" constr(e) constr(code_gen) "in" constr(H)  :=  
  let code1 := fresh "code1" in 
    set (code1 :=  fst (fst code_gen) (snd code_gen + closure_prolog_size)) in H;
  rewrite (surjective_pairing code1) in H;
  rewrite (surjective_pairing (fst code1)) in H;  simpl in H.


Tactic Notation (at level 0) "surjective_pairing_two_arg" constr(e1) constr(e2) 
	constr(aux_code) constr(start) constr(alloc) constr(dealloc) "in" "goal" := 
  let code2_gen := fresh "code2_gen" in 
    set (code2_gen := compile e2 aux_code (start+length aux_code) alloc dealloc);
  rewrite (surjective_pairing code2_gen);
  rewrite (surjective_pairing (fst code2_gen));
  let code1_gen := fresh "code1_gen" in 
    set (code1_gen := compile e1 (snd(fst code2_gen)) (snd code2_gen) alloc dealloc );
  rewrite (surjective_pairing code1_gen);
  rewrite (surjective_pairing (fst code1_gen));
  unfold append_code_table; simpl.

Tactic Notation (at level 0) "surjective_pairing_two_arg" constr(e1) constr(e2) 
	constr(aux_code) constr(start) constr(alloc) constr(dealloc) "in" constr(H):=   
  let code2_gen := fresh "code2_gen" in 
    set (code2_gen := compile e2 aux_code (start+length aux_code) alloc dealloc) in H;
  rewrite (surjective_pairing code2_gen) in H;
  rewrite (surjective_pairing (fst code2_gen)) in H;
  let code1_gen := fresh "code1_gen" in 
    set (code1_gen := compile e1 (snd(fst code2_gen)) (snd code2_gen) alloc dealloc ) in H;
  rewrite (surjective_pairing code1_gen) in H;
  rewrite (surjective_pairing (fst code1_gen)) in H;
  unfold append_code_table in H; simpl in H.

Tactic Notation (at level 0) "surjective_pairing_two_arg" constr(e1) constr(e2) 
	constr(code1_gen) constr(code2_gen) constr(label) "in" "goal" := 
  let code2 := fresh "code2" in 
    set (code2 := fst(fst code2_gen) (label));
  rewrite (surjective_pairing code2);
  rewrite (surjective_pairing (fst code2));
  let code1 := fresh "code1" in 
    set (code1 := fst(fst(code1_gen)) (snd code2));
  rewrite (surjective_pairing code1);
  rewrite (surjective_pairing (fst code1)); simpl.

Tactic Notation (at level 0) "surjective_pairing_two_arg" constr(e1) constr(e2) 
	constr(code1_gen) constr(code2_gen) constr(label) "in" constr(H):=   
  let code2 := fresh "code2" in 
    set (code2 := fst(fst code2_gen) (label)) in H;
  rewrite (surjective_pairing code2) in H;
  rewrite (surjective_pairing (fst code2)) in H;
  let code1 := fresh "code1" in 
    set (code1 := fst(fst(code1_gen)) (snd code2)) in H;
  rewrite (surjective_pairing code1) in H;
  rewrite (surjective_pairing (fst code1)) in H; simpl in H.

Tactic Notation (at level 0) "surjective_pairing_three_arg" constr(e1) constr(e2) constr(e3) 
	constr(aux_code) constr(start) constr(alloc) constr(dealloc) "in" "goal":= 
  let code1_gen := fresh "code1_gen" in 
    set (code1_gen := compile e1 aux_code (start+length aux_code) alloc dealloc);
  rewrite (surjective_pairing code1_gen);
  rewrite (surjective_pairing (fst code1_gen));
  let code2_gen := fresh "code2_gen" in 
    set (code2_gen := compile e2 (snd(fst code1_gen)) (snd code1_gen) alloc dealloc );
  rewrite (surjective_pairing code2_gen);
  rewrite (surjective_pairing (fst code2_gen));
  let code3_gen := fresh "code3_gen" in 
    set (code3_gen := compile e3 (snd(fst code2_gen)) (snd code2_gen) alloc dealloc );
  rewrite (surjective_pairing code3_gen);
  rewrite (surjective_pairing (fst code3_gen));
    unfold append_code_table; simpl.

Tactic Notation (at level 0) "surjective_pairing_three_arg" constr(e1) constr(e2) constr(e3) 
	constr(aux_code) constr(start) constr(alloc) constr(dealloc)  "in" constr(H) :=   
  let code1_gen := fresh "code1_gen" in 
    set (code1_gen := compile e1 aux_code (start+length aux_code) alloc dealloc) in H;
  rewrite (surjective_pairing code1_gen) in H;
  rewrite (surjective_pairing (fst code1_gen)) in H;
  let code2_gen := fresh "code2_gen" in 
    set (code2_gen := compile e2 (snd(fst code1_gen)) (snd code1_gen) alloc dealloc ) in H;
  rewrite (surjective_pairing code2_gen) in H;
  rewrite (surjective_pairing (fst code2_gen)) in H;
  let code3_gen := fresh "code3_gen" in 
    set (code3_gen := compile e3 (snd(fst code2_gen)) (snd code2_gen) alloc dealloc ) in H;
  rewrite (surjective_pairing code3_gen) in H;
  rewrite (surjective_pairing (fst code3_gen)) in H;
    unfold append_code_table in H; simpl in H.

Tactic Notation (at level 0) "surjective_pairing_three_arg2" constr(e1) constr(e2) constr(e3) 
	constr(code1_gen) constr(code2_gen) constr(code3_gen) constr(label) "in" "goal":= 
  let code1 := fresh "code1" in 
    set (code1 := fst(fst code1_gen) label);
  rewrite (surjective_pairing code1);
  rewrite (surjective_pairing (fst code1));
  let code2 := fresh "code2" in 
    set (code2 := fst(fst(code2_gen)) (snd code1+2));
  rewrite (surjective_pairing code2);
  rewrite (surjective_pairing (fst code2));
  let code3 := fresh "code3" in 
    set (code3 := fst(fst(code3_gen)) (snd code2+1));
  rewrite (surjective_pairing code3);
  rewrite (surjective_pairing (fst code3));  simpl.

Tactic Notation (at level 0) "surjective_pairing_three_arg2" constr(e1) constr(e2) constr(e3) 
	constr(code1_gen) constr(code2_gen) constr(code3_gen) constr(label) "in" constr(H) :=   
  let code1 := fresh "code1" in 
    set (code1 := fst(fst code1_gen) label) in H;
  rewrite (surjective_pairing code1) in H;
  rewrite (surjective_pairing (fst code1)) in H;
  let code2 := fresh "code2" in 
    set (code2 := fst(fst(code2_gen)) (snd code1+2)) in H;
  rewrite (surjective_pairing code2) in H;
  rewrite (surjective_pairing (fst code2)) in H;
  let code3 := fresh "code3" in 
    set (code3 := fst(fst(code3_gen)) (snd code2+1)) in H;
  rewrite (surjective_pairing code3) in H;
  rewrite (surjective_pairing (fst code3)) in H; simpl in H.

Set Implicit Arguments.


Lemma length_list : forall n, length((iterate (list instruction)
    (fun l => (i_move (d_imm workreg) (s_indo 1 workreg)) :: l)
    ((i_move (d_ind spreg) (s_ind workreg)) :: nil)
    n)) = n+1.
Proof.
  induction n; [ trivial | simpl; omega].
Qed.


Lemma compile_length : forall e alloc dealloc start aux_code label,
	snd (compile e aux_code (start+length aux_code) alloc dealloc) =
        start + length (snd(fst (compile e aux_code (start+length aux_code) alloc dealloc))) /\
        snd (fst(fst(compile e aux_code (start+length aux_code) alloc dealloc)) label) = 
        label + length (fst(fst(fst(fst(compile e aux_code (start+length aux_code) alloc dealloc)) label))).	
Proof.
  
  let strat_0 := split;(omega) with
   strat_VAR := try rewrite length_list;(omega) with
      strat_1 e label := surjective_pairing_one_arg e aux_code start alloc dealloc in goal;
      surjective_pairing_one_arg e code1_gen label in goal; split;
     hreplace IHe with (IHe alloc dealloc start aux_code label); unfold code1, code1_gen in *;
     destruct IHe as (IHe1,IHe2);
     [rewrite IHe1;(omega) | rewrite IHe2; rewrite IHe1; rewrite app_length; simpl; (omega)] with
  strat_function e := surjective_pairing_function e aux_code start alloc dealloc in goal;
     surjective_pairing_function e code1_gen in goal; split;
     try (omega); hreplace IHe with (IHe alloc dealloc start aux_code (snd code1_gen+closure_prolog_size));
     destruct IHe as (IH1,IH2) ; repeat rewrite app_length;simpl; rewrite app_length; simpl;
     unfold code1, code1_gen, closure_epilog_size, closure_prolog_size in *; rewrite IH2; rewrite IH1; (omega); 
     intros; simpl; surjective_pairing_function e aux_code start alloc dealloc in goal with
  strat_2 e1 e2 := surjective_pairing_two_arg e1 e2 aux_code start alloc dealloc in goal;
     surjective_pairing_two_arg e1 e2 code1_gen code2_gen label in goal;
     split;  hreplace IHe2 with (IHe2 alloc dealloc start aux_code label);
     hreplace IHe1 with (IHe1 alloc dealloc start (snd(fst code2_gen)) (label+length(fst (fst code2))));
     destruct IHe2 as (IHe21,IHe22); destruct IHe1 as (IHe11,IHe12);
     [unfold code1_gen, code2_gen,code2,code1 in *; rewrite IHe21; rewrite IHe11;  (omega) |
     unfold code1, code1_gen, code2, code2_gen in*;  rewrite IHe22; rewrite IHe21;
     rewrite IHe11; do 2 rewrite app_length; simpl; rewrite IHe12; (omega)]  with
  strat_3 e1 e2 e3 := surjective_pairing_three_arg e1 e2 e3 aux_code start alloc dealloc in goal;
     surjective_pairing_three_arg2 e1 e2 e3 code1_gen code2_gen code3_gen label in goal; split;
     hreplace IHe1 with (IHe1 alloc dealloc start aux_code label);
     hreplace IHe2 with (IHe2 alloc dealloc start (snd(fst code1_gen)) (label+length(fst (fst code1))+2));
     hreplace IHe3 with (IHe3 alloc dealloc start (snd(fst code2_gen)) (label+length(fst (fst code1))+2+length(fst (fst code2))+1));
     unfold code3, code3_gen,code2, code2_gen,code1, code1_gen in *;
     destruct IHe1 as (IHe11,IHe12); destruct IHe2 as (IHe21,IHe22); destruct IHe3 as (IHe31,IHe32);
     [ rewrite IHe11;  rewrite IHe21; rewrite <- IHe11; rewrite IHe31;(omega) |
     rewrite IHe12; rewrite IHe11; rewrite IHe22; rewrite IHe21;  rewrite <- IHe11; rewrite <- IHe12;
     rewrite app_length; rewrite app_comm_cons; simpl; rewrite app_length; do 2 rewrite split_list; do 2 rewrite app_length;
     simpl;  fold code3 code3_gen code2 code2_gen code1  code1_gen;
       fold code3 code3_gen code2 code2_gen code1  code1_gen in IHe12, IHe32;
     rewrite IHe12; rewrite IHe32; (omega)]  in 

  induction e; intros alloc dealloc start aux_code label ; simpl;
  [strat_0 | strat_0 | strat_VAR | strat_function e | strat_1 e (label+5) |strat_2 e1 e2 | 
   strat_2 e1 e2 | strat_2 e1 e2 | strat_3 e1 e2 e3 | strat_2 e1 e2 | strat_1 e label | strat_1 e label ].

  Qed.

Lemma compile_length_aux : forall e alloc dealloc start aux_code,
	snd (compile e aux_code (start+length aux_code) alloc dealloc) =
        start + length (snd(fst (compile e aux_code (start+length aux_code) alloc dealloc))).
Proof.
  intros.
  set (H:= compile_length e alloc dealloc start aux_code 0).
  destruct H.
  apply H.
Qed.

Lemma compile_length_main : forall e alloc dealloc start aux_code label,
        snd (fst(fst(compile e aux_code (start+length aux_code) alloc dealloc)) label) = 
        label + length (fst(fst(fst(fst(compile e aux_code (start+length aux_code) alloc dealloc)) label))).	
Proof.
  intros.
  set (H:= compile_length e alloc dealloc start aux_code label).
  destruct H.
  apply H0.
Qed.

Lemma auxcode_monotony : forall alloc dealloc e 
  (gen_code: ((nat->list instruction*nat*nat)*list instruction*nat)) base start,
	fragfromprog(progfromlist (snd(fst gen_code))) base <<
	fragfromprog(progfromlist (snd (fst (compile e (snd(fst gen_code)) (start + length (snd(fst gen_code))) alloc dealloc)))) base.
Proof.
  
let strat_0 := intros;simpl; reflexivity with
  strat_1 e := (simpl;intros; surjective_pairing_one_arg e (snd(fst gen_code)) start alloc dealloc in goal;
  apply (IHe gen_code base start)) with
  strat_function e :=  (intros;simpl; surjective_pairing_function e (snd(fst gen_code)) start alloc dealloc in goal;
     surjective_pairing_function e code1_gen in goal;
     hreplace IHe with (IHe gen_code base start);  fold code1_gen in IHe;
     eapply subfrag_trans; try apply IHe;  apply prefix_code) with
  strat_2 e1 e2 :=    intros;simpl;  surjective_pairing_two_arg e1 e2 (snd(fst gen_code)) start alloc dealloc in goal;
  hreplace IHe2 with (IHe2 gen_code base start);  fold code2_gen in IHe2;
  hreplace IHe1 with (IHe1 code2_gen base start);  fold code1_gen in IHe1;
  eapply subfrag_trans; [apply IHe2 | eapply subfrag_trans ; [apply IHe1 | idtac]];
  unfold code2_gen;  rewrite <- compile_length_aux;  fold code2_gen code1_gen; reflexivity with
  strat_3 e1e2 e3 :=   intros; simpl; surjective_pairing_three_arg e1 e2 e3 (snd(fst gen_code)) start alloc dealloc in goal;
   hreplace IHe1 with (IHe1 gen_code base start);  fold code1_gen in IHe1;
   hreplace IHe2 with (IHe2 code1_gen base start); unfold code1_gen in IHe2;  rewrite <- compile_length_aux in IHe2;  fold code1_gen code2_gen in IHe2;
   hreplace IHe3 with (IHe3 code2_gen base start); unfold code2_gen, code1_gen in IHe3;  rewrite compile_length_aux in IHe3;
   rewrite <- (compile_length_aux e2) in IHe3;  rewrite <- (compile_length_aux e1) in IHe3;
   fold code1_gen code2_gen code3_gen in IHe3; eapply subfrag_trans; try apply IHe3;eapply subfrag_trans; try apply IHe2;
   eapply subfrag_trans; try apply IHe1; reflexivity  in 
  induction e; [strat_0 | strat_0 | strat_0 | strat_function e | strat_1 e |
                       strat_2 e1 e2 |  strat_2 e1 e2 | strat_2 e1 e2 | 
                       strat_3 e1e2 e3 | strat_2 e1 e2 | strat_1 e | strat_1 e ].
Qed.

Lemma stack_size_positive : forall e alloc dealloc base start aux_code,
     snd (fst (((fst (fst (compile e aux_code (start+length aux_code) alloc dealloc)))) base)) >= 1.
Proof.
let strat_0 :=(omega) with
  strat_1 e base:=  surjective_pairing_one_arg e aux_code start alloc dealloc in goal;
  surjective_pairing_one_arg e code1_gen base in goal;unfold ge; assert (snd(fst code1) >= 1) ; try (unfold code1,code1_gen ; apply IHe);
  match goal with | |- 1 <= max ?a ?b => apply le_trans with (m:=a); try apply le_max_l end; (omega) with
  strat_function e := surjective_pairing_function e aux_code start alloc dealloc in goal; 
  surjective_pairing_function e code1_gen in goal; (omega) with
  strat_2 e1 e2 := surjective_pairing_two_arg e1 e2 aux_code start alloc dealloc in goal;
  surjective_pairing_two_arg e1 e2 code1_gen code2_gen base in goal; unfold ge;
  assert (snd(fst code2) >=1);[unfold code2, code2_gen ; apply IHe2 | idtac] ;
  match goal with | |- 1 <= max ?a ?b => apply le_trans with (m:=a); 
  [ match goal with | |- 1 <= max ?a' ?b' => apply le_trans with (m:=a'); 
     [(omega) |apply le_max_l] end |  apply le_max_l] end with
  strat_3 e1e2 e3 := surjective_pairing_three_arg e1 e2 e3 aux_code start alloc dealloc in goal;
  surjective_pairing_three_arg2 e1 e2 e3 code1_gen code2_gen code3_gen base in goal;
  hreplace IHe1 with (IHe1 alloc dealloc base start aux_code);  fold code1_gen code1 in IHe1; (omega)  in 

  induction e; intros alloc dealloc base start aux_code; simpl; 
  [strat_0 | strat_0 | strat_0 | strat_function e | strat_1 e (base +5)| strat_2 e1 e2 |  strat_2 e1 e2 | 
   strat_2 e1 e2 | strat_3 e1e2 e3 | strat_2 e1 e2 | strat_1 e base | strat_1 e base].
Qed.  


Lemma aux_next_in_compile : forall init alloc dealloc e base,
  fst (compile' e base init alloc dealloc) = snd (compile e nil base alloc dealloc).
Proof.
  intros.
  unfold compile'.
  rewrite (surjective_pairing (compile e nil base alloc dealloc)).
  rewrite (surjective_pairing (fst (compile e nil base alloc dealloc))).
  rewrite (surjective_pairing (fst (fst (compile e nil base alloc dealloc))
            (snd (compile e nil base alloc dealloc) + intro_code_size))).
  rewrite (surjective_pairing (fst
            (fst (fst (compile e nil base alloc dealloc))
               (snd (compile e nil base alloc dealloc) + intro_code_size)))).
  simpl.
  trivial.
Qed.  
