(* judgment.v

   Define the partial and total correctness judgments.
   
   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export tools.
Require Export state_predicates.
Require Export allocator_spec.

Definition stateRel_Pred (Pleft Pright:statePred) : stateRel.
Proof.
  intros.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun  k => 
    Pleft s p k /\ Pright s' p' k) _).
  intuition; eapply statePred_cond; try eassumption.
Defined.

Add Morphism stateRel_Pred with signature statePredEq ==> statePredEq ==> stateRelEq
as staterelpredcompat.
Proof.
  unfold statePredEq, stateRelEq, stateRel_Pred; simpl.
  firstorder.
Qed.
 
Definition statePred_Rel_left (R:stateRel) : statePred.
Proof.
  intro.
  refine (mkStatePred (fun s => fun p => fun  k => exists s', exists p', R s s' p p' k) _).
  intros. 
  destruct H as (s' & p' & H). 
  exists s'; exists p';  eapply stateRel_cond; try eassumption.
  apply subPfunction_refl.
Defined.

Add Morphism statePred_Rel_left with signature stateRelEq ==> statePredEq
as statepredrelleftcompat.
Proof.
  unfold statePredEq, stateRelEq, statePred_Rel_left; simpl.
  firstorder; exists x0; exists x1.
  rewrite <- H; assumption.
  rewrite H; assumption.
Qed.

Definition statePred_Rel_right (R:stateRel) : statePred.
Proof.
  intro.
  refine (mkStatePred (fun s' => fun p' => fun  k => exists s, exists p, R s s' p p' k) _).
  intros. 
  destruct H as (s' & p' & H). 
  exists s'; exists p';  eapply stateRel_cond; try eassumption.
  apply subPfunction_refl.
Defined.


Add Morphism statePred_Rel_right with signature stateRelEq ==> statePredEq
as statepredrelrightcompat.
Proof.
  unfold statePredEq, stateRelEq, statePred_Rel_right; simpl.
  firstorder; exists x0; exists x1.
  rewrite <- H; assumption.
  rewrite H; assumption.
Qed.

Lemma leftification : forall (R:stateRel) s s' p p' k, 
  R s s' p p' k ->  (statePred_Rel_left R) s p k.
Proof.
  unfold statePred_Rel_left; simpl; intros.
  exists s'; exists p'; assumption.
Qed.

Lemma rightification : forall (R:stateRel) s s' p p' k, 
  R s s' p p' k ->  (statePred_Rel_right R) s' p' k.
Proof.
  unfold statePred_Rel_right; simpl; intros.
  exists s; exists p; assumption.
Qed.

Definition prog_indpt_relation (R:stateRel) := forall s s' p p' k pp pp', 
  R s s' p p' k -> R s s' pp pp' k.

Add Morphism prog_indpt_relation with signature stateRelEq ==> iff
as progindptrelationcompat.
Proof.
  unfold prog_indpt_relation; simpl.
  intuition.
  rewrite <- H;  eapply H0.
  erewrite H;  eassumption.
  rewrite H;  eapply H0.
  erewrite <- H; eassumption.
Qed.

Definition splittable_relation (R:stateRel) := forall s s' p p' k, 
  (statePred_Rel_left R) s p k -> 
  (statePred_Rel_right R) s' p' k -> 
  R s s' p p' k.

Add Morphism splittable_relation with signature stateRelEq ==> iff
as splittablerelationcompat.
Proof.
  unfold splittable_relation; simpl.
  firstorder.
  rewrite <- H; apply H0.
  exists x0; exists x2.
  rewrite H;  assumption.
  exists x1; exists x3.
  rewrite H;  assumption.
  rewrite H; apply H0.
  exists x0; exists x2.
  rewrite <- H;  assumption.
  exists x1; exists x3.
  rewrite <- H;  assumption.
Qed.

(* Lemma foo : forall Rpre Rc s s' p p' k, *)
(*   (Rpre ## Rc) s s' p p' k -> *)
(*     exists f1 : (Pfunction Astate Bstate), exists f2 : (Pfunction Astate Bstate),  *)
(*     exists f'1 : (Pfunction Astate Bstate), exists f'2 : (Pfunction Astate Bstate),  *)
(*       disjointPfunction f1 f2 /\  f = (joinPfunction f1 f2) /\   *)
(*       disjointPfunction f1 f'2 /\  f' = (joinPfunction f'1 f'2) /\ *)
(*       P1 f1 f'1 p p' k /\ P2 f2 f'2 p p' k) *)

Definition preciseTensor R := forall R' s s' p p' k,
  (R##R') s s' p p' k ->     
  (
    exists ss : (Astate -> Bstate), exists ss' : (Astate -> Bstate),
    exists f1 : (Pfunction Astate Bstate), exists f2 : (Pfunction Astate Bstate),
    exists f'1 : (Pfunction Astate Bstate), exists f'2 : (Pfunction Astate Bstate),
      disjointPfunction f1 f2 /\  s = (joinPfunction f1 f2) /\
      disjointPfunction f'1 f'2 /\  s' = (joinPfunction f'1 f'2) /\
      disjointPfunction f1 f'2 /\  functionToPfunction ss = (joinPfunction f1 f'2) /\
      disjointPfunction f'1 f2 /\ functionToPfunction ss' = (joinPfunction f'1 f2) /\
      R f1 f'1 p p' k /\ R' f2 f'2 p p' k).

Lemma diagonalization_judgment : forall Rc Rpre Rpost p p' start start' finish finish', 
  let PpreL := statePred_Rel_left Rpre in
  let PpreR := statePred_Rel_right Rpre in
  let PpostL := statePred_Rel_left Rpost in
  let PpostR := statePred_Rel_right Rpost in
    preciseTensor Rpre ->
    preciseTensor (stateRel_Pred PpostL PpostL) ->
    preciseTensor (stateRel_Pred PpostR PpostR) ->
    prog_indpt_relation (stateRel_Pred PpreL PpreL) ->
    prog_indpt_relation (stateRel_Pred PpreR PpreR) ->
    splittable_relation Rpost ->
    splittable_relation Rc ->
    explicit_judgment (stateRel_Pred PpreL PpreL ## Rc) (stateRel_Pred PpostL PpostL ## Rc) p p' start start' finish finish' -> 
    explicit_judgment (stateRel_Pred PpreR PpreR ## Rc) (stateRel_Pred PpostR PpostR ## Rc) p p' start start' finish finish' -> 
    explicit_judgment (Rpre##Rc) (Rpost##Rc) p p' start start' finish finish'.
Proof.
  intros until PpostR.
  unfold PpreL, PpreR, PpostL, PpostR.
  unfold explicit_judgment.
  intros HprePrecise HpostPreciseL HpostPreciseR Hpre_progL Hpre_progR Hpost_split HRc_split H H'.
  intros k s1 s1' Hpre.
  unfold Interp in *.
  apply HprePrecise in Hpre.
  destruct Hpre as (ss & ss' & Hpre).
  destruct_tensor Hpre.
  destruct Hpre as (Hdisj2 & Hjoin2 & Hdisj2' & Hjoin2' & Hpre & HRc).
  unfold Interp in *.
  specialize (H k s1 ss). 
  destruct H as (j1 & jj1 & s2 & ss2 & Hj1 & Hjj1 & Hpost).
  unfold stateRelTensor; simpl.
  exists f1; exists f2; exists f1; exists f'2.
  intuition.
  eapply leftification with (s':= f'1); eassumption.
  eapply Hpre_progL; split.
  exists f'1; exists p'; eassumption.
  exists f'1; exists p'; eassumption.


  specialize (H' k ss' s1').
  destruct H' as (jj1' & j1' & ss2' & s2' & Hjj1' & Hj1' & Hpost').
  unfold stateRelTensor; simpl.
  exists f'1; exists f2; exists f'1; exists f'2.
  intuition.
  eapply Hpre_progR; split.
  exists f1; exists p; eassumption.
  exists f1; exists p; eassumption.
  eapply rightification with (s:= f1); eassumption.
  
  exists j1; exists j1'; exists s2; exists s2'.
  intuition.
  clear - Hpost Hpost' HpostPreciseL HpostPreciseR Hpost_split HRc_split.

  apply HpostPreciseL in Hpost.
  destruct Hpost as (sss & sss' & Hpost).
  destruct_tensor Hpost.
  destruct Hpost as (Hdisj22 & Hjoin22 & Hdisj22' & Hjoin22' & Hpost & HRc2).

  apply HpostPreciseR in Hpost'.
  destruct Hpost' as (ssss & ssss' & Hpost').
  destruct_tensor Hpost'.
  destruct Hpost' as (Hdisj222 & Hjoin222 & Hdisj222' & Hjoin222' & Hpost' & HRc2').

  exists f1; exists f2; exists f'0; exists f'3.
  intuition.
  destruct Hpost as (Hf1 & Hf1').
  destruct Hpost' as (Hf0 & Hf0').
  eapply Hpost_split; assumption.
  apply leftification in HRc2.
  apply rightification in HRc2'.
  apply HRc_split; assumption.
  
Admitted.

Lemma ptsto_splittable : forall (m m' n n': nat),
  splittable_relation ({m,m'} |-> {n,n'}).
Proof.
  unfold splittable_relation, ptsto; simpl; firstorder.
Qed.

Lemma ptsto_prog_indpt : forall (m m' n n': nat),
  prog_indpt_relation ({m,m'} |-> {n,n'}).
Proof.
  unfold prog_indpt_relation, ptsto; simpl; firstorder.
Qed.

Lemma tensor_prog_indpt : forall R R',
  prog_indpt_relation R ->
  prog_indpt_relation R' ->
  prog_indpt_relation (R ## R').
Proof.
  unfold prog_indpt_relation. 
  intros R R' HindptR HindptR' s s' p p' k pp pp' H.
  unfold stateRelTensor in H; simpl in H.
  destruct_tensor H. 
  destruct H as (H & H').
  eapply HindptR in H.
  eapply HindptR' in H'.
  unfold stateRelTensor; simpl.
  exists f1; exists f2; exists f'1; exists f'2.
  intuition; try eassumption.
Qed.

Lemma tensor_Rel_PredL : forall R R' s p k, 
  statePred_Rel_left (R ## R') s p k -> 
  statePredTensor (statePred_Rel_left R) (statePred_Rel_left R') s p k.
Proof.  
  intros until k.
  unfold statePredTensor, statePred_Rel_left, stateRelTensor; simpl.
  intro H.
  destruct H as (s' & p' & H).
  destruct_tensor H.
  destruct H as (H & H').
  exists f1; exists f2.
  intuition.
  exists f'1; exists p'; assumption.
  exists f'2; exists p'; assumption.
Qed.

Lemma tensor_Rel_PredR : forall R R' s p k, 
  statePred_Rel_right (R ## R') s p k -> 
  statePredTensor (statePred_Rel_right R) (statePred_Rel_right R') s p k.
Proof.  
  intros until k.
  unfold statePredTensor, statePred_Rel_right, stateRelTensor; simpl.
  intro H.
  destruct H as (s' & p' & H).
  destruct_tensor H.
  destruct H as (H & H').
  exists f'1; exists f'2.
  intuition.
  exists f1; exists p'; assumption.
  exists f2; exists p'; assumption.
Qed.

Lemma tensor_splittable : forall R R',
  splittable_relation R ->
  splittable_relation R' ->
  splittable_relation (R ## R').
Proof.
  unfold splittable_relation. 
  intros R R' HsplitR HsplitR' s s' p p' k HR HR'.
  apply tensor_Rel_PredL in HR.
  apply tensor_Rel_PredR in HR'.
  unfold statePredTensor in *; simpl in HR; simpl in HR'.
  destruct HR as (f1 & f2 & Hdisj & Hjoin & HRL & HR'L).
  destruct HR' as (f1' & f2' & Hdisj' & Hjoin' & HRR & HR'R).
  assert (foo := HsplitR f1 f1' p p' k HRL HRR).
  assert (foo' := HsplitR' f2 f2' p p' k HR'L HR'R).
  exists f1; exists f2; exists f1'; exists f2'.
  intuition.
Qed.

Lemma exists_Rel_PredL : forall (X:Type) (R:X->stateRel) s p k,
  statePred_Rel_left (Exists R) s p k ->
  exists x, (statePred_Rel_left (R x)) s p k.
Proof.
  intros until k.
  unfold stateExists, statePred_Rel_left; simpl.
  intro H.
  destruct H as (s' & p' & x & H).
  exists x; exists s';  exists p'; assumption.
Qed.

Lemma exists_Rel_PredR : forall (X:Type) (R:X->stateRel) s p k,
  statePred_Rel_right (Exists R) s p k ->
  exists x, (statePred_Rel_right (R x)) s p k.
Proof.
  intros until k.
  unfold stateExists, statePred_Rel_right; simpl.
  intro H.
  destruct H as (s' & p' & x & H).
  exists x; exists s';  exists p'; assumption.
Qed.

Lemma exists_splittable : forall X (R : X -> stateRel),
  (forall x, splittable_relation (R x)) ->
  splittable_relation (Ex x, R x).
Proof.
  unfold splittable_relation; intros until k.
  unfold stateExists; simpl.
  intros H0 H1.
  destruct H0 as (ss' & pp' & a0 & H0). 
  destruct H1 as (ss & pp & a1 & H1).
  exists a0.
  apply (H a0).
  exists ss'; exists pp'; intuition.
  exists ss; exists pp; intuition.
(*   firstorder. *)
(*   exists x4. *)
  
(*   exists x3; exists x10; exists x7; exists x12. *)
(*   intuition. *)
(* Qed. *)
Admitted.

Lemma exists_splittable_spe : forall m m' n n', (* (R : nat -> nat -> stateRel), *)
  (* (forall x, forall y, splittable_relation (R x y )) -> *)
  splittable_relation (Ex x, Ex y, ({m,m'}|->{x,y} ## Block x y n n')).
Proof.
  double induction n n'.
  intros; unfold splittable_relation; intros until k.
  Opaque stateRelTensor.
  unfold stateExists; simpl.
  intros H0 H1.
  destruct H0 as (ss' & pp' & a0 & aa1 & H0). 
  destruct H1 as (ss & pp & aa0 & a1 & H1). 
  exists a0; exists a1.
  destruct_tensor H0; destruct_tensor H1.
  exists f1; exists f2; exists f'0; exists f'3.
  intuition.
  destruct H as (l1 & l2 & H).
  destruct H0 as (l1' & l2' & H0).
  apply ptsto_splittable.
  exists f'0; exists p';exists l1; exists l2'.
  simpl in *; intuition.
  exists f1; exists p;exists l1; exists l2'.
  simpl in *; intuition.

  intros; unfold splittable_relation; intros until k.
  simpl Block in *.
  Opaque stateRelTensor.
  unfold stateExists; simpl.
  (* rewrite right_block_unfold_end. *)
  
  intros H0 H1.
  destruct H0 as (ss' & pp' & a0 & aa1 & H0). 
  destruct H1 as (ss & pp & aa0 & a1 & H1). 
  exists a0; exists a1.
  destruct_tensor H0; destruct_tensor H1.
  exists f1; exists f2; exists f'0; exists f'3.
  intuition.
  apply (ptsto_splittable m m' a0 a1).
  apply leftification in H2.
  
  assumption.
  destruct H as (l1 & l2 & H).
  destruct H0 as (l1' & l2' & H0).
  apply ptsto_splittable.
  exists f'0; exists p';exists l1; exists l2'.
  simpl in *; intuition.
  exists f1; exists p;exists l1; exists l2'.
  simpl in *; intuition.

  apply Hsplit.
  
  exists f'3; exists p'.
  
Qed.

Lemma ptstoTrue_splittable : forall (m m': nat),
  splittable_relation ({m,m'} |-> -).
Proof.
  unfold splittable_relation, ptsto; simpl; firstorder.
Qed.

Lemma spittable_rpost : forall Ra Rc_alloc size size' n5 n5' n6 n6' l l',  
   splittable_relation (SPost_alloc Ra Rc_alloc size size' n5 n5' n6 n6') /\
   splittable_relation (SPre_alloc Ra Rc_alloc size size' n5 n5' n6 n6' l l').
Proof.
  intros; unfold SPost_alloc, SPre_alloc;split.
  apply tensor_splittable.
  Focus 2.
  apply tensor_splittable.
  Focus 2.
  apply tensor_splittable.
  Focus 2.
  repeat (apply tensor_splittable; try apply ptstoTrue_splittable 
    ; try apply ptsto_splittable).
  apply exists_splittable.


;
(**)