(* evalution.v

   Lemmas relating the relational description of the stack to its 
   explicit description

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export compiler.
Require Export tools.

Lemma ptsto_pstateval : forall m m' n n' P s s' p p' k,  
  (({m,m'} |-> {n,n'}) ## P) s s' p p' k -> 
  ((s m) = Some n /\ (s' m') = Some n').
Proof.
  intros.
  assert (Hw:= stateRelTensor_weakening ({m,m'} |-> {n,n'}) P).
  unfold stateRelLeq in Hw.
  hreplace Hw with ((Hw s s' p p' k) H).
  simpl in Hw.
  firstorder; subst;  assumption.
Qed.

Lemma ptsto_stateval : forall m m' n n' P s s' p p' k, 
  Interp (({m,m'} |-> {n,n'}) ## P) s s' p p' k -> 
  (s m = n /\ s' m' = n').
Proof.
  intros.
  assert (Hps:= ptsto_pstateval _ _ _ _ _ _ _ _ _ _ H).
  destruct Hps as (Hps,Hps'); inversion Hps; inversion Hps'.
  rewrite H1,H2; intuition.
Qed.

Lemma ptsto_stateval_ind : forall a a' b b' n n' P s s' p p' k, 
  Interp (({a,a'} |->{b,b'}) ## ({b,b'} |-> {n,n'}) ## P) s s' p p' k -> 
  ( s (s a) = n /\ s' (s' a') = n').
Proof.
  intros.
  assert (s a = b /\ s' a' =b').
  eapply ptsto_stateval; apply H.
  destruct H0 as (H0,H0'); rewrite H0,H0'.
  setoid_replace  (({a,a'} |->{b,b'}) ## ({b,b'} |-> {n,n'}) ## P) with (({b,b'} |-> {n,n'}) ## ({a,a'} |->{b,b'}) ## P)
  using relation stateRelEq in H; try ring.
  eapply ptsto_stateval; apply H.
Qed.

Lemma ptsto_statevalOp : forall (op:nat->nat->nat) m1 m1' n1 n1' m2 m2' n2 n2' P s s' p p' k, 
  Interp (({m1,m1'}|->{n1,n1'}) ## ({m2,m2'}|->{n2,n2'}) ## P) s s' p p' k -> 
  (op (s m1) (s m2) = op n1 n2 /\ op (s' m1') (s' m2') = op n1' n2').
Proof.
  intros.
  assert (Hps1:= ptsto_pstateval _ _ _ _ _ _ _ _ _ _ H).
  setoid_replace (({m1,m1'}|->{n1,n1'}) ## ({m2,m2'}|->{n2,n2'}) ## P) with 
    (({m2,m2'}|->{n2,n2'}) ## ({m1,m1'}|->{n1,n1'}) ## P) 
    using relation stateRelEq in H;  try ring.
  assert (Hps2:= ptsto_pstateval _ _ _ _ _ _ _ _ _ _ H).
  destruct Hps1 as (Hps1,Hps1'); inversion Hps1; inversion Hps1'.
  destruct Hps2 as (Hps2,Hps2'); inversion Hps2; inversion Hps2'.
  rewrite H1,H2,H3,H4; intuition.
Qed.

Lemma ptsto_statevalOpLeft : forall (op:nat->nat->nat) m1 m1' n1 n1' n2 n2' P s s' p p' k, 
  Interp (({m1,m1'}|->{n1,n1'}) ## P) s s' p p' k -> 
  (op (s m1) n2 = op n1 n2 /\ op (s' m1') n2' = op n1' n2').
Proof.
  intros until k; intro Hinterp.
  assert (Hs : s m1 = n1 /\ s' m1' =n1').
  eapply ptsto_stateval; apply Hinterp.
  intuition.
Qed.


Lemma stateRellift_val_interp : forall  P Q s s' p p' k, 
  Interp ((stateRel_lift Q) ## P) s s' p p' k -> Q.
Proof.
  intros; unfold stateRel_lift in H;  simpl in H; firstorder.
Qed.

Lemma stateRellift_equality : forall  a m n, 
  stateRelEq (stateRel_lift (m=n) ## (m|->a)) (stateRel_lift (m=n) ## (n|->a)).
Proof.
  unfold stateRelEq; intuition;
  (assert (m=n); [eapply stateRellift_val | idtac]);subst; try eassumption.
Qed.

(*--------------------*)
(* Independent update *)
(*--------------------*)

Definition psupdate (ps : Pstate) (l: nat) (v:nat) := (fun w : nat => if beq_nat w l then
                                                       match (ps l) with 
							 |  (Some y) => Some v
							 |   None => None
							end 
                                                        else ps w).

Lemma psupdate_dom : forall ps l v n, 
  domPfunction ps n <-> domPfunction (psupdate ps l v) n.
Proof.
  unfold domPfunction, psupdate; intros; split.
  intros.
  witness n0 in H.
  simpl.
  compare n l.
  intro Heq; rewrite Heq.
  rewrite <- (beq_nat_refl l).
  rewrite <- Heq; rewrite H.
  exists v; reflexivity.
  intro Hneq; replace (beq_nat n l) with false.
  exists n0; assumption.
  apply (beq_nat_neq).
  assumption.
  
  simpl.
  intros; witness n0 in H. 
  compare n l.
  intro; replace (beq_nat n l) with true in H.
  rewrite e.
  dcase (ps l).
  intros.
  exists n1; reflexivity.
  intro; 
  rewrite H0 in H.
  exists n0.
  assumption.
  rewrite e; apply beq_nat_refl.
  intro.
  replace (beq_nat n l) with false in H.
  exists n0; assumption.
  apply beq_nat_neq.
  assumption.
Qed.
  
Lemma psupdateupdate : forall s l v, 
  functionToPfunction (update s l v) = psupdate (functionToPfunction s) l v.
Proof. 
  intros.
  unfold functionToPfunction, psupdate, update; simpl.
  rewrite <- (Extensionality nat (option nat)).
  intro l'.
  compare l l'.
  intro H; rewrite H.
  rewrite <- (beq_nat_refl l').
  reflexivity.
  intros H.
  replace (beq_nat l' l) with false.
  replace (beq_nat l l') with false.
  trivial.
  apply beq_nat_neq.
  assumption.
  apply beq_nat_neq.
  intuition.
Qed.


Lemma psupdatejoin : forall ps1 ps2 l v, domPfunction ps1 l -> 
  psupdate (joinPfunction ps1 ps2) l v = joinPfunction (psupdate ps1 l v) ps2.
Proof.
  unfold psupdate, joinPfunction; simpl; intros.
  rewrite <- Extensionality.
  simpl.
  intro.
  compare a l.
  intro Heq; rewrite Heq.
  rewrite <- (beq_nat_refl l).
  witness v2 in H.
  rewrite H.
  reflexivity.
  intro Hneq; replace (beq_nat a l) with false.
  reflexivity.
  apply beq_nat_neq.
  assumption.
Qed.

Lemma indepupdate : forall s s' p p' k P Q v v' loc loc', natrel Q p p' v v' k ->
  Interp (({loc,loc'} |=> natPtrue) ##P) s s' p p' k -> 
  Interp (({loc,loc'} |=> Q)## P) (update s loc v) (update s' loc' v') p p' k.
Proof.
  simpl; intros; destruct_tensor H0 ;destruct H0 as ([l [l' (Hl,(Hl',Htrue))]],HP).
  exists (psupdate f1 loc v); exists f2; exists (psupdate f'1 loc' v'); exists f'2.
  intuition.

  unfold disjointPfunction.
  intro l0;  setoid_rewrite <- (psupdate_dom f1 loc v l0).
  apply Hdisj.

  transitivity (psupdate (functionToPfunction s) loc v).
  apply (psupdateupdate s loc v).
  rewrite <- psupdatejoin.
  rewrite Hjoin; reflexivity.
  unfold domPfunction; exists l; assumption.

  unfold disjointPfunction.
  intro l0;  setoid_rewrite <- (psupdate_dom f'1 loc' v' l0).
  apply Hdisj'.

  transitivity (psupdate (functionToPfunction s') loc' v').
  apply (psupdateupdate s' loc' v').
  rewrite <- psupdatejoin.
  rewrite Hjoin'; reflexivity.
  unfold domPfunction; exists l'; assumption.

  exists v; exists v'.
  intuition; unfold psupdate; rewrite <- beq_nat_refl; try rewrite Hl; try rewrite Hl'; trivial.
Qed.


Lemma ptstointerp : forall n n' P Q s s' p p' k, 
  Interp (({n,n'}|=>P) ## Q) s s' p p' k  ->
  P p p' (s n) (s' n') k.
Proof.
  unfold Interp, stateRelTensor; simpl; intros.
  destruct_tensor H ;destruct H as ([l[l' (H , H')]],Hq).
  intuition.
  rewrite (subelement n H Hjoin); rewrite (subelement n' H0 Hjoin').
  assumption.
Qed.

Lemma ptstointerp_cloud : forall n n' P Q R s s' p p' k , 
  Interp ((cloudToState (({n,n'}||=>P)//\\Q)) ## R) s s' p p' k  -> 
  P p p' (s n) (s' n') k.
Proof.
  unfold Interp, stateRelTensor; simpl; intros.
  destruct_tensor H ;destruct H as (([l[l' Hp]],Hq),Hr).
  intuition.
  rewrite (subelement n H Hjoin); rewrite (subelement n' H1 Hjoin').
  assumption.
Qed.

Lemma ptstointerp_stack : forall list ptr ptr' n n' P Q R s s' p p' k , 
    Interp ((stack_description list ptr ptr' (({n,n'}||=>P)//\\Q)) ## R) s s' p p' k  ->
    P p p' (s n) (s' n') k.
Proof.
  intros; explicit_stack list ptr ptr'.
  rewrite (Hstack ({n,n'}||=>P //\\ Q)) in H.
  normal_form in H.
  rewrite cloudRelConj_comm in H; rewrite <- cloudRelConj_assoc in H.
  eapply ptstointerp_cloud; try apply H.
Qed.

Lemma ptstointerp_ind : forall n n' m m' P Q s s' p p' k, 
  Interp (({n,n'}|->{m,m'}) ## ({m,m'}|=>P) ## Q) s s' p p' k  ->
  P p p' (s(s n)) (s'(s' n')) k.
Proof.
  unfold Interp, stateRelTensor; simpl; intros;
  destruct_tensor H ;destruct H as ([l [l' (Hl,(Hl',(Hm,H'm)))]],foo);
  destruct_tensor foo ;destruct foo as ([l2 [l'2 (Hl2,(Hl2',Hp))]],Hq);
  intuition;subst.
  assert (Hn : s n = m).
  apply (subelement n Hl Hjoin).
  rewrite Hn; rewrite join_disj_comm in Hjoin; [idtac | assumption].
  rewrite  disj_join_right in Hdisj; inversion Hdisj.
  rewrite <- disj_join_assoc in Hjoin; resolve_frag.
  rewrite (subelement m Hl2 Hjoin).
  assert (Hn' : s' n' = m').
  apply (subelement n' Hl' Hjoin').
  rewrite Hn'; rewrite join_disj_comm in Hjoin'; [idtac | assumption].
  rewrite (disj_join_right) in Hdisj'; inversion Hdisj'.
  rewrite <- disj_join_assoc in Hjoin'; resolve_frag.
  rewrite (subelement m' Hl2' Hjoin'); assumption.
Qed.

Lemma ptstointerp_ind_cloud : forall n n' m m' P Q R s s' p p' k , 
  Interp ({n,n'}|->{m,m'} ## cloudToState (({m,m'}||=>P)//\\Q) ## R) s s' p p' k  ->
  P p p' (s (s n)) (s' (s' n'))  k.
Proof.
  unfold Interp, stateRelTensor; simpl; intros.
  destruct_tensor H ;destruct H as ([l [l' (Hl,(Hl',(Hm,H'm)))]],foo);
  destruct_tensor foo ;destruct foo as (([l2 [l'2 (Hl2,(Hl2',Hp))]],Hq),Hr);
  intuition; subst.
  assert (Hn : s n = m).
  apply (subelement n Hl Hjoin).
  rewrite Hn; rewrite join_disj_comm in Hjoin; [idtac | assumption].
  rewrite  disj_join_right in Hdisj; inversion Hdisj.
  rewrite <- disj_join_assoc in Hjoin; resolve_frag.
  rewrite (subelement m Hl2 Hjoin).
  assert (Hn' : s' n' = m').
  apply (subelement n' Hl' Hjoin').
  rewrite Hn'; rewrite join_disj_comm in Hjoin'; [idtac | assumption].
  rewrite  disj_join_right in Hdisj'; inversion Hdisj'.
  rewrite <- disj_join_assoc in Hjoin'; resolve_frag.
  rewrite (subelement m' Hl2' Hjoin'); assumption.
Qed.

Lemma ptstointerp_ind_stack : forall list ptr ptr' n n' m m' P Q R s s' p p' k , 
  Interp (({n,n'}|->{m,m'}) ## stack_description list ptr ptr' (({m,m'}||=>P)//\\Q) ## R) s s' p p' k  ->
  P p p' (s (s n)) (s' (s' n')) k.
Proof.
  intros; explicit_stack list ptr ptr' .
  rewrite (Hstack ({m,m'}||=>P //\\ Q)) in H.
  normal_form in H.
  rewrite cloudRelConj_comm in H; rewrite <- cloudRelConj_assoc in H.
  eapply ptstointerp_ind_cloud; try apply H.
Qed.

 
Lemma ptstointerp_indo : forall cst cst' n n' m m' P Q s s' p p' k, 
  Interp (({n,n'}|->{m,m'}) ## ({m+cst,m'+cst'}|=>P) ## Q) s s' p p' k  ->
  P p p' (s (s n + cst)) (s' (s' n' + cst')) k.
Proof.
  unfold Interp, stateRelTensor; simpl; intros;
  destruct_tensor H ;destruct H as ([l [l' (Hl,(Hl',(Hm,H'm)))]],foo);
  destruct_tensor foo ;destruct foo as ([l2 [l'2 (Hl2,(Hl2',Hp))]],Hq);
  intuition; subst.
  assert (Hn : s n = m).
  apply (subelement n Hl Hjoin).
  rewrite Hn; rewrite join_disj_comm in Hjoin; [idtac | assumption].
  rewrite  disj_join_right in Hdisj; inversion Hdisj.
  rewrite <- disj_join_assoc in Hjoin; resolve_frag.
  rewrite (subelement (m+cst) Hl2 Hjoin).
  assert (Hn' : s' n' = m').
  apply (subelement n' Hl' Hjoin').
  rewrite Hn'; rewrite join_disj_comm in Hjoin'; [idtac | assumption].
  rewrite  disj_join_right in Hdisj'; inversion Hdisj'.
  rewrite <- disj_join_assoc in Hjoin'; resolve_frag.
  rewrite (subelement (m'+cst') Hl2' Hjoin'); assumption.
Qed.

