(* focusin_efficient.v

   Define a tactic that place the nth component of a tensor (or a conj) 
   at the top of the stack
   To be efficient, we strongly make use of reflection.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export judgment.

Require Export theory_of_types.

(**********  reflection for stateRel  ***********)

Inductive stack : stateRel -> Type :=
  | Unit : stack statePtrue
  | Atom : forall R, stack R 
  | Tensor : forall R R', stack R -> stack R' -> stack (R ## R').

Ltac certificate R k :=
  match R with
    | ?R1 ## ?R2 =>
      certificate R2 ltac:(fun H2 =>
      certificate R1 ltac:(fun H1 =>
        k (Tensor R1 R2 H1 H2)))
    | statePtrue => k Unit
    | ?R1 => k (Atom R1)
  end.

Fixpoint stackToList_rec R (H:stack R) (accu:list (sigT stack)): list (sigT stack) :=
  match H with
    | Unit => accu
    | Atom R => ((existT stack R (Atom R))::accu)
    | Tensor R1 R2 H1' H2' =>
      stackToList_rec R2 H2' (stackToList_rec R1 H1' accu)
  end.

Definition stackToList R (H:stack R) : list (sigT stack) := List.rev (stackToList_rec R H nil).

Fixpoint listToStack (l:list (sigT stack)) : sigT stack :=
  match l with
    | nil => existT stack statePtrue Unit
    | h :: nil => h
    | h :: t => let recStack := listToStack t in 
      let R := projT1 recStack in
      let H := projT2 recStack in
          existT stack (projT1 h ## R) (Tensor (projT1 h) R (projT2 h) H)
  end.

Fixpoint sigStackToList (S:sigT stack) : list (sigT stack) :=
  stackToList (projT1 S) (projT2 S).

Definition normalize S := projT1 (listToStack (sigStackToList S)).

Lemma concatList : forall l l', stateRelEq (projT1 (listToStack (l++l')))
  ((projT1 (listToStack l)) ## (projT1 (listToStack l'))).
Proof.
  induction l; simpl; intros; try ring.
  destruct l; simpl.  
  destruct l'; simpl; try ring.
  rewrite (IHl l'). 
  destruct l; simpl; try ring.
Qed.

Lemma concatStackRec : forall R H l, rev (stackToList_rec R H l) =
   rev l ++ rev (stackToList_rec R H nil).
Proof.
  induction H; intros; simpl; trivial.
  apply app_nil_end.
  setoid_rewrite IHstack2.
  setoid_rewrite IHstack1.
  rewrite app_ass; simpl; trivial.
Qed.

Lemma concatStack : forall R R' (H:stack R) (H':stack R'),  stackToList (R##R') (Tensor R R' H H') =
  stackToList R H ++ stackToList R' H'.
Proof.
  unfold stackToList;simpl ;intros.
  setoid_rewrite concatStackRec; simpl; trivial.
Qed.

Lemma normalize_compat : forall R (H:stack R), stateRelEq R (normalize (existT stack R H)).
Proof.
  unfold normalize; induction H; simpl; try ring.
  setoid_rewrite concatStack.
  setoid_rewrite concatList.
  setoid_rewrite <- IHstack1; setoid_rewrite <- IHstack2. 
  reflexivity.
Qed.

Fixpoint placeOnTop_rec R R' (H:stack R) (H':stack R') n :=
  match n with
    | 0 =>
      match H with
        | Unit => existT stack R' H'
        | Atom R => existT stack (R ## R') (Tensor R R' (Atom R) H')
        | Tensor R1 R2 H1 H2 => existT stack (R1  ## (R' ## R2)) 
                        (Tensor R1 (R' ## R2) H1 (Tensor R' R2 H' H2))
      end
    | S m =>
      match H with
        | Unit => existT stack R' H'
        | Atom R => existT stack (R ## R') (Tensor R R' (Atom R) H')
        | Tensor R1 R2 H1 H2 =>
          placeOnTop_rec R2 (R' ## R1) H2 (Tensor R' R1 H' H1) m
      end
  end.

Definition placeOnTop R (H:stack R) n := placeOnTop_rec R statePtrue H Unit n.

Lemma focusing_rec : forall n R R' (H:stack R) (H':stack R'),  
  stateRelEq (R##R') (normalize (placeOnTop_rec R R' H H' n)).
Proof.
  induction n; 
  induction H; simpl; intros; try (rewrite <- normalize_compat; ring).
  specialize (IHn R'0 (R' ## R) H0 (Tensor R' R H' H)).
  rewrite <- IHn.
  ring.
Qed.

Lemma focusing : forall n R (H:stack R), stateRelEq R (normalize (placeOnTop R H n)).
Proof.
  intros.
  setoid_rewrite <- (stateRelTensor_unit R) at 1.
  setoid_rewrite <- (stateRelTensor_comm R) at 1.
  apply (focusing_rec n R statePtrue H Unit).
Qed.

(**********  reflection for cloudRel  ***********)

Inductive stackCloud : cloudRel -> Type :=
  | Unit_c : stackCloud cloudPtrue
  | Atom_c : forall R, stackCloud R 
  | Conj : forall R R', stackCloud R -> stackCloud R' -> stackCloud (R //\\ R').

Ltac certificate_c R k :=
  match R with
    | ?R1 //\\ ?R2 =>
      certificate_c R2 ltac:(fun H2 =>
      certificate_c R1 ltac:(fun H1 =>
        k (Conj R1 R2 H1 H2)))
    | cloudPtrue => k Unit_c
    | ?R1 => k (Atom_c R1)
  end.

Fixpoint stackCloudToList R (H:stackCloud R) : list (sigT stackCloud) :=
  match H with
    | Unit_c => nil
    | Atom_c R => [existT stackCloud R (Atom_c R)]
    | Conj R1 R2 H1' H2' =>
      (stackCloudToList R1 H1')++(stackCloudToList R2 H2')
  end.

Fixpoint listToStackCloud (l:list (sigT stackCloud)) : sigT stackCloud :=
  match l with
    | nil => existT stackCloud cloudPtrue Unit_c
    | h :: nil => h
    | h :: t => let recStackCloud := listToStackCloud t in 
      let R := projT1 recStackCloud in
      let H := projT2 recStackCloud in
          existT stackCloud (projT1 h //\\ R) (Conj (projT1 h) R (projT2 h) H)
  end.

Fixpoint sigStackCloudToList (S:sigT stackCloud) : list (sigT stackCloud) :=
  stackCloudToList (projT1 S) (projT2 S).

Definition normalize_c S := projT1 (listToStackCloud (sigStackCloudToList S)).

Lemma concatStackCloud : forall l l', cloudRelEq (projT1 (listToStackCloud (l++l')))
  ((projT1 (listToStackCloud l)) //\\ (projT1 (listToStackCloud l'))).
Proof.
  induction l; simpl; intros; try ring.
  destruct l; simpl.  
  destruct l'; simpl; try ring.
  rewrite (IHl l'). 
  destruct l; simpl; try ring.
Qed.

Lemma normalize_compat_c : forall R (H:stackCloud R), 
  cloudRelEq R (normalize_c (existT stackCloud R H)).
Proof.
  unfold normalize_c; induction H; simpl; try ring.
  setoid_rewrite concatStackCloud.
  setoid_rewrite <- IHstackCloud1; setoid_rewrite <- IHstackCloud2. 
  reflexivity.
Qed.

Fixpoint placeOnTop_c_rec R R' (H:stackCloud R) (H':stackCloud R') n :=
  match n with
    | 0 =>
      match H with
        | Unit_c => existT stackCloud R' H'
        | Atom_c R => existT stackCloud (R //\\ R') (Conj R R' (Atom_c R) H')
        | Conj R1 R2 H1 H2 => existT stackCloud (R1  //\\ (R' //\\ R2)) 
                        (Conj R1 (R' //\\ R2) H1 (Conj R' R2 H' H2))
      end
    | S m =>
      match H with
        | Unit_c => existT stackCloud R' H'
        | Atom_c R => existT stackCloud (R //\\ R') (Conj R R' (Atom_c R) H')
        | Conj R1 R2 H1 H2 =>
          placeOnTop_c_rec R2 (R' //\\ R1) H2 (Conj R' R1 H' H1) m
      end
  end.

Definition placeOnTop_c R (H:stackCloud R) n := placeOnTop_c_rec R cloudPtrue H Unit_c n.

Lemma focusing_c_rec : forall n R R' (H:stackCloud R) (H':stackCloud R'),  
  cloudRelEq (R//\\R') (normalize_c (placeOnTop_c_rec R R' H H' n)).
Proof.
  induction n; 
  induction H; simpl; intros; try (rewrite <- normalize_compat_c; ring).
  specialize (IHn R'0 (R' //\\ R) H0 (Conj R' R H' H)).
  rewrite <- IHn.
  ring.
Qed.

Lemma focusing_c : forall n R (H:stackCloud R),
  cloudRelEq R (normalize_c (placeOnTop_c R H n)).
Proof.
  intros.
  setoid_rewrite <- (cloudRelConj_unit R) at 1.
  setoid_rewrite <- (cloudRelConj_comm R) at 1.
  apply (focusing_c_rec n R cloudPtrue H Unit_c).
Qed.



(********* some context lemmas to use apply instead of rewrite **********)
(********* disgusting but a lot faster **********)

(************* lemmas for stateRel **********)

Lemma focusing_lemma : forall n R (H:stack R) R',
  stateRelEq (normalize (placeOnTop R H n)) R' -> stateRelEq R R'.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

Lemma focusing_lemma_leq : forall n R (H:stack R) R',
  stateRelLeq (normalize (placeOnTop R H n)) R' -> stateRelLeq R R'.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

Lemma focusing_lemma_derivation : forall n R (H:stack R) p p' l l' k,
  derivation (normalize (placeOnTop R H n)) p p' l l' k -> derivation R p p' l l' k.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

Lemma focusing_lemma_judgment : forall n R (H:stack R) p p' l l',
  judgment (normalize (placeOnTop R H n)) p p' l l' -> judgment R p p' l l'.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

Lemma focusing_lemma_totaljudgment : forall n R (H:stack R) R' p p' l1 l1' l2 l2',
  explicit_judgment (normalize (placeOnTop R H n)) R' p p' l1 l1' l2 l2' -> 
  explicit_judgment R R' p p' l1 l1' l2 l2'.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

Lemma focusing_lemma_interp : forall n R (H:stack R) s s' p p' k,
  Interp (normalize (placeOnTop R H n)) s s' p p' k -> Interp R s s' p p' k.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

(************* lemmas for stateRel as hyp **********)

Lemma focusing_lemma_hyp : forall P n R (H:stack R) R',
  stateRelEq R R' -> (stateRelEq (normalize (placeOnTop R H n)) R' -> P)
  -> P.
Proof.
  intros until R'; intros H1 H2.
  rewrite (focusing n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_lemma_leq_hyp : forall P n R (H:stack R) R',
  stateRelLeq R R' -> (stateRelLeq (normalize (placeOnTop R H n)) R' -> P)
  -> P.
Proof.
  intros until R'; intros H1 H2.
  rewrite (focusing n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_lemma_judgment_hyp : forall P n R (H:stack R) p p' l l',
  judgment R p p' l l' -> 
  (judgment (normalize (placeOnTop R H n)) p p' l l' -> P)
  -> P.
Proof.
  intros until l'; intros H1 H2.
  rewrite (focusing n R H) in H1.  
  apply (H2 H1). 
Qed.

Lemma focusing_lemma_totaljudgment_hyp : forall P n R (H:stack R) R' p p' l1 l1' l2 l2',
  explicit_judgment R R' p p' l1 l1' l2 l2' ->
  (explicit_judgment (normalize (placeOnTop R H n)) R' p p' l1 l1' l2 l2' -> P)
  -> P.
Proof.
  intros until l2'; intros H1 H2.
  rewrite (focusing n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_lemma_interp_hyp : forall P n R (H:stack R) s s' p p' k,
  Interp R s s' p p' k -> 
  (Interp (normalize (placeOnTop R H n)) s s' p p' k -> P)
  -> P.
Proof.
  intros until k; intros H1 H2.
  rewrite (focusing n R H) in H1.  
  apply (H2 H1).
Qed.

(************* lemmas for cloudRel **********)

Lemma focusing_c_lemma : forall n R (H:stackCloud R) R',
  stateRelEq (cloudToState(normalize_c (placeOnTop_c R H n))) R' -> 
  stateRelEq (cloudToState R) R'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_lemma_leq : forall n R (H:stackCloud R) R',
  stateRelLeq (cloudToState(normalize_c (placeOnTop_c R H n))) R' -> 
  stateRelLeq (cloudToState R) R'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_lemma_cloud : forall n R (H:stackCloud R) R',
  cloudRelEq (normalize_c (placeOnTop_c R H n)) R' -> cloudRelEq R R'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_lemma_cloudleq : forall n R (H:stackCloud R) R',
  cloudRelLeq (normalize_c (placeOnTop_c R H n)) R' -> cloudRelLeq R R'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_lemma_judgment : forall n R (H:stackCloud R) p p' l l',
  judgment (cloudToState (normalize_c (placeOnTop_c R H n))) p p' l l' -> 
  judgment (cloudToState R) p p' l l'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_lemma_totaljudgment : forall n R (H:stackCloud R) R' p p' l1 l1' l2 l2',
  explicit_judgment (cloudToState (normalize_c (placeOnTop_c R H n))) R' p p' l1 l1' l2 l2' -> 
  explicit_judgment (cloudToState R) R' p p' l1 l1' l2 l2'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

(************* lemmas for cloudRel as hyp **********)

Lemma focusing_c_lemma_hyp : forall P n R (H:stackCloud R) R',
  stateRelEq (cloudToState R) R' ->
  (stateRelEq (cloudToState (normalize_c (placeOnTop_c R H n))) R' -> P)
  -> P.
Proof.
  intros until R'; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_lemma_leq_hyp : forall P n R (H:stackCloud R) R',
  stateRelLeq (cloudToState R) R' -> 
  (stateRelLeq (cloudToState (normalize_c (placeOnTop_c R H n))) R' -> P)
  -> P.
Proof.
  intros until R'; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_lemma_cloud_hyp : forall P n R (H:stackCloud R) R',
  cloudRelEq R R' -> 
  (cloudRelEq (normalize_c (placeOnTop_c R H n)) R' -> P)
  -> P.
Proof.
  intros until R'; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_lemma_cloudleq_hyp : forall P n R (H:stackCloud R) R',
  cloudRelLeq R R' -> 
  (cloudRelLeq (normalize_c (placeOnTop_c R H n)) R' -> P)
  -> P.
Proof.
  intros until R'; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_lemma_judgment_hyp : forall P n R (H:stackCloud R) Q p p' l l',
  judgment (cloudToState R ## Q) p p' l l' -> 
  (judgment (cloudToState (normalize_c (placeOnTop_c R H n)) ## Q) p p' l l' -> P)
  -> P.
Proof.
  intros until l'; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_lemma_totaljudgment_hyp : forall P n R (H:stackCloud R) Q R' p p' l1 l1' l2 l2',
  explicit_judgment (cloudToState R ## Q) R' p p' l1 l1' l2 l2' ->
  (explicit_judgment (cloudToState (normalize_c (placeOnTop_c R H n)) ## Q) R' p p' l1 l1' l2 l2' -> P) -> P.
Proof.
  intros until l2'; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

(************* lemmas for cloudRel with context **********)

Lemma focusing_c_context_lemma_judgment : forall n R (H:stackCloud R) Q p p' l l',
  judgment (cloudToState (normalize_c (placeOnTop_c R H n)) ## Q) p p' l l' -> 
  judgment (cloudToState R ## Q) p p' l l'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_context_lemma_derivation : forall n R (H:stackCloud R) Q p p' l l' k,
  derivation (cloudToState (normalize_c (placeOnTop_c R H n)) ## Q) p p' l l' k -> 
  derivation (cloudToState R ## Q) p p' l l' k.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_context_lemma_totaljudgment : forall n R (H:stackCloud R) R' Q p p' l1 l1' l2 l2' ,
  explicit_judgment (cloudToState (normalize_c (placeOnTop_c R H n)) ## Q) R' p p' l1 l1' l2 l2' -> 
  explicit_judgment (cloudToState R ## Q) R' p p' l1 l1' l2 l2'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_stack_lemma : forall n R (H:stackCloud R) R' ptr ptr' list,
  stateRelEq (stack_description list ptr ptr'
    (normalize_c (placeOnTop_c R H n))) R' -> 
  stateRelEq (stack_description list ptr ptr' R) R'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_stack_lemma_leq : forall n R (H:stackCloud R) R' ptr ptr' list,
  stateRelLeq (stack_description list ptr ptr'
    (normalize_c (placeOnTop_c R H n))) R' -> 
  stateRelLeq (stack_description list ptr ptr' R) R'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_stack_lemma_judgment : forall n R (H:stackCloud R) Q p p' l l' ptr ptr' list,
  judgment (stack_description list ptr ptr' (normalize_c (placeOnTop_c R H n)) ## Q) p p' l l' -> 
  judgment (stack_description list ptr ptr' R ## Q) p p' l l'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.

Lemma focusing_c_stack_lemma_totaljudgment : forall n R (H:stackCloud R) Q R' p p' l1 l1' l2 l2' ptr ptr' list,
  explicit_judgment (stack_description list ptr ptr' (normalize_c (placeOnTop_c R H n)) ## Q) R' p p' l1 l1' l2 l2' -> 
  explicit_judgment (stack_description list ptr ptr' R ## Q) R' p p' l1 l1' l2 l2'.
Proof.
  intros.
  rewrite (focusing_c n R H);  assumption.
Qed.


(************* lemmas for cloudRel with context as hyp **********)

Lemma focusing_c_context_lemma_judgment_hyp : forall P n R (H:stackCloud R) Q p p' l l',
  judgment (cloudToState R ## Q) p p' l l' ->
  (judgment (cloudToState (normalize_c (placeOnTop_c R H n)) ## Q) p p' l l' -> P)
  -> P.
Proof.
  intros until l'; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_context_lemma_totaljudgment_hyp : forall P n R (H:stackCloud R) R' Q p p' l1 l1' l2 l2' ,
  explicit_judgment (cloudToState R ## Q) R' p p' l1 l1' l2 l2' ->
  (explicit_judgment (cloudToState (normalize_c (placeOnTop_c R H n)) ## Q) R' p p' l1 l1' l2 l2' -> P) -> P.    
Proof.
  intros until l2'; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_stack_lemma_hyp : forall P n R (H:stackCloud R) R' ptr ptr' list,
  stateRelEq (stack_description list ptr ptr' R) R' ->
  (stateRelEq (stack_description list ptr ptr'
    (normalize_c (placeOnTop_c R H n))) R' -> P) -> P.
Proof.
  intros until list; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_stack_lemma_leq_hyp : forall P n R (H:stackCloud R) R' ptr ptr' list,
  stateRelLeq (stack_description list ptr ptr' R) R' ->
  (stateRelLeq (stack_description list ptr ptr'
    (normalize_c (placeOnTop_c R H n))) R' -> P) -> P.
Proof.
  intros until list; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_stack_lemma_judgment_hyp : forall P n R (H:stackCloud R) Q p p' l l' ptr ptr' list,
  judgment (stack_description list ptr ptr' R ## Q) p p' l l' ->
  (judgment (stack_description list ptr ptr' 
    (normalize_c (placeOnTop_c R H n)) ## Q) p p' l l' -> P) -> P.
Proof.
  intros until list; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

Lemma focusing_c_stack_lemma_totaljudgment_hyp : forall P n R (H:stackCloud R) Q R' p p' l1 l1' l2 l2' ptr ptr' list,
  explicit_judgment (stack_description list ptr ptr' R ## Q) R' p p' l1 l1' l2 l2' ->
  (explicit_judgment (stack_description list ptr ptr' 
    (normalize_c (placeOnTop_c R H n)) ## Q) R' p p' l1 l1' l2 l2' -> P) -> P.
Proof.
  intros until list; intros H1 H2.
  rewrite (focusing_c n R H) in H1.  
  apply (H2 H1).
Qed.

(************* lemmas for stateRel for the righthand side argument **********)

Lemma focusing_lemma' : forall n R (H:stack R) R',
  stateRelEq R' (normalize (placeOnTop R H n)) -> stateRelEq R' R.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

Lemma focusing_lemma_leq' : forall n R (H:stack R) R',
  stateRelLeq R' (normalize (placeOnTop R H n)) -> stateRelLeq R' R.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

Lemma focusing_lemma_totaljudgment' : forall n R (H:stack R) R' p p' l1 l1' l2 l2',
  explicit_judgment R' (normalize (placeOnTop R H n)) p p' l1 l1' l2 l2' -> 
  explicit_judgment R' R p p' l1 l1' l2 l2'.
Proof.
  intros.
  rewrite (focusing n R H);  assumption.
Qed.

Ltac cleanup := unfold normalize; simpl projT1.

Ltac cleanuphyp hyp := cleanup ;clear hyp; intro hyp.

Ltac cleanup_c := unfold normalize_c; simpl projT1.

Ltac cleanuphyp_c hyp := cleanup_c ;clear hyp; intro hyp.

Ltac focus n :=
  match goal with
    | |- stateRelEq ?P ?Q => 
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma n P H Q); cleanup)
    | |- stateRelLeq ?P ?Q =>
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma_leq n P H Q); cleanup)
    | |- judgment ?P ?p ?p' ?l ?l' =>
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma_judgment n P H p p' l l'); cleanup)
    | |- derivation ?P ?p ?p' ?l ?l' ?k =>
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma_derivation n P H p p' l l' k); cleanup)
    | |- explicit_judgment ?P ?Q ?p ?p' ?l1 ?l1' ?l2 ?l2' =>
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma_totaljudgment n P H Q p p' l1 l1' l2 l2'); cleanup)
    | |- Interp ?P ?s ?s' ?p ?p' ?k =>
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma_interp n P H s s' p p' k); cleanup)
  end.

Ltac focus_conj n :=
  match goal with
    | |- judgment (cloudToState ?P ## ?Q) ?p ?p' ?l ?l' =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_context_lemma_judgment n P H Q p p' l l'); cleanup_c)
    | |- derivation (cloudToState ?P ## ?Q) ?p ?p' ?l ?l' ?k =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_context_lemma_derivation n P H Q p p' l l' k); cleanup_c)
    | |- explicit_judgment (cloudToState ?P ## ?Q) ?R ?p ?p' ?l1 ?l1' ?l2 ?l2' =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_context_lemma_totaljudgment n P H R Q p p' l1 l1' l2 l2'); cleanup_c)
    | |- stateRelEq (stack_description ?l ?p ?p' ?P) ?Q =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_stack_lemma n P H Q p p' l); cleanup_c)
    | |- stateRelLeq (stack_description ?l ?p ?p' ?P) ?Q =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_stack_lemma_leq n P H Q p p' l); cleanup_c)
    | |- judgment (stack_description ?list ?ptr ?ptr' ?P ## ?Q) ?p ?p' ?l ?l' =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_stack_lemma_judgment n P H Q p p' l l' ptr ptr' list); cleanup_c)
    | |- explicit_judgment (stack_description ?list ?ptr ?ptr' ?P ## ?Q) ?R' ?p ?p' ?l1 ?l1' ?l2 ?l2' =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_stack_lemma_totaljudgment n P H R' Q p p' l1 l1' l2 l2' ptr ptr' list); cleanup_c)
    | |- cloudRelEq ?P ?Q =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_lemma_cloud n P H Q); cleanup_c)
    | |- cloudRelLeq ?P ?Q =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_lemma_cloudleq n P H Q); cleanup_c)
    | |- stateRelEq (cloudToState ?P) ?Q =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_lemma n P H Q); cleanup_c)
    | |- stateRelLeq (cloudToState ?P) ?Q =>
      certificate_c P ltac:(fun H =>
        simple apply (focusing_c_lemma_leq n P H Q); cleanup_c)
  end.

Ltac focus_hyp n :=
  match goal with
    | hyp:stateRelEq ?P ?Q |- ?goal => 
      certificate P ltac:(fun H =>
        eapply (focusing_lemma_hyp goal n P H Q) in hyp;
          [assumption | cleanuphyp hyp])
    | hyp:stateRelLeq ?P ?Q |- ?goal =>
      certificate P ltac:(fun H =>
        apply (focusing_lemma_leq_hyp goal n P H Q) in hyp; 
          [assumption | cleanuphyp hyp])
    | hyp:judgment ?P ?p ?p' ?l ?l' |- ?goal =>
      certificate P ltac:(fun H =>
        apply (focusing_lemma_judgment_hyp goal n P H p p' l l') in hyp; 
          [assumption | cleanuphyp hyp])
    | hyp:explicit_judgment ?P ?Q ?p ?p' ?l1 ?l1' ?l2 ?l2' |- ?goal =>
      certificate P ltac:(fun H =>
        apply (focusing_lemma_totaljudgment_hyp goal n P H Q p p' l1 l1' l2 l2') in hyp; 
          [assumption | cleanuphyp hyp])
    | hyp:Interp ?P ?s ?s' ?p ?p' ?k |- ?goal =>
      certificate P ltac:(fun H =>
        apply (focusing_lemma_interp_hyp goal n P H s s' p p' k) in hyp; 
          [assumption | cleanuphyp hyp])
  end.

Ltac focus_conj_hyp n :=
  match goal with
    | hyp:cloudRelEq ?P ?Q |- ?goal =>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_lemma_cloud_hyp goal n P H Q) in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:cloudRelLeq ?P ?Q |- ?goal =>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_lemma_cloudleq_hyp goal n P H Q) in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:stateRelEq (cloudToState ?P) ?Q |- ?goal =>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_lemma_hyp goal n P H Q) in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:stateRelLeq (cloudToState ?P) ?Q |- ?goal =>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_lemma_leq_hyp goal n P H Q) in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:judgment (cloudToState ?P ## ?Q) ?p ?p' ?l ?l' |- ?goal =>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_lemma_judgment_hyp goal n P H Q p p' l l') in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:explicit_judgment (cloudToState ?P ## ?Q) ?R' ?p ?p' ?l1 ?l1' ?l2 ?l2' |- ?goal=>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_lemma_totaljudgment_hyp goal n P H Q R' p p' l1 l1' l2 l2') in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:stateRelEq (stack_description ?list ?ptr ?ptr' ?P) ?Q |- ?goal =>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_stack_lemma_hyp goal n P H Q ptr ptr' list) in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:stateRelLeq (stack_description ?list ?ptr ?ptr' ?P) ?Q |- ?goal =>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_stack_lemma_leq_hyp goal n P H Q ptr ptr' list) in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:judgment (stack_description ?list ?ptr ?ptr' ?P ## ?Q) ?p ?p' ?l ?l' |- ?goal =>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_stack_lemma_judgment_hyp goal n P H Q p p' l l' ptr ptr' list) in hyp;  
          [assumption | cleanuphyp_c hyp])
    | hyp:explicit_judgment (stack_description ?list ?ptr ?ptr' ?P ## ?Q) ?R' ?p ?p' ?l1 ?l1' ?l2 ?l2' |- ?goal=>
      certificate_c P ltac:(fun H =>
        apply (focusing_c_stack_lemma_totaljudgment_hyp goal n P H Q R' p p' l1 l1' l2 l2' ptr ptr' list) in hyp;  
          [assumption | cleanuphyp_c hyp])
  end.

Ltac focus' n :=
  match goal with
    | |- stateRelEq ?Q ?P => 
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma' n P H Q); cleanup)
    | |- stateRelLeq ?Q ?P =>
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma_leq' n P H Q); cleanup)
    | |- explicit_judgment ?Q ?P ?p ?p' ?l1 ?l1' ?l2 ?l2' =>
      certificate P ltac:(fun H =>
        simple apply (focusing_lemma_totaljudgment' n P H Q p p' l1 l1' l2 l2'); cleanup)
  end.

Lemma test_stack : forall P1 P2 P3,
  stateRelEq (P1 ## P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P2)
  (P2 ## P1 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3 ##
   P3 ## P3 ## P3 ## P3 ## P3 ## P3).
Proof.
  intros.
  focus 42.
  reflexivity.
Qed.


Lemma test_normalization : forall P1 P2 P3,
  stateRelEq ((P1 ## P2) ## (P1 ## P3) ## statePtrue)
  ((P1 ## P2) ## (P1 ## P3) ## statePtrue).
Proof.
  intros.
  focus 0; focus' 0.
  reflexivity.
Qed.
