(* judgment.v

   Define the partial and total correctness judgments.
   
   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export pointers.
Require Import Basics.

Definition derivation (Scond:stateRel) (p p':program) (l l' k:nat) :=
   let Pcond := Perp Scond in natrel Pcond p p' l l' k.

Add Morphism derivation with signature 
  stateRelLeq --> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat) ==> (@eq nat) ==> impl
as derivationcompat_leq.
Proof.
  intros; subst; unfold impl, derivation.
  apply perpContravariant; assumption.
Qed.

Add Morphism derivation with signature
  inverse stateRelLeq --> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat) ==> (@eq nat) ==> inverse impl
as derivationcompat_leq2.
Proof.
  intros x y H; intros; rewrite <- H.
  reflexivity.
Qed.

Add Morphism derivation with signature 
              stateRelEq ==> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat) ==> (@eq nat) ==> iff
as derivationcompat.
Proof.
  intros; subst; unfold derivation; simpl.
  split; intro.
  rewrite H in H0; assumption.
  rewrite <- H in H0; assumption.
Qed.

Lemma derivation_leq : forall P Q p p' l l' k, stateRelLeq P Q ->
                                  derivation Q p p' l l' k ->
                                  derivation P p p' l l' k.
Proof.
  intros until k; intro H.
  setoid_rewrite <- H; trivial.
Qed.

Definition judgment (Scond:stateRel) (p p':program) (l l':nat) :=
   forall k , derivation Scond p p' l l' k .

Add Morphism judgment with signature 
              stateRelLeq --> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat) ==> impl
as judgmentcompat_leq.
Proof.
  unfold judgment,impl; intuition.
  rewrite H; trivial.
Qed.

Add Morphism judgment with signature
              stateRelLeq ++> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat) ==> inverse impl
as judgmentcompat_leq2.
Proof.
  intros; rewrite <- H; intro; trivial.
Qed.

Add Morphism judgment with signature 
              stateRelEq ==> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat) ==> iff
as judgmentcompat.
Proof.
  unfold judgment; intuition.
  rewrite <- H; trivial.
  rewrite H; trivial.
Qed.

Notation "|= l1 , l2 : S ; p1 , p2 " := (judgment S p1 p2 l1 l2) (at level 75, l2 at next level,
 S at next level, p1 at next level, p2 at next level).

Lemma judgment_false : forall p p' l l', judgment statePfalse p p' l l'.
Proof.
  intros; unfold judgment, derivation,Perp; simpl.
  unfold perpOf; intros.
  destruct H0.
Qed.

Lemma judgment_leq : forall P Q p p' l l', stateRelLeq P Q ->
                                  judgment Q p p' l l' ->
                                  judgment P p p' l l'.
Proof.
  intros until l'; intro H.
  rewrite <- H; trivial.
Qed.

Lemma judgment_weakening : forall P Q p p' l l', judgment P p p' l l' ->
                                  judgment (P##Q) p p' l l'.
Proof.
  intros until l'; intro H.
  pose (stateRelTensor_weakening P Q).
  rewrite <- s in H.
  assumption.
Qed.


Definition explicit_judgment Rpre Rpost p p' start start' finish finish':= 
  forall k s1 s1', Interp Rpre s1 s1' p p' k ->  
    (exists j1, exists j1', exists s2 ,exists s2',
         kstepreduce j1 p s1 s2 start finish
      /\ kstepreduce j1' p' s1' s2' start' finish' 
      /\ Interp Rpost s2 s2' p p' k).

Add Morphism explicit_judgment with signature 
  stateRelLeq --> stateRelLeq ++> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat)  ==> (@eq nat) ==> (@eq nat) ==> impl
as explicitjudgmentcompat_leq.
Proof.
  unfold explicit_judgment,impl; intuition.
  unfold flip in *; rewrite H in H2.
  specialize (H1 k s1 s1' H2).
  destruct H1 as (j1 & j1' & s2 & s2' & Hstep & Hstep' & Hx0).
  exists j1; exists j1'; exists s2; exists s2'.
  rewrite <- H0; trivial.
  intuition.
Qed.

Add Morphism explicit_judgment with signature
  stateRelLeq ++> stateRelLeq --> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat)  ==> (@eq nat) ==> (@eq nat) ==> inverse impl
as explicitjudgmentcompat_leq2.
Proof.
  intros; rewrite <- H, <- H0; intro; trivial.
Qed.

Add Morphism explicit_judgment with signature 
  stateRelEq ==> stateRelEq ==> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat)  ==> (@eq nat) ==> (@eq nat) ==> iff
as explicitjudgmentcompat.
Proof.
  intuition.
  apply stateRelEq_comm in H; apply stateRelEq_Leq in H.
  apply stateRelEq_Leq in H0.
  rewrite <- H0;  rewrite <- H in H1; assumption.
  apply stateRelEq_Leq in H.
  apply stateRelEq_comm in H0; apply stateRelEq_Leq in H0.
  rewrite H0 in H1; rewrite <- H in H1 ; assumption.
Qed.

Lemma explicitjudgment_false : forall P p p' l1 l1' l2 l2', explicit_judgment statePfalse P p p' l1 l1' l2 l2'.
Proof.
  intros; unfold explicit_judgment; intros.
  destruct H.
Qed.

Lemma explicitjudgment_leq : forall Inv P Q p p' l1 l1' l2 l2',
  stateRelLeq P Q ->
  explicit_judgment Inv P p p' l1 l1' l2 l2' ->
  explicit_judgment Inv Q p p' l1 l1' l2 l2'.
Proof.
  intros until l2'; intro H.
  rewrite <- H; trivial.
Qed.


Definition total_judgment Rpre Rpost p p' start start' := 
  forall l l', explicit_judgment (Rpre l l') Rpost p p' start start' l l'.


Lemma total_to_partial : forall n RPost Ra p p' l l',
             total_judgment (fun l => fun l' => (n |-> {l,l'}) ## Ra ) RPost p p' l l' ->
             judgment ((n |=> Perp RPost) ## Ra ) p p' l l'.
Proof.
  intros until l'; intros Hintern.
  unfold judgment, derivation; unfold Perp at 1; simpl; unfold perpOf; intuition;
  unfold Interp, ptsto in H0; simpl in H0;
  destruct H0 as (f1 & f2 & f'1 & f'2 & Hdisj & Hjoin & Hdisj' & Hjoin' & H0);
  destruct H0 as ([l1 [l2 (Hl1,(Hl2,HPost))]]);
  unfold total_judgment, explicit_judgment in *;
  specialize (Hintern l1 l2 j s s');
  destruct Hintern as (k1 & k2 & s_ret & s_ret' & Hreduce & Hreduce' & Hinterp); intuition;
  unfold Interp , stateRelTensor; simpl;
  try (exists f1; exists f2; exists f'1; exists f'2; intuition);
  try (exists l1;exists l2; intuition);
  unfold perpOf in HPost;
  destruct j; try solve [inversion H1];
  specialize (HPost s_ret s_ret' (S j) (le_refl (S j))).

  eapply terminate_reduce; try apply Hreduce'.
  destruct HPost as (Hp,Hp'); trivial.
  apply Hp.
  eapply (stepterm_reduce k1 (k1 + S j) (S j)); [trivial| apply Hreduce | idtac].
  eapply kstep_induction; try eassumption.
  omega.

  eapply terminate_reduce; try apply Hreduce.
  destruct HPost as (Hp,Hp'); trivial.
  apply Hp'.
  eapply (stepterm_reduce k2 (k2 + S j) (S j)); [trivial| apply Hreduce' | idtac].
  eapply kstep_induction; try eassumption.
  omega.
Qed.


Lemma sem_partial_src : forall start start' Rpre1 Rpre2 p p' m m' source source',
  (forall s1 s1' j, Interp Rpre1 s1 s1' p p' j ->
    Interp Rpre2 (update s1 m (sem_src source s1)) (update s1' m' (sem_src source' s1')) p p' j /\
    sem_instr (p start) s1 start = Some (update s1 m (sem_src source s1) , start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' m' (sem_src source' s1') , start'+1)) ->
  judgment Rpre2 p p' (start+1) (start'+1) ->
  judgment Rpre1 p p' start start'.
Proof.
  intros until source'; intros Hstep Hnext.
  unfold judgment, derivation, Perp in *; simpl in *;
  unfold perpOf in *; intro k; intros s1 s1' j Hineq Hpre.
  split; intro hyp; destruct j; try inversion hyp;
  assert (Hstep2 := Hstep s1 s1' (S j) Hpre);
  rename Hstep into Hstep_aux;
  destruct Hstep2 as (Hpre2 &Hstep & Hstep');
  assert (Hj : j <= k); try omega;
  set (s2 := (update s1 m (sem_src source s1))) in *;
  set (s2' := (update s1' m' (sem_src source' s1'))) in *;
  specialize (Hnext k s2 s2' j Hj);
  destruct Hnext as (Hnext & Hnext'); try assumption;
  simpl in hyp.
  unfold Interp in *; eapply stateRel_cond;
  [apply Hpre2 | apply subPfunction_refl | apply subPfunction_refl | omega].
  
  erewrite Hstep in hyp.
  specialize (Hnext hyp); destruct Hnext as (j' & Hnext).
  exists (S j'); simpl.
  erewrite Hstep'; assumption.

  unfold Interp in *; eapply stateRel_cond;
  [apply Hpre2 | apply subPfunction_refl | apply subPfunction_refl | omega].
  
  erewrite Hstep' in hyp.
  specialize (Hnext' hyp); destruct Hnext' as (j' & Hnext').
  exists (S j'); simpl.
  erewrite Hstep; assumption.
Qed.  


Lemma sem_total_src : forall start start' Rpre1 Rpre2 Rpost lpost lpost' p p' m m' source source' ,
  (forall s1 s1' j, Interp Rpre1 s1 s1' p p' j ->
    Interp Rpre2 (update s1 m (sem_src source s1)) (update s1' m' (sem_src source' s1')) p p' j /\
    sem_instr (p start) s1 start = Some (update s1 m (sem_src source s1) , start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' m' (sem_src source' s1') , start'+1)) ->
  explicit_judgment Rpre2 Rpost p p' (start+1) (start'+1) lpost lpost' ->
  explicit_judgment Rpre1 Rpost p p' start start' lpost lpost'.
Proof.
  intros until source'; intros Hstep Hnext.
  intro k; intros s1 s1' Hpre.
  set (s2 := (update s1 m (sem_src source s1))) in *;
  set (s2' := (update s1' m' (sem_src source' s1'))) in *.

  assert (Hstep2 := Hstep s1 s1' k Hpre);
  rename Hstep into Hstep_aux;
  destruct Hstep2 as (Hpre2 &Hstep & Hstep').
  specialize (Hnext k s2 s2' Hpre2);
  destruct Hnext as (j1 & j1' & s & s' & Hreduce & Hreduce' & Hnext); try assumption.
  
  exists (S j1); exists (S j1'); exists s; exists s'.
  simpl; rewrite Hstep, Hstep'.
  intuition.
Qed.  

Lemma sem_partial : forall start start' Rpre1 Rpre2 p p' m m' l l',
  (forall s1 s1' j, Interp Rpre1 s1 s1' p p' j ->
    Interp Rpre2 (update s1 m l) (update s1' m' l') p p' j /\
    sem_instr (p start) s1 start = Some (update s1 m l, start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' m' l', start'+1)) ->
  judgment Rpre2 p p' (start+1) (start'+1) ->
  judgment Rpre1 p p' start start'.
Proof.
  intros until l'; intros Hstep Hnext.
  unfold judgment, derivation, Perp in *; simpl in *;
  unfold perpOf in *; intro k; intros s1 s1' j Hineq Hpre.
  split; intro hyp; destruct j; try inversion hyp;
  assert (Hstep2 := Hstep s1 s1' (S j) Hpre);
  rename Hstep into Hstep_aux;
  destruct Hstep2 as (Hpre2 &Hstep & Hstep');
  assert (Hj : j <= k); try omega;
  set (s2 := (update s1 m l)) in *;
  set (s2' := (update s1' m' l')) in *;
  specialize (Hnext k s2 s2' j Hj);
  destruct Hnext as (Hnext & Hnext'); try assumption;
  simpl in hyp.
  unfold Interp in *; eapply stateRel_cond;
  [apply Hpre2 | apply subPfunction_refl | apply subPfunction_refl | omega].
  
  erewrite Hstep in hyp.
  specialize (Hnext hyp); destruct Hnext as (j' & Hnext).
  exists (S j'); simpl.
  erewrite Hstep'; assumption.

  unfold Interp in *; eapply stateRel_cond;
  [apply Hpre2 | apply subPfunction_refl | apply subPfunction_refl | omega].
  
  erewrite Hstep' in hyp.
  specialize (Hnext' hyp); destruct Hnext' as (j' & Hnext').
  exists (S j'); simpl.
  erewrite Hstep; assumption.
Qed.  

Lemma sem_total: forall start start' Rpre1 Rpre2 Rpost lpost lpost' p p' m m' l l',
  (forall s1 s1' j, Interp Rpre1 s1 s1' p p' j ->
    Interp Rpre2 (update s1 m l) (update s1' m' l') p p' j /\
    sem_instr (p start) s1 start = Some (update s1 m l , start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' m' l' , start'+1)) ->
  explicit_judgment Rpre2 Rpost p p' (start+1) (start'+1) lpost lpost' ->
  explicit_judgment Rpre1 Rpost p p' start start' lpost lpost'.
Proof.
  intros until l'; intros Hstep Hnext.
  intro k; intros s1 s1' Hpre.
  set (s2 := update s1 m l) in *;
  set (s2' := update s1' m' l') in *.

  assert (Hstep2 := Hstep s1 s1' k Hpre);
  rename Hstep into Hstep_aux;
  destruct Hstep2 as (Hpre2 &Hstep & Hstep');
  specialize (Hnext k s2 s2' Hpre2);
  destruct Hnext as (j1 & j1' & s & s' & Hreduce & Hreduce' & Hnext); try assumption.
  
  exists (S j1); exists (S j1'); exists s; exists s'.
  simpl; rewrite Hstep, Hstep'.
  intuition.
Qed.


Lemma sem_total2 : forall Rpre Rpost p p' start start' l l' lpost lpost',
  (forall s1 s1' j, Interp Rpre s1 s1' p p' j -> 
    sem_instr (p start) s1 start = Some (s1, l) /\
    sem_instr (p' start') s1' start' = Some (s1', l')) ->
  explicit_judgment Rpre Rpost p p' l l' lpost lpost' ->
  explicit_judgment Rpre Rpost p p' start start' lpost lpost'.
Proof.
  intros until lpost'; intros hyp Hnext.
  unfold explicit_judgment.
  intros until s1'; intro Hinterp.

  specialize (Hnext k s1 s1' Hinterp).
  destruct Hnext as (j1 & j1' & s & s' & Hreduce & Hreduce' & Hnext); try assumption.
  exists (S j1); exists (S j1'); exists s; exists s'.
  specialize (hyp s1 s1' k Hinterp).
  destruct hyp as (Hsem & Hsem').
  simpl; rewrite Hsem, Hsem'.
  intuition.
Qed.

Lemma sem_partial2 : forall p p' start start' m m' n n' Rpre l l',
  (forall s1 s1' j, Interp (({m,m'} |->{n,n'}) ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, l) /\
    sem_instr (p' start') s1' start' = Some (s1', l')) ->
  judgment (({m,m'} |->{n,n'}) ## Rpre) p p' l l' -> 
  judgment (({m,m'} |->{n,n'}) ## Rpre) p p' start start'.
Proof.
  intros until l'; intros Hstep Hnext.
  unfold judgment, derivation, Perp in *; simpl in *.
  unfold perpOf in *; intro k; intros s1 s1' j Hineq Hpre.
  split; intro hyp; destruct j; try inversion hyp;
  specialize (Hstep s1 s1' (S j) Hpre);
  destruct Hstep as (Hstep & Hstep');
  assert (Hj : j <= k); try omega;
  specialize (Hnext k s1 s1' j Hj);
  destruct Hnext as (Hnext & Hnext'); try assumption;
  simpl in hyp.
  unfold Interp in *; eapply stateRel_cond;
  [apply Hpre | apply subPfunction_refl | apply subPfunction_refl | omega].
  
  erewrite Hstep in hyp.
  specialize (Hnext hyp); destruct Hnext as (j' & Hnext).
  exists (S j'); simpl.
  erewrite Hstep'; assumption.

  unfold Interp in *; eapply stateRel_cond;
  [apply Hpre | apply subPfunction_refl | apply subPfunction_refl | omega].
  
  erewrite Hstep' in hyp.
  specialize (Hnext' hyp); destruct Hnext' as (j' & Hnext').
  exists (S j'); simpl.
  erewrite Hstep; assumption.
Qed.

Lemma total_refl : forall Rpre Rpost p p' l l', stateRelLeq Rpre Rpost -> explicit_judgment Rpre Rpost p p' l l' l l'.
Proof.
  intros until l'; intro Heq.
  unfold total_judgment; intros k s1 s1' Hpre. 
  exists 0; exists 0; exists s1; exists s1'.
  simpl; intuition.
  rewrite <- Heq; assumption.
Qed.
