(* lambda.v

   Basic CBV lambda calculus (PCF)

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import utility.

(* -------------------
   Some useful results
   ------------------- *)

Section Lists_extras.

  Variable A : Type.

  Set Implicit Arguments.

  (* cons_nth *)

  Fixpoint cons_nth (a:A) (l:list A) (m:nat) {struct l} : list A :=
    match l with
      | nil => a :: nil
      | b :: bs =>
	match m with
	  | 0 => a :: l
	  | S m' => b :: cons_nth a bs m'
	end
    end.

  Lemma cons_nth_S : forall a b l m, cons_nth a (b::l) (S m) = b :: (cons_nth a l m).
  Proof.
    firstorder.
  Qed.

  Lemma nth_error_nil : forall n, nth_error (nil : list A) n = error.
  Proof.
    destruct n; trivial.
  Qed.

  Lemma nth_error_cons : forall (l:list A) n a b, (nth_error l n = Some a) = (nth_error (b::l) (S n) = Some a).
    simpl; auto.
  Qed.

  Lemma nth_error_cons_nth_lt : forall (l:list A) n a b m (h:n<m), 
         (nth_error l n = Some a) -> (nth_error (cons_nth b l m) n = Some a).
  Proof.    
    induction l.
    intros; rewrite (nth_error_nil n) in H; inversion H.
    intros until m;  case m.
    intro h;  absurd (n < 0); [apply  (lt_n_O n) | assumption].
    case n ;[ trivial | simpl; intros ; apply (IHl n0 a0 b n1) ; [omega |trivial]].
  Qed.

  Lemma nth_error_cons_nth_ge : forall (l:list A) n a b m (h:n>=m), 
         (nth_error l n = Some a) -> (nth_error (cons_nth b l m) (S n) = Some a).
  Proof.
    induction l.
    intro n; rewrite (nth_error_nil n); intros; inversion H.
    intros until m;case m.
    trivial.
    case n.
    intros n0 h; absurd (0 >= S n0) ; [omega | trivial].
    intros n0 n1 h; simpl; simpl in IHl.
    apply IHl; omega.
  Qed.

  Lemma nth_error_0 : forall (l:list A) (a:A), nth_error l 0 = Some a <-> (exists l', l = a::l').
  Proof.
    intros; split.
    induction l.
    simpl;  intro.
    inversion H.
    simpl;  intro.
    inversion H.
    exists l;trivial.
    intro; destruct H.
    rewrite H.
    simpl;trivial.
  Qed.
  
  Lemma nth_error_Sn : forall (l:list A) (a:A) (n:nat), 
    nth_error l (S n) = Some a <-> (exists l', exists a0,  l = a0::l' /\ nth_error l' n = Some a).
  Proof.
    intros; split.
    induction l.
    simpl;  intro.
    inversion H.
    simpl;  intro.
    exists l;exists a0;intuition.
    intro; destruct H as [l' [a0 (H0,H1)]].
    rewrite H0.
    simpl;trivial.
  Qed.

  Lemma split_list : forall (a:A) l , a :: l = (a :: nil) ++ l.
  Proof.
    trivial.
  Qed.

  Fixpoint takedrop n (l1 l2 : list A) {struct n} := match n with
    | 0 => (l1,l2)
    | S n => match l2 with 
       | nil => (l1,l2)
       | h::t => takedrop n (l1++(h::nil)) t
       end
    end. 

  Definition fuse (couple:list A * list A) := match couple with (l1,l2) => l1++l2 end.
  
  Lemma fuse_unfold : forall l1 l2, fuse (l1,l2) = l1++l2.
  Proof.
    unfold fuse; intuition.
  Qed.

  Opaque fuse.

  Unset Implicit Arguments.

End Lists_extras.

Ltac clever_append n list:=
  let foo := fresh "clever_append" in
  assert (foo : list = (fuse (takedrop 11 nil list))); try (simpl;trivial);
  simpl in foo; rewrite fuse_unfold in foo.


(* -------------------
   CBV lambda calculus
   ------------------- *)

Set Implicit Arguments.

(*** Types and contexts ***)

Inductive ExpType : Type :=
  | Int : (nat -> Prop) -> ExpType
  | Bool : (bool -> Prop) -> ExpType
  | Arrow : ExpType -> ExpType -> ExpType
  | Pair : ExpType -> ExpType -> ExpType.

Notation "a ---> b" := (Arrow a b) (at level 55).

Notation "a ** b" := (Pair a b) (at level 55).

Definition EnvType := list ExpType.


(*** Typed terms  ***)

Reserved Notation "a <-< b'" (at level  75).

Inductive TypeCoercion : ExpType -> ExpType -> Type :=
  | IntCoerce : forall (P1 P2 : nat -> Prop), (forall n, P1 n -> P2 n) -> Int P1 <-< Int P2
  | BoolCoerce : forall (P1 P2 : bool -> Prop), (forall b, P1 b -> P2 b) -> Bool P1 <-< Bool P2
  | PairCoerce : forall a a' b b', a <-< a' -> b <-< b' -> Pair a b <-< Pair a' b'
  | ArrowCoerce : forall a a' b b', a' <-< a -> b <-< b' -> Arrow a b <-< Arrow a' b'
where "a <-< b" := (TypeCoercion a b).

Inductive Exp : Set :=
  | INT : nat -> Exp
  | BOOL : bool -> Exp
  | VAR : nat -> Exp
  | LAMBDA : Exp -> Exp
  | FIX : Exp -> Exp
  | APP : Exp -> Exp -> Exp
  | OP : (nat->nat->nat) -> Exp -> Exp -> Exp
  | GT : Exp -> Exp -> Exp
  | IF_THEN_ELSE : Exp -> Exp -> Exp -> Exp
  | P : Exp -> Exp -> Exp
  | FST : Exp -> Exp
  | SND : Exp -> Exp.

(* Now this is the typing judgement *)

Definition AbstractOp (op : nat -> nat -> nat) (P1 P2 : nat -> Prop) n := exists x, exists y, P1 x /\ P2 y /\ (op x y = n).

Definition AbstractGt (P1 P2 : nat -> Prop) (b:bool) := (exists n, exists m, P1 n /\ P2 m /\ ((n > m) <-> Is_true b)).

Reserved Notation "a |-- b ':::' c" (at level  75).

Inductive Typing : EnvType -> ExpType -> Exp -> Type :=
  | TINT : forall env (P : nat->Prop) (n:nat), P n ->    env |-- INT n ::: Int P
  | TBOOL : forall env (P : bool->Prop) (b:bool), P b -> env |-- BOOL b ::: Bool P
  | TVAR : forall env m a, (nth_error env m = Some a) -> env |-- VAR m ::: a
  | TLAMBDA : forall env a b body,                  a :: env |-- body ::: b -> 
                                                         env |-- LAMBDA body ::: a ---> b
  | TFIX : forall env t body,          t :: env |-- (LAMBDA body) ::: t -> 
                                            env |-- FIX (LAMBDA body) ::: t
  | TAPP : forall env a b rator rand, env |-- rator::: a ---> b ->
                                      env |-- rand ::: a ->
                                      env |-- APP rator rand ::: b 
  | TOP : forall op env P1 P2 exp1 exp2, env |-- exp1 ::: Int P1 ->
                                         env |-- exp2 ::: Int P2 ->
                                         env |-- OP op exp1 exp2 ::: Int (AbstractOp op P1 P2)
  | TGT : forall env P1 P2 exp1 exp2,                    env |-- exp1 ::: Int P1 ->
                                                         env |-- exp2 ::: Int P2 ->
                                                         env |-- GT exp1 exp2 ::: Bool (AbstractGt P1 P2) 
  | TP : forall env a b e1 e2,       env |-- e1 ::: a -> env |--  e2 ::: b -> env |-- P e1 e2 ::: (Pair a b)
  | TFST : forall env a b e,                             env |-- e ::: (Pair a b) -> env |-- FST e ::: a
  | TSND : forall env a b e,                             env |-- e ::: (Pair a b) -> env |-- SND e ::: b
  | TIF_THEN_ELSE : forall env P a expcond exp1 exp2,       env |-- expcond ::: Bool P -> 
                                                 (P true -> env |-- exp1 ::: a) -> 
                                                (P false -> env |-- exp2 ::: a) -> 
                                                            env |-- IF_THEN_ELSE expcond exp1 exp2 ::: a
  | TCOERCE : forall env a b exp, a <-< b -> env |-- exp ::: a -> env |-- exp ::: b
where " env '|--' exp ':::' type" := (Typing env type exp).


Definition Sing (n:nat) := (fun x => x=n).

Definition SInt n := Int (Sing n).

Definition Succ := LAMBDA (OP plus (VAR 0) (INT 1)).

Definition Factorial := FIX (LAMBDA (IF_THEN_ELSE (GT (VAR 0) (INT 0))  
            (OP mult (VAR 0) (APP (VAR 1)(OP minus (VAR 0) (INT 1)))) (INT 1))).

Fixpoint real_factorial n := match n with 
  | 0 => 1
  | S m => n * real_factorial m
end.

Lemma AbstractGt_absurd : forall n m, n <= m -> (AbstractGt (Sing n) (Sing m) true <-> False).
Proof.
  unfold Sing; firstorder.
Qed.

Lemma AbstractGt_false : forall m n, AbstractGt (Sing m) (Sing n) false <-> (m <= n).
Proof.
  unfold Sing; firstorder.
  rewrite H0 in H1; rewrite H in H1; simpl in H1;  omega.
  unfold AbstractGt; exists m;exists n; firstorder.
Qed.

Definition ClosedExp := Typing nil.


Inductive wannaBeFunctionType : ExpType -> Prop :=
  | IsArrow : forall a b, wannaBeFunctionType (a ---> b).

Lemma wannaBeFunctionTypeCoercion : forall a b, a <-< b -> wannaBeFunctionType a -> wannaBeFunctionType b.
Proof.
  intros a b Hcoerce Ha; induction Hcoerce; try inversion Ha.
  apply IsArrow.
Qed.

Lemma shape_of_lambda : forall env e e' a, (env |-- e :::a) -> (e=LAMBDA e') -> wannaBeFunctionType a.
Proof.
  intros until a; intro typing; induction typing; try (intros He ;inversion He).
  apply IsArrow.
  eapply wannaBeFunctionTypeCoercion; try eassumption; intuition.
Qed.


Unset Implicit Arguments.
