(* nat_relations.v

   Define logical relation on code pointers

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export programs.

Record natRel : Type := mkNatRel {
  natrel :> program -> program -> nat -> nat -> nat -> Prop;
  natRel_cond : forall p p' l l' k1 k2, natrel p p' l l' k1 -> (k2 <= k1) -> natrel p p' l l' k2
}.

Implicit Arguments natRel_cond [].

Definition natRelEq (P Q: natRel) := forall p p' l l' k , P p p' l l' k <-> Q p p' l l' k.

Lemma natRelEq_refl : forall (P : natRel), natRelEq P P.
Proof.
  firstorder.
Qed.

Lemma natRelEq_comm : forall (P Q : natRel), natRelEq P Q -> natRelEq Q P.
Proof.
  unfold natRelEq; intros.
  hreplace H with (H p p' l l' k).
  intuition.
Qed.

Lemma natRelEq_trans : forall (P Q R : natRel), natRelEq P Q -> natRelEq Q R -> natRelEq P R.
Proof.
  unfold natRelEq; intros; simpl.
  hreplace H0 with (H0 p p' l l' k); hreplace H with (H p p' l l' k).
  firstorder.
Qed.

Hint Resolve natRelEq_refl natRelEq_trans.

Add Relation natRel natRelEq
   reflexivity proved by natRelEq_refl
   symmetry proved by natRelEq_comm
   transitivity proved by natRelEq_trans
 as natRelEqRel.

(* Ltac obligations_tactic ::= repeat red;intros. *)

Add Morphism natrel with signature 
  natRelEq ==> (@eq program)  ==> (@eq program)  ==> (@eq nat) ==> (@eq nat) ==> (@eq nat) ==> iff
as natrelcompat.
Proof.
  constructor; subst.
  unfold natRelEq in *.
  rewrite H; intuition.
  unfold natRelEq in *.
  rewrite <- H; intuition.
Qed.

Definition natRelLeq (P Q : natRel) := forall p p' l l' k , P p p' l l' k -> Q p p' l l' k.

Lemma natRelLeq_refl : forall P, natRelLeq P P.
Proof.
  firstorder.
Qed.

Lemma natRelLeq_trans : forall P Q R,  natRelLeq P Q -> natRelLeq Q R -> natRelLeq P R.
Proof.
  firstorder.
Qed.

Hint Resolve natRelLeq_refl natRelLeq_trans.

Add Relation natRel natRelLeq
   reflexivity proved by natRelLeq_refl
   transitivity proved by natRelLeq_trans
 as natRelLeqRel.

Add Morphism natRelLeq with signature natRelEq ==> natRelEq ==> iff
as prelleqcompat.
Proof.
  unfold natRelEq, natRelLeq; intuition.
  assert (foo := H p p' l l' k ); assert (foo2 := H0 p p' l l' k); assert (foo3 := H1 p p' l l' k).
  intuition.
  assert (foo := H p p' l l' k ); assert (foo2 := H0 p p' l l' k); assert (foo3 := H1 p p' l l' k).
  intuition.
Qed.

Lemma natRelEq_Leq: forall P Q, natRelEq P Q -> natRelLeq P Q.
Proof.
  unfold natRelEq,natRelLeq; intros.
  rewrite <- (H p p' l l' k); assumption.
Qed.

(* Lifting pure relicates, true and false *)

Definition natPconstantRel (P:nat -> nat -> Prop) : natRel.
Proof.
  intro P.
  refine (mkNatRel (fun p => fun p' => fun l => fun l' => fun  k => (P l l')) _).
  intuition.
Defined.

Add Morphism natPconstantRel with signature (@eq (nat->nat->Prop)) ==> natRelEq
as natPconstantRelcompat.
Proof.
  intros; subst; auto.
Qed.

Add Morphism natPconstantRel with signature (@eq (nat->nat->Prop)) ==> natRelLeq
as natPconstantRelLeqcompat.
Proof.
  intros; subst; auto.
Qed.

Definition natRel_lift (P:Prop):= natPconstantRel(fun n => fun n' => P).

(*
Add Morphism natRel_lift with signature iff ==> natRelEq
as natrelliftcompat.
Proof.
  unfold natRel_lift, natRelEq; simpl; firstorder. 
Qed.

Add Morphism natRel_lift with signature iff ==> natRelLeq
as natrelliftLeqcompat.
Proof.
  unfold natRel_lift, natRelLeq; simpl; firstorder. 
Qed.
*)
Definition natPtrue := natRel_lift (True).

Definition natPfalse := natRel_lift (False).

Definition natPconstant (n:nat) := natPconstantRel(fun m => fun m' => (m=n /\ m'=n)).

(* *******************  Modality ********************* *)

Definition Later (P:natRel) : natRel.
  intros P.
  refine (mkNatRel (fun s => fun s' => fun p => fun p' => fun k => 
    forall j, j < k -> P s s' p p' j) _).
  intuition.
Defined.

Add Morphism Later with signature 
  natRelEq ==> natRelEq
as latercompat.
Proof.
  intros R R'; subst;  unfold Later,natRelEq; intros; subst; simpl; intuition;
  eapply H; eapply H0; assumption. 
Qed.

Add Morphism Later with signature 
  natRelLeq ++> natRelLeq
as latercompat2.
Proof.
  intros R R'; subst;  unfold Later,natRelLeq; intros; subst; simpl; intuition;
  eapply H; eapply H0; assumption. 
Qed.

Definition Necessity (P:natRel) : natRel.
  intros P.
  refine (mkNatRel (fun s => fun s' => fun p => fun p' => fun k => 
    forall j, j <= k -> P s s' p p' j) _).
  intuition.
Defined.

Add Morphism Necessity with signature 
  natRelEq ==> natRelEq
as necessitycompat.
Proof.
  intros R R'; subst;  unfold Necessity,natRelEq; intros; subst; simpl; intuition;
  eapply H; eapply H0; assumption. 
Qed.

