(* Partial_Function.v

   Define partial functions as functions of type A -> option B
   Define the notion of disjoint, join, extension, etc ...

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export utility.

Set Implicit Arguments.

Section Partial_Function.

Variable A B : Type.

Definition Pfunction := A -> option B.

Definition domPfunction (f:Pfunction) x := exists y, f x = Some y.

Definition disjointPfunction (f1 f2 : Pfunction) := forall x , ~ (domPfunction f1 x /\ domPfunction f2 x).

Definition joinPfunction (f1 f2 : Pfunction) : Pfunction := (fun x => match f1 x with None => f2 x | x => x end).

Notation "f1 ';;' f2" := (joinPfunction f1 f2) (at level 65, right associativity).

Notation "f1 '<#>' f2" := (disjointPfunction f1 f2) (at level 65, right associativity).

Definition nullPfunction : Pfunction :=  (fun x => None).

Definition Pfunctiondiff (f1 f2 : Pfunction) : Pfunction := (fun x => match f2 x with Some y => None | None => f1 x end).

Definition subPfunction (f1 f2 : Pfunction) := exists f3:Pfunction, f1 <#> f3 /\ f2 = f1 ;; f3.

Definition extendsPfunction (f1 f2 : Pfunction) :=  forall x y, f1 x = Some y -> f2 x = Some y.

Notation "f1 '<<' f2" := (subPfunction f1 f2) (at level 65, right associativity).


Definition PfunctionExtendsFunction (f : A -> B) (pf:Pfunction) := 
  forall n, match pf n with None => True | Some i => (f n = i) end.

Definition constantExtension (pf : Pfunction) (b:B) : (A -> B) :=
 fun l => match pf l with Some i => i | None => b end.

Definition functionToPfunction (f:A->B) := fun a => Some (f a). 

Lemma functionTotal : forall (f:A->B) (a:A), domPfunction (functionToPfunction f) a.
intros.
exists (f a).
intuition.
Qed.



(*---------------------------------------------------------*)
(* Lemmas about Pfunction, disjointness and joining two Pfunctions *)
(*---------------------------------------------------------*)


Lemma disjointPfunction_comm : forall (f1 f2 : Pfunction), f1 <#> f2 <-> f2 <#> f1.
Proof.
firstorder.
Qed.


Lemma disjointPfunction_unit : forall (f:Pfunction), f <#> nullPfunction.
Proof.
  intros.
  unfold disjointPfunction,domPfunction.
  intros.
  intro.
  destruct H.
  witness n in H0.
  simpl in H0.
  inversion H0.
Qed.

Implicit Arguments disjointPfunction_unit [x].


Lemma join_unit : forall f:Pfunction, f = f ;; nullPfunction.
Proof.
  unfold joinPfunction, nullPfunction, Pfunction.
  intros.
  unfold Pfunction in f.
  rewrite <- Extensionality.
  intros.
  elim (f a);  tauto.
Qed.


Lemma join_disj_comm : forall f1 f2 : Pfunction, f1 <#> f2 -> f1 ;; f2 = f2 ;; f1.
Proof.
   unfold joinPfunction, Pfunction.
   intros f1 f2 Hdisj.
   rewrite <- Extensionality.
   intro x.
   dcase (f1 x).
   intros y Hy.
   dcase (f2 x).
   intros.
   absurd (f2 x = Some b).
   intro.
   apply (Hdisj x).
   unfold domPfunction.
   split.
   exists y; assumption.
   exists b; assumption.
   assumption.
   reflexivity.
   intro H. 
   elim (f2 x).
   reflexivity.
   reflexivity.
Qed.

Implicit Arguments join_disj_comm [].

   
Lemma disj_join_assoc : forall f1 f2 f3: Pfunction, 
             f1 <#> f2 -> f2 <#> f3 -> f1 <#> f3 -> f1 ;; f2 ;; f3 = (f1;; f2) ;; f3.
Proof.
  unfold disjointPfunction,joinPfunction,Pfunction.
  intros.
  rewrite <- Extensionality.
  intros.
  dcase (f1 a).
  auto.
  firstorder.
Qed.


Lemma sub_is_extends : forall f1 f2:Pfunction, f1 << f2 <-> extendsPfunction f1 f2.
Proof.
   unfold subPfunction, extendsPfunction, disjointPfunction,joinPfunction,Pfunction.
   intros. 
   split.
   intro H.
   witness f3 in H.
   intros.
   destruct H as [H1 H2].
   rewrite H2.
   simpl.
   rewrite H0.
   reflexivity.
   
   intros.
   exists (Pfunctiondiff f2 f1).
   unfold Pfunctiondiff; unfold domPfunction; split.
   intros.
   intros [Ha Hb].
   witness y in Ha.
   witness y2 in Hb.
   rewrite Ha in Hb.
   inversion Hb.

   rewrite <- Extensionality.
   intro x; dcase (f1 x).
   intros n Hn. 
   apply H.
   assumption.
   tauto.
 Qed.
   
Lemma domjoin : forall f1 f2 a, domPfunction (f1 ;; f2) a <-> domPfunction f1 a \/ domPfunction f2 a.
Proof.
   intros.
   unfold domPfunction, joinPfunction.
   split.
   intro H; destruct H.
   dcase (f1 a).
   intros.
   left; exists b; reflexivity.
   intros; right.
   rewrite H0 in H.
   exists x; assumption.
   
   intro H; destruct H.
   destruct H.
   rewrite H.
   exists x; reflexivity.
   dcase (f1 a).
   intros.
   exists b; reflexivity.
   intros.
   destruct H.
   exists x.
   assumption.
Qed.



Lemma disj_join_right : forall f1 f2 f3:Pfunction, f1 <#> (f2 ;; f3) <->  f1 <#> f2 /\ f1 <#> f3. 
Proof.
  intros; unfold disjointPfunction.
  split.
  intro; split; intro x.
  generalize (H x); rewrite (domjoin f2 f3 x).
  intuition.
  generalize (H x); rewrite (domjoin f2 f3 x).
  intuition.
  intros [H1 H2] x.
  rewrite (domjoin f2 f3 x).
  firstorder.
Qed.


Lemma disj_join_left : forall f1 f2 f3, (f1 ;; f2) <#> f3  <->  f1 <#> f3 /\ f2 <#> f3. 
Proof.
  intros.
  rewrite (disjointPfunction_comm (joinPfunction f1 f2) f3).
  rewrite (disj_join_right f3 f1 f2).
  rewrite (disjointPfunction_comm f3 f1).
  rewrite (disjointPfunction_comm f3 f2).
  tauto.
Qed.

Lemma disjPfunction_joinl : forall f1 f2 f3, f2 <#> f3 -> f1 <#> (f2 ;; f3) -> (f1 ;; f2) <#> f3.
Proof.
  intros.
  assert (Hjoin := (disj_join_right f1 f2 f3)).
  rewrite (disj_join_left f1 f2 f3); split; [intuition | assumption ].
Qed.


Lemma disjPfunction_joinr : forall f1 f2 f3, f1 <#> f2  -> (f1 ;; f2) <#> f3 -> f1 <#> (f2 ;; f3).
Proof.
  intros.
  assert (Hjoin := (disj_join_left f1 f2 f3)).
  rewrite (disj_join_right f1 f2 f3); split; [intuition | intuition].  
Qed.


Lemma Pfunction_ext : forall f1 f2, f1 <#> f2 -> f1 << (f1;; f2).
Proof.
  intros.
  unfold joinPfunction, subPfunction.
  firstorder.
Qed.

Lemma subPfunction_unit : forall f, nullPfunction << f.
Proof.
  unfold Pfunction,subPfunction; intros; simpl.
  exists f.
  intuition.
  rewrite disjointPfunction_comm; apply disjointPfunction_unit.
  rewrite join_disj_comm. 
  apply (join_unit f).
  rewrite disjointPfunction_comm; apply disjointPfunction_unit.
Qed.

Lemma subPfunction_refl : forall f, f << f.
Proof.
  unfold Pfunction.
  exists nullPfunction.
  unfold subPfunction,disjointPfunction.
  split.
  intro.
  intro.
  firstorder.
  inversion H0.
  apply join_unit.
Qed.

Lemma subPfunction_trans: forall f1 f2 f3, f1 << f2 -> f2 << f3 -> f1 << f3.
Proof.
  unfold subPfunction.
  intros.
  witness f4 in H.
  witness f5 in H0.
  exists (joinPfunction f4 f5).
  deconstruct H as [Hdisj14 Hjoin214].
  deconstruct H0 as [Hdisj25 Hjoin325].
  rewrite Hjoin214 in Hjoin325.
  rewrite Hjoin214 in Hdisj25.
  assert (H:= disj_join_left f1 f4 f5).  
  split.
  rewrite (disj_join_right f1 f4 f5); split; [intuition |  intuition].  
  rewrite (disj_join_assoc); [assumption | intuition | intuition |intuition].
Qed.

Lemma subPfunction_extends : forall f1 f2 f, extendsPfunction f1 f2 -> 
	PfunctionExtendsFunction f f2 -> PfunctionExtendsFunction f f1.
Proof.
  unfold PfunctionExtendsFunction, extendsPfunction.
  intuition.
  assert (foo := H n).
  destruct (f1 n); [ idtac | trivial].
  assert (foo2 := foo b).
  assert (foo3 := H0 n).
  replace (f2 n) with (Some b) in foo3.
  assumption.
  intuition.
Qed.

Lemma Extends_refl : forall f, PfunctionExtendsFunction f (functionToPfunction f).
Proof.
  unfold PfunctionExtendsFunction, functionToPfunction, Pfunction.
  trivial.
Qed.

Lemma Extends_trans: forall pf1 f2 f3, PfunctionExtendsFunction f3 (functionToPfunction f2) ->
           PfunctionExtendsFunction f2 pf1 ->
           PfunctionExtendsFunction f3 pf1.
Proof. 
  unfold PfunctionExtendsFunction, functionToPfunction, Pfunction.
  intros.  
  hreplace H0 with (H0 n).
  rewrite H.
  assumption.
Qed.

Lemma sub_is_extendsfunction : forall s0 s, s << (functionToPfunction s0) <-> PfunctionExtendsFunction s0 s.
Proof.
  intros; rewrite sub_is_extends.
  unfold extendsPfunction, PfunctionExtendsFunction; intuition.
  hreplace H with (H n).
  destruct (s n); trivial.
  assert (foo : Some b = Some b); try reflexivity;  hreplace H with (H b foo).
  inversion H; trivial.
  hreplace H with (H x);  rewrite H0 in H ; simpl in H.
  rewrite <- H; unfold functionToPfunction; reflexivity.
Qed.

Lemma application : forall x (f g : A -> option B ) , f = g -> f x = g x.
Proof.
  intuition.
  rewrite H.
  trivial.
Qed.

Lemma Extends_eq : forall f g, PfunctionExtendsFunction f (functionToPfunction g) -> (f=g).
Proof.
  unfold PfunctionExtendsFunction, functionToPfunction, Pfunction.
  intros; rewrite <- Extensionality.
  intuition.
Qed.

Lemma subelement : forall s s1 s2 x n, s1 x = Some n ->
	 	  		functionToPfunction s = s1 ;; s2 ->
				s x = n.  
Proof.
  intros.
  unfold functionToPfunction, joinPfunction in H0; simpl in H0.
  apply (application x) in H0.
  rewrite H in H0.
  inversion H0.
  trivial.
Qed.

Lemma subelement_partial : forall s s1 s2 x n, s1 x = Some n ->
	 				       s = s1 ;; s2 ->
					       s x = Some n.  
Proof.
  intros.
  unfold joinPfunction in H0; simpl in H0.
  apply (application x) in H0.
  rewrite H in H0.
  trivial.
Qed.

Lemma notdom : forall f n, ~(domPfunction f n) <-> f n = None.
Proof.
  unfold domPfunction;intros.
  case (f n).
  intros;  split.
  intros.
  destruct H.
  exists b; trivial.
  intros;  intro.
  rewrite H in H0;   destruct H0;   inversion H0.
  split; intros.
  trivial.
  intro.
  destruct H0.
  inversion H0.
Qed.

Lemma disj_diff : forall s1 s2, s2 <#> (Pfunctiondiff s1 s2).
Proof.
  unfold disjointPfunction, Pfunctiondiff, domPfunction ; simpl ; intros.
  intro.
  destruct H as ([y1 H1] , [y2 H2]).
  rewrite H1  in H2.
  inversion H2.
Qed.


Lemma join_diff : forall s1 s2,  extendsPfunction s1 s2 -> s2 = s1 ;; (Pfunctiondiff s2 s1).
Proof.
  intros until s2.
  unfold joinPfunction, Pfunctiondiff, domPfunction ; simpl.
  unfold extendsPfunction.
  intro.
  rewrite <- (Extensionality A (option B) s2).
  intro.
  hreplace H with (H a).  
  destruct (s1 a); [idtac | trivial].
  intros.
  apply H.
  trivial.
Qed.

Lemma join_extends : forall s1 s2 s3,  s2 = s1 ;; s3 -> s1 << s2.
Proof.
  intros until s3.
  rewrite sub_is_extends.
  unfold joinPfunction, extendsPfunction.
  intros.
  rewrite <- (Extensionality A (option B) s2) in H.
  hreplace H with (H x).
  rewrite H0 in H.
  assumption.
Qed.

End Partial_Function.

Hint Resolve disjointPfunction_comm disjointPfunction_unit join_unit : partial_function.


