(* perping.v

   Define the orthogonality relational between relations on states
   and relations on code pointers.
   We prove at the end that this induces an adjunction between
   stateRel and natRel.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)


Require Import List.
Require Export state_relations.
Require Export cloud_relations.
Require Export nat_relations.
Require Import Basics.

(* Interpretation of a statePred *)

Definition Interp (R : stateRel) s s' p p' k : Prop := R (functionToPfunction s) (functionToPfunction s') p p' k.

Add Morphism Interp with signature 
  stateRelEq ==> (@eq state) ==> (@eq state) ==> (@eq program) ==> (@eq program) ==> (@eq nat) ==> iff
as Interpcompat.
  intros; subst; unfold Interp.
  rewrite H; intuition.
Qed.

Add Morphism Interp with signature 
  stateRelLeq ==> (@eq state) ==> (@eq state) ==> (@eq program) ==> (@eq program) ==> (@eq nat) ==> Basics.impl
as Interpcovariant.
  intros; subst; unfold Interp,impl.
  apply H.
Qed.

Add Morphism Interp with signature 
  inverse stateRelLeq ==> (@eq state) ==> (@eq state) ==> (@eq program) ==> (@eq program) ==> (@eq nat) ==> inverse impl
as Interpcovariant2.
  intros; subst; unfold Interp,impl; intro.
  apply H; assumption.
Qed.

Lemma Interp_leq: forall R1 R2 s s' p p' k, stateRelLeq R1 R2 -> Interp R1 s s' p p' k -> Interp R2 s s' p p' k.
Proof.
  unfold stateRelLeq, Interp.
  intuition.
Qed.

(*   Defintion of Perp  *)

Definition perpOf (S:stateRel) := fun p p' l l' k => forall s s' j,
                       j <= k ->
                       Interp S s s' p p' j ->
                       ((kstepterm j p s l -> terminates p' s' l') /\ 
                        (kstepterm j p' s' l' -> terminates p s l)).

Add Morphism perpOf with signature 
            stateRelEq ==> (@eq program) ==> (@eq program) ==> (@eq nat) ==>(@eq nat) ==> (@eq nat) ==> iff
as perpofcompat.
Proof.
  intros R R' H p p' l l' k'; subst;  unfold perpOf,natRelEq,stateRelEq,Interp; intros; subst; simpl; intuition;
  hreplace H with (H (functionToPfunction s) (functionToPfunction s') p p' j);
  destruct H as (H,H');
  try hreplace H0 with (H0 s s' j H1 (H' H2));
  try hreplace H0 with (H0 s s' j H1 (H H2)); 
  intuition.
Qed.

Definition Perp (S : stateRel ) : natRel.
Proof.
  intros; refine (mkNatRel (perpOf S) _).
  unfold perpOf; intros.
  assert (H3 : j <= k1); [ omega | idtac].
  hreplace H with (H s s' j H3 H2); intuition.
Defined.

Add Morphism Perp with signature stateRelEq ==> natRelEq
as perpcompat.
Proof.
  unfold Perp,perpOf,natRelEq,stateRelEq; intros; subst; simpl; intuition;
  hreplace H with (H (functionToPfunction s) (functionToPfunction s') p p' j); 
  destruct H as (H,H');
  try hreplace H0 with (H0 s s' j H1 (H' H2)); 
  try hreplace H0 with (H0 s s' j H1 (H H2)); 
  intuition.
Qed.

Add Morphism Perp with signature stateRelLeq --> natRelLeq
as perpContravariant.
Proof.
  unfold Perp,perpOf,natRelLeq,stateRelLeq; intros; subst; simpl; intuition;
  hreplace H0 with (H0 s s' j H1 (H (functionToPfunction s) (functionToPfunction s') p p' j H2)); intuition.
Qed.

Definition stateProg_lift (p1 p2 : program_fragment) : stateRel.
  intros p1 p2.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => 
    (PfunctionExtendsFunction p p1 /\ PfunctionExtendsFunction p' p2)) _).
  intuition.
Defined.

Add Morphism stateProg_lift with signature (@eq program_fragment) ==> (@eq program_fragment) ==> stateRelEq
as stateprogliftcompat.
Proof.
  intros; subst; auto.
Qed.

Add Morphism stateProg_lift with signature (@eq program_fragment) ==> (@eq program_fragment) ==> stateRelLeq
as stateprogliftcompat2.
Proof.
  intros; subst; auto.
Qed. 

Notation "'intern' ( p1 , p2 )" := (stateProg_lift (functionToPfunction p1) 
  (functionToPfunction p2)) (at level 75, p1 at next level, p2 at next level).

(* the same for cloudRel *)

(* Interpretation of a stateRel *)

Definition Interp_c (R : cloudRel) s s' p p' k : Prop := R (functionToPfunction s) (functionToPfunction s') p p' k.

Add Morphism Interp_c with signature
  cloudRelEq ==> (@eq state) ==> (@eq state) ==> (@eq program) ==> (@eq program) ==> (@eq nat) ==> iff
as Interpcompat_c.
  unfold Interp_c;  intros; subst.
  rewrite <- H; tauto.
Qed.

Lemma Interp_leq_c: forall R1 R2 s s' p p' k, cloudRelLeq R1 R2 -> Interp_c R1 s s' p p' k -> Interp_c R2 s s' p p' k.
Proof.
  unfold cloudRelLeq, Interp_c.
  intuition.  
Qed.

(*   Defintion of Perp  *)

Definition perpOf_c (S:cloudRel) := fun p p' l l' k => forall s s' j,
                       j <= k ->
                       Interp_c S s s' p p' j ->
                       ((kstepterm j p s l -> terminates p' s' l') /\ 
                        (kstepterm j p' s' l' -> terminates p s l)).

Add Morphism perpOf_c with signature 
            cloudRelEq ==> (@eq program) ==> (@eq program) ==> (@eq nat) ==> (@eq nat) ==> (@eq nat) ==> iff
as perpofcompat_c.
Proof.
  intros R R' H p p' l l' k'; subst;  unfold perpOf_c,natRelEq,cloudRelEq,Interp; intros; subst; simpl; intuition;
  hreplace H with (H (functionToPfunction s) (functionToPfunction s') p p' j);
  destruct H as (H,H');
  try hreplace H0 with (H0 s s' j H1 (H H2));
  try hreplace H0 with (H0 s s' j H1 (H' H2));
  intuition.
Qed.

Definition Perp_c (S : cloudRel ) : natRel.
Proof.
  intros.
  refine (mkNatRel (perpOf_c S) _).
  unfold perpOf_c; intros.
  assert (H3 : j <= k1);[ omega | idtac].
  hreplace H with (H s s' j H3 H2); intuition.
Defined.

Add Morphism Perp_c with signature cloudRelEq ==> natRelEq
as perpcompat_c.
Proof.
  unfold Perp_c,perpOf_c,natRelEq,cloudRelEq; simpl; intuition;
  hreplace H with (H (functionToPfunction s) (functionToPfunction s') p p' j); 
  destruct H as (H,H');
  try hreplace H0 with (H0 s s' j H1 (H H2)); 
  try hreplace H0 with (H0 s s' j H1 (H' H2)); 
  intuition.
Qed.

Add Morphism Perp_c with signature cloudRelLeq --> natRelLeq
as perpContravariant_c.
Proof.
  unfold Perp_c,perpOf_c,natRelLeq,cloudRelLeq; simpl; intuition;
  hreplace H0 with (H0 s s' j H1 (H (functionToPfunction s) (functionToPfunction s') p p' j H2)); intuition.
Qed.

Definition perpOf_nat (P:natRel) := fun ps ps' p p' k => forall l l' j,                  
                       j <= k ->
                       P p p' l l' j ->
                       forall s s', (PfunctionExtendsFunction s ps) -> 
                       (PfunctionExtendsFunction s' ps') -> 
                       ((kstepterm j p s l -> terminates p' s' l') /\ 
                        (kstepterm j p' s' l' -> terminates p s l)).

Add Morphism perpOf_nat with signature 
            natRelEq ==> (@eq (Pfunction Astate Bstate)) ==> (@eq (Pfunction Astate Bstate)) ==> 
            (@eq program) ==> (@eq program) ==> (@eq nat) ==> iff
as perpofcompat_nat.
Proof.
  unfold perpOf_nat; intros; subst; intuition.
  rewrite <- H in H2;
  hreplace H0 with (H0 l l' j H1 H2 s s'); intuition.
  rewrite <- H in H2;
  hreplace H0 with (H0 l l' j H1 H2 s s'); intuition.
  rewrite H in H2;
  hreplace H0 with (H0 l l' j H1 H2 s s'); intuition.
  rewrite H in H2; 
  hreplace H0 with (H0 l l' j H1 H2 s s'); intuition.
Qed.

Definition Perp_nat (P : natRel ) : stateRel.
Proof.
  intros.
  refine (mkStateRel (perpOf_nat P) _).
  unfold perpOf_nat; intros; assert (Hk1 : j <= k1); try omega.
  assert (Hs1 : PfunctionExtendsFunction s s1).
  rewrite sub_is_extends in H0; eapply subPfunction_extends; try eassumption.
  assert (Hs2 : PfunctionExtendsFunction s' s2).
  rewrite sub_is_extends in H1; eapply subPfunction_extends; try eassumption.
  hreplace H with (H l l' j Hk1 H4 s s' Hs1 Hs2);  intuition.
Defined.

Add Morphism Perp_nat with signature natRelEq ==> stateRelEq
as perpcompat_nat.
Proof.
  unfold Perp_nat, perpOf_nat,stateRelEq,natRelEq; simpl; intuition;
  hreplace H with (H p p' l l' j); destruct H as (H,H');
  try hreplace H0 with (H0 l l' j H1 (H H2) s0 s'0);
  try hreplace H0 with (H0 l l' j H1 (H' H2) s0 s'0); intuition.
Qed.

Add Morphism Perp_nat with signature natRelLeq --> stateRelLeq
as perpContravariant_nat.
Proof.
  unfold Perp_nat,perpOf_nat,natRelLeq,stateRelLeq; simpl; intuition;
  hreplace H0 with (H0 l l' j H1 (H p p' l l' j H2) s0 s'0); intuition.
Qed.

Definition stateClosure (P:stateRel) := Perp_nat (Perp P).

Definition natClosure (P:natRel) := Perp (Perp_nat P).

Lemma Adjunction : forall A B, stateRelLeq B (Perp_nat A) <-> natRelLeq A (Perp B).
Proof.
  unfold natRelLeq, stateRelLeq, Perp, perpOf, Perp_nat, perpOf_nat; simpl; intros.
  split; intros.
  assert (Hj : j<=j); try omega.
  assert( HA :  A p p' l l' j); try (eapply natRel_cond; try eassumption; omega).
  assert (Hrefl := Extends_refl s).
  assert (Hrefl' := Extends_refl s').
  apply (H (functionToPfunction s) (functionToPfunction s') p p' j H2 l l' j Hj HA s s' Hrefl Hrefl').
  assert (Hj : j<=j); try omega.
  assert( HB :  Interp B s0 s'0 p p' j).
  unfold Interp; eapply stateRel_cond.
  eassumption.
  rewrite sub_is_extendsfunction; assumption.
  rewrite sub_is_extendsfunction; assumption.
  omega.
  apply (H p p' l l' j H2 s0 s'0 j Hj HB).
Qed.

Lemma closure_unit : forall P, stateRelLeq P (stateClosure P).
Proof.
  intros; unfold stateClosure; rewrite Adjunction; reflexivity.
Qed.
  
Lemma closure_counit : forall P, natRelLeq P (natClosure P).
Proof.
  intros; unfold natClosure; rewrite <- Adjunction; reflexivity.
Qed.

Lemma closure_multi : forall P, stateRelLeq (stateClosure (stateClosure P)) (stateClosure P).
Proof.
  intros; subst; unfold stateClosure.
  apply perpContravariant_nat.
  unfold flip.
  rewrite <- Adjunction; reflexivity.
Qed.

Lemma closure_comulti : forall P, natRelLeq (natClosure (natClosure P)) (natClosure P).
Proof.
  intros; unfold natClosure.
  apply perpContravariant.
  unfold flip; rewrite Adjunction; reflexivity.
Qed.


(************)
(* ExistP  *)
(************)

Definition natExists (X : Type) (h : X -> natRel ) : natRel.
  intros.
  refine (mkNatRel (fun p => fun p' => fun l => fun l' => fun k => exists a: X, h a p p' l l' k) _).
  intros.
  witness a in H.
  exists a.
  apply (natRel_cond (h a) p p' l l' k1 k2); assumption.
Defined.

Implicit Arguments natExists [X].

Notation "'Exists_nat' h" := (natExists  h ) (at level 75, right associativity).

Notation "'Ex_nat' x , P" := (Exists_nat (fun x=> P) ) (at level 75, right associativity).

Add Morphism (@natExists nat) with signature (@eq (nat -> natRel)) ==> natRelEq  
as existsRelnat.
Proof.
  intros; subst; auto.
Qed.


(* Properties of existential*)

Lemma forallCompat : forall (X:Type) (h : X -> stateRel), 
stateRelLeq (stateClosure (Forall n, h n)) (Forall n, stateClosure (h n)).
Proof.
  unfold stateRelLeq, natClosure, Perp_nat; simpl;
  unfold Perp,perpOf_nat,  perpOf; simpl; intros.
  apply H; trivial.
  intros.
  assert (Hj0 : j0 <= k) ; try omega.
  apply H1; try assumption.
  apply (H5 a).
Qed.

Lemma existsCompat : forall (X:Type) (h : X -> natRel), 
natRelLeq (Ex_nat n, natClosure (h n)) (natClosure (Ex_nat n, h n)).
Proof.
  unfold natRelLeq, natClosure, Perp, perpOf, Perp_nat,perpOf_nat; simpl; intros.
  destruct H.
  apply H; trivial.
  intros.
  assert (Hj0 : j0 <= k) ; try omega.
  apply H1; try assumption.
  exists x;  apply H3.
Qed.

Lemma existsCompat2 : forall (X:Type) (h : X -> stateRel), 
stateRelLeq (Ex n, stateClosure (h n)) (stateClosure (Ex n, h n)).
Proof.
  unfold stateRelLeq, stateClosure, Perp, perpOf, Perp_nat,perpOf_nat; simpl; intros.
  destruct H.
  apply H; trivial.
  intros.
  assert (Hj0 : j0 <= k) ; try omega.
  apply H1; try assumption.
  exists x;  apply H5.
Qed.

Lemma uncoverExists : forall (X:Type) (h : X -> stateRel), 
natRelEq (Perp(Forall n, stateClosure (h n))) (natClosure (Ex_nat n, Perp (h n))).
Proof.
  unfold natRelEq, natClosure, Perp, perpOf, Perp_nat,perpOf_nat; simpl; 
  unfold Perp, perpOf, Perp_nat,perpOf_nat; simpl; intros.
  split ; intros.
  apply H; try assumption.
  intros.
  apply H1; try assumption.
  exists a; apply H3.
  apply H; try assumption.
  intros.
  destruct H3.
  apply (H1 x); trivial.
Qed.

Lemma forallClosed : forall (X:Type) (h : X -> stateRel), 
stateRelLeq (stateClosure (Forall n, stateClosure (h n))) (Forall n, stateClosure (h n)).
 Proof.
  unfold stateRelLeq, natClosure, Perp, perpOf, Perp_nat,perpOf_nat; simpl; 
  unfold Perp, perpOf, Perp_nat,perpOf_nat; simpl; intros.
  apply H; trivial.
  unfold Perp, perpOf, Perp_nat,perpOf_nat; simpl; intros.
  assert (Hj0 : j0 <= k) ; try omega.
  apply (H5 a); try assumption; trivial.
  intros.
  apply H1.
  omega.
  assumption.
  apply Extends_refl.
  apply Extends_refl.
Qed.

Lemma instanciateExists : forall (X:Type) (h : X -> stateRel) n0, 
natRelLeq (Perp(Ex n, h n)) (Perp (h n0)).
Proof.
  unfold natRelLeq, natClosure, Perp, perpOf, Perp_nat,perpOf_nat; simpl; 
  unfold Perp, perpOf, Perp_nat,perpOf_nat; simpl; intros.
  apply H; try assumption.
  intros.
  exists n0.
  apply H1.
Qed.


Lemma conjExtract : forall P Q, stateRelLeq (stateClosure (cloudToState (P //\\ Q)))
(cloudToState ( stateToCloud(stateClosure (cloudToState P)) //\\ stateToCloud(stateClosure (cloudToState Q)))).
Proof.
  unfold stateRelLeq, natClosure, Perp, perpOf, Perp_nat,perpOf_nat; simpl; 
  unfold Perp, perpOf, Perp_nat,perpOf_nat; simpl; intros.
  split;intros.
  apply H; try assumption.
  intros;  apply H1; try assumption.
  destruct H5; assumption.
  apply H; try assumption.
  intros.
  intros;  apply H1; try assumption.
  destruct H5; assumption.
Qed.
