(* pointers.v

   Define the basic block of our relational reasoning,
   namely, given a natRel R
   
   ({m,m'} |-> R) s s'   <=>   R s(m) s'(m')
   
   Define the structure of a Block and all the unfolding lemmas.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)


Require Import List.
Require Export perping.


(*   Definition of n |=> P:natPred    *)

Definition ptsto (n n':nat) (P:natRel) : stateRel.
  intros n n' P.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => exists l, exists l',
    s n = Some l /\ s' n' = Some l' /\ P p p' l l' k) _).
  intros.
  destruct H as [l [l' (H,(H',Hp))]].
  exists l; exists l'.
  intuition.
  rewrite sub_is_extends in H0.
  rewrite (H0 n l); [trivial | assumption].
  rewrite sub_is_extends in H1.
  rewrite (H1 n' l'); [trivial | assumption].
  apply (natRel_cond P p1 p2 l l' k1 k2); assumption.
Defined.

Notation "{ x1 , x2 } |=> S" := (ptsto x1 x2 S) (at level 75, x2 at next level).

Notation "x |=> S" := (ptsto x x S) (at level 75).


Add Morphism ptsto with signature (@eq nat) ==> (@eq nat) ==> natRelEq ==> stateRelEq
as ptstocompat.
Proof.
  unfold ptsto,stateRelEq;intros ; simpl; firstorder.
  exists x0; exists x1; intuition; rewrite <- H; intuition.
  exists x0; exists x1; intuition; rewrite H; intuition.
Qed.

Add Morphism ptsto with signature (@eq nat) ==> (@eq nat) ==> natRelLeq ++> stateRelLeq
as ptstoCovariant.
Proof.
  unfold ptsto,stateRelLeq; simpl ; intros; auto.
  firstorder.
Qed.

(* ptsto *)

Notation "{ n1 , n2 } |-> { m1 , m2 }" := ({n1 , n2} |=> natPconstantRel (fun l => fun l' => l=m1 /\ l'=m2 )) 
  (at level 75, n2 at next level, m1 at next level).

Notation "n |-> { m , p }" := ({n , n} |-> {m , p}) (at level 75, m at next level).

Notation "n |-> m" := (n |-> {m , m}) (at level 75).

Notation "{ n , m } |-> - " := ({n , m} |=> natPtrue) (at level 75, m at next level).

Notation "n |-> -" := ({n ,n} |-> -) (at level 75).

Notation "{ n , m } |-> ? " := (Exists (fun x=> Exists (fun y=> ({n,m} |-> {x,y})))) 
  (at level 75, m at next level).

Notation "n |-> ?" := ({n,n} |-> ?) (at level 75).


(* the same for cloudRel *)

(*   Definition of n |=> P:natRel    *)

Definition ptsto_c (n n':nat) (P:natRel) : cloudRel.
  intros n n' P.
  refine (mkCloudRel (fun s => fun s' => fun p => fun p' => fun k => exists l, exists l',
    s n = Some l /\ s' n' = Some l' /\ P p p' l l' k) _).
  intros.
  destruct H as [l [l' (H,(H',Hp))]].
  exists l; exists l'.
  intuition.
  rewrite sub_is_extends in H0.
  rewrite (H0 n l); [trivial | assumption].
  rewrite sub_is_extends in H1.
  rewrite (H1 n' l'); [trivial | assumption].
  apply (natRel_cond P p1 p2 l l' k1 k2); assumption.
Defined.

Notation "{ x1 , x2 } ||=> S" := (ptsto_c x1 x2 S) (at level 75, x2 at next level).

Notation "x ||=> S" := ({x , x} ||=> S) (at level 75).

Add Morphism ptsto_c with signature (@eq nat) ==> (@eq nat) ==> natRelEq ==> cloudRelEq
as ptstocompat_c.
Proof.
  unfold ptsto_c,cloudRelEq; simpl; intros; subst;firstorder.
  exists x0; exists x1; intuition.
  rewrite <- H; intuition.
  exists x0; exists x1; intuition.
  rewrite H; intuition.
Qed.

Add Morphism ptsto_c with signature (@eq nat) ==> (@eq nat) ==> natRelLeq ++> cloudRelLeq
as ptstoCovariant_c.
Proof.
  unfold ptsto_c,cloudRelLeq; simpl; intros; subst;firstorder.
Qed.

(* ptsto *)

Notation "{ n1 , n2 } ||-> { m1 , m2 }" := ({n1 , n2} ||=> natPconstantRel (fun l => fun l' => l=m1 /\ l'=m2 )) 
  (at level 75, n2 at next level, m1 at next level).

Notation "n ||-> { m , p }" := ({n ,n} ||-> {m,p}) (at level 75, m at next level).

Notation "n ||-> m" := (n ||-> {m,m}) (at level 75).

Notation "{ n , m } ||-> - " := ({n,m} ||=> natPtrue) (at level 75, m at next level).

Notation "n ||-> -" :=  ({n,n} ||-> -) (at level 75).

Notation "{ n , m } ||-> ? " := (Exists_c (fun x=> Exists_c (fun y=> ({n,m} ||-> {x,y})))) 
  (at level 75, m at next level).

Notation "n ||-> ?" := ({n,n} ||-> ?) (at level 75).

Lemma stateCloudPtsot : forall n1 n2 P, cloudRelEq (stateToCloud ({n1,n2}|=>P)) ({n1,n2}||=>P).
Proof.
  unfold cloudRelEq, ptsto, ptsto_c; simpl;intuition.
Qed.

Lemma cloudStatePtsot : forall n1 n2 P, stateRelEq (cloudToState ({n1,n2}||=>P)) ({n1,n2}|=>P).
Proof.
  unfold stateRelEq, ptsto, ptsto_c; simpl;intuition.
Qed.

(* Lemmas on blank pointers*)

Lemma ptsto_blank_leq_nat : forall n1 n2 m1 m2 , stateRelLeq ({n1,n2}|->{m1,m2}) ({n1,n2}|->?).
Proof.
  unfold ptsto, stateRelLeq; simpl; firstorder.
  subst; exists m1; exists m2; exists m1; exists m2; intuition.
Qed.

Lemma ptsto_blank_leq_nat_c : forall n1 n2 m1 m2 , cloudRelLeq ({n1,n2}||->{m1,m2}) ({n1,n2}||->?).
Proof.
  unfold ptsto_c, cloudRelLeq; simpl; firstorder.
  subst; exists m1; exists m2; exists m1; exists m2; intuition.
Qed.

Lemma ptsto_blank_leq : forall n1 n2 P, stateRelLeq ({n1,n2}|=>P) ({n1,n2}|->-).
Proof.
  unfold ptsto, stateRelLeq; simpl; firstorder.
Qed.

Lemma ptsto_blank_leq_c : forall n1 n2 P, cloudRelLeq ({n1,n2}||=>P) ({n1,n2}||-> -).
Proof.
  unfold ptsto_c, cloudRelLeq;  simpl; firstorder.
Qed.

Lemma ptsto_blank_elim : forall n1 n2 P, cloudRelEq ({n1,n2}||=>P) ({n1,n2}||=>P //\\ {n1,n2}||->-).
Proof.
  unfold ptsto_c, cloudRelEq;  simpl; firstorder.
Qed.

Lemma ptsto_blank_elim2 : forall n1 n2 m1 m2, cloudRelEq ({n1,n2}||->{m1,m2}) ({n1,n2}||->{m1,m2} //\\ {n1,n2}||->?).
Proof.
  unfold ptsto_c, cloudRelEq;  simpl; firstorder.
  subst ; exists m1; exists  m2; exists m1; exists m2; intuition.
Qed.

(* Dealing with part of the stack *)

Definition left_ptsto (n:nat) (P : nat->Prop)  : stateRel.
  intros n P.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => exists l, s n = Some l /\ P l) _).
  intros; destruct H as [l (H,Hp)].
  exists l; intuition.
  rewrite sub_is_extends in H0.
  rewrite (H0 n l); [trivial | assumption].
Defined.

Definition right_ptsto (n:nat) (P : nat->Prop)  : stateRel.
  intros n P.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => exists l, s' n = Some l /\ P l) _).
  intros; destruct H as [l (H,Hp)].
  exists l; intuition.
  rewrite sub_is_extends in H1.
  rewrite (H1 n l); [trivial | assumption].
Defined.

Notation "{ n , - } |-> m" := (left_ptsto n (fun x => x = m )) (at level 75).

Notation "{ - , n } |-> m" := (right_ptsto n (fun x => x = m )) (at level 75, n at next level).

Notation "{ n , - } |-> -" := (left_ptsto n (fun x => True)) (at level 75).

Notation "{ - , n } |-> -" := (right_ptsto n (fun x => True)) (at level 75, n at next level).

(* Lemma split_pointer : forall m n, stateRelEq ({m,n} |-> -)  ({m,-} |-> - ## {-,n} |-> -). *)
(* Proof. *)
(*   unfold left_ptsto, right_ptsto, ptsto, stateRelEq; simpl. *)
(*   firstorder. *)
(*   exists  *)

(* Qed. *)

Fixpoint left_Block (base size : nat) : stateRel :=
  match size with 
    0 => statePtrue
    | S m => ({(base+m) , - } |-> -) ## left_Block base m
  end.

Fixpoint right_Block (base size : nat) : stateRel :=
  match size with 
    0 => statePtrue
    | S m => ({- , (base+m)} |-> -) ## right_Block base m
  end.

Fixpoint Block (base base' size size': nat) : stateRel :=
  match size with 
    0 => right_Block base' size'
    | S m => match size' with 
               0 => left_Block base size
               | S m' => (({(base+m) , (base'+m')} |-> -)) ## Block base base' m m'
             end
  end.

Definition Topfrom (m m': nat) : stateRel.
Proof.
  intros m m'.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => 
     (forall n, n >= m -> (({n,-} |-> -) s s' p p' k)) 
  /\ (forall n', n' >= m' -> (({-,n'} |-> -) s s' p p' k)))
  _).
  intuition.
  specialize (H3 n H).
  eapply stateRel_cond; [apply H3 | assumption | assumption | assumption].
  specialize (H4 n' H).
  eapply stateRel_cond; [apply H4 | assumption | assumption | assumption].
Defined.

(* rearranging the stack *)

Lemma right_block_unfold_end : forall m n, stateRelEq (right_Block m (S n))
                                         ( {-, m+n} |-> - ## (right_Block m n) ).
Proof.
  simpl; trivial.
Qed.

Lemma left_block_unfold_end : forall m n, stateRelEq (left_Block m (S n))
                                         ( {m+n,-} |-> - ## (left_Block m n) ).
Proof.
  simpl; trivial.
Qed.

Lemma block_unfold_end : forall m m' n n', stateRelEq (Block m m' (S n) (S n'))
                                         ( {m+n , m'+n'} |-> - ## (Block m m' n n') ).
Proof.
  simpl; trivial.
Qed.

Lemma ptsto_split : forall m n, stateRelEq ( {m,n} |-> - ) ( {m,-} |-> -  ##  {-,n} |-> -).
Proof.
  unfold stateRelEq; simpl; intuition.
  destruct H as [l [l' (Hl,(hl',Hdummy))]].
  exists s; exists (nullPfunction Astate Bstate); exists (nullPfunction Astate Bstate) ; exists s';
  intuition.
  rewrite disjointPfunction_comm; intuition.
  rewrite join_disj_comm; intuition.
  rewrite disjointPfunction_comm; intuition.
  exists l; intuition.
  exists l'; intuition.
  firstorder.
  exists x3; exists x4; intuition.
  eapply subelement_partial ; try eassumption; intuition.
  rewrite join_disj_comm in H2; intuition.
  eapply subelement_partial ; try eassumption; intuition.
Qed.


Lemma right_block_unfold_start : forall m n , stateRelEq (right_Block m (S n)) 
                                          ( {- , m} |-> - ## (right_Block (S m) n )).
Proof.
  induction n.
  simpl; replace (m+0) with m by omega; reflexivity.
  simpl; rewrite IHn.
  omega_replace (m+S n) with (S(m+n)) ; ring.
Qed.

Lemma left_block_unfold_start : forall m n , stateRelEq (left_Block m (S n)) 
                                          ( {m , -} |-> - ## (left_Block (S m) n )).
Proof.
  induction n.
  simpl; replace (m+0) with m by omega; reflexivity.
  simpl; rewrite IHn.
  omega_replace (m+S n) with (S(m+n)); ring.
Qed.

Lemma block_unfold_start : forall m m' n n', stateRelEq (Block m m' (S n) (S n')) 
                                          ( {m , m'} |-> - ## (Block (S m) (S m') n n')).
Proof.
  induction n.
  simpl;  replace (m+0) with m by omega.
  setoid_rewrite ptsto_split.
  intros; repeat rewrite <- stateRelTensor_assoc.
  rewrite <- right_block_unfold_start.
  simpl; repeat rewrite stateRelTensor_assoc; setoid_rewrite <- ptsto_split.
  reflexivity.
  induction n'.
  simpl;  replace (m'+0) with m' by omega.
  setoid_rewrite ptsto_split.
  rewrite (stateRelTensor_comm ({m , -} |-> -)); rewrite <- (stateRelTensor_assoc ({- , m'} |-> -)).
  rewrite (stateRelTensor_assoc ({m , -} |-> -));  rewrite (stateRelTensor_comm ({m , -} |-> -)).
  setoid_rewrite <- stateRelTensor_assoc;  rewrite <- left_block_unfold_start.
  simpl;omega_replace (m+ S n) with (S(m+n)); ring.
  simpl; setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm ({m, m'} |->-)).
  setoid_rewrite <- stateRelTensor_assoc; rewrite <- IHn.
  simpl; omega_replace (m+S n) with (S(m+n)); omega_replace (m'+S n') with (S(m'+n')); ring.
Qed.

Lemma right_block_weakening : forall m n p, n <= p -> stateRelLeq (right_Block m p) (right_Block m n).
Proof.
  induction p.
  intros; assert (e : n=0); try omega; subst; simpl; reflexivity.
  intro; compare n (S p).
  intro e; rewrite e; reflexivity.
  intro; assert (e : n<=p); try omega;  rewrite right_block_unfold_end; rewrite stateRelTensor_comm;rewrite stateRelTensor_weakening;  apply (IHp e).
Qed.

Lemma left_block_weakening : forall m n p, n <= p -> stateRelLeq (left_Block m p) (left_Block m n).
Proof.
  induction p.
  intros; assert (e : n=0); try omega; subst; simpl; reflexivity.
  intro; compare n (S p).
  intro e; rewrite e; reflexivity.
  intro; assert (e : n<=p); try omega;  rewrite left_block_unfold_end; rewrite stateRelTensor_comm;rewrite stateRelTensor_weakening;  apply (IHp e).
Qed.
  
Lemma Block_to_leftBlock : forall m n p, stateRelEq (left_Block m p) (Block m n p 0).
Proof.
  induction p; simpl; reflexivity.
Qed.

Lemma block_unfold_end_sided : forall m m' n n', 
  stateRelEq (Block m m' (S n) n') ( {m + n , -} |-> - ## (Block m m' n n') ) /\
  stateRelEq (Block m m' n (S n')) ( {- , m' + n'} |-> - ## (Block m m' n n') ).
Proof.
  double induction n n'; intuition; simpl.
  rewrite ptsto_split; ring.
  rewrite <- Block_to_leftBlock;  rewrite ptsto_split; ring.
  setoid_rewrite ptsto_split.
  setoid_rewrite (stateRelTensor_comm ({m + n1 , -} |-> -)).
  setoid_rewrite <- stateRelTensor_assoc.
  specialize (H0 n0); destruct H0 as (H0,H0').
  rewrite <- H0; reflexivity.
  setoid_rewrite ptsto_split.
  setoid_rewrite <- stateRelTensor_assoc.
  specialize (H0 n0); destruct H0 as (H0,H0').
  rewrite <- H0'.
  ring.
Qed.

Lemma block_unfold_end_left : forall m m' n n', 
  stateRelEq (Block m m' (S n) n') ( {m + n,-} |-> - ## (Block m m' n n') ).
Proof.
  intros; assert (H := block_unfold_end_sided  m m' n n') .
  intuition.
Qed.

Lemma block_unfold_end_right : forall m m' n n', 
  stateRelEq (Block m m' n (S n')) ( {- , m' + n'} |-> - ## (Block m m' n n') ).
Proof.
  intros; assert (H := block_unfold_end_sided  m m' n n') .
  intuition.
Qed.

Lemma block_weakening : forall p1 p2 m1 m2 n1 n2 , n1<=p1 -> n2<=p2 ->
                                   stateRelLeq (Block m1 m2 p1 p2) (Block m1 m2 n1 n2).
Proof.
  induction p1.
  intros; assert (e : n1=0); try omega; subst; simpl; apply right_block_weakening; assumption.
  induction p2.
  intros; assert (e : n2=0); try omega; subst; simpl.
  rewrite <- Block_to_leftBlock; rewrite <- left_block_unfold_end.
  apply left_block_weakening; assumption.
  intros; compare n1 (S p1); compare n2 (S p2); intros; subst.
  reflexivity.
  assert (e : n2<=p2); try omega; rewrite block_unfold_end.
  rewrite <- (IHp2 m1 m2 (S p1) n2); try omega; try assumption.
  rewrite stateRelTensor_comm.
  rewrite block_unfold_end_left.
  setoid_rewrite ptsto_split.
  setoid_rewrite stateRelTensor_assoc.
  rewrite (stateRelTensor_comm ({m1 + p1 , -} |-> -)).
  apply stateRelTensor_weakening.
  assert (e : n1<=p1); try omega; rewrite block_unfold_end.
  rewrite <- (IHp1 (S p2) m1 m2 n1 (S p2)); try omega; try assumption.
  rewrite block_unfold_end_right.
  setoid_rewrite ptsto_split.
  setoid_rewrite <- stateRelTensor_assoc.
  rewrite stateRelTensor_comm.
  apply stateRelTensor_weakening.
  simpl.
  rewrite stateRelTensor_comm;  rewrite stateRelTensor_weakening.
  eapply IHp1; omega.
Qed.

Lemma topfromBlock : forall m m', stateRelEq (Topfrom m m') ({m , m'} |-> -  ## Topfrom (S m) (S m')).
Proof.
  unfold stateRelEq; simpl; intuition.
  set (s1 := fun i => if beq_nat i m then s i else None).
  exists s1; exists (Pfunctiondiff s s1).
  set (s'1 := fun i => if beq_nat i m' then s' i else None).
  exists s'1; exists (Pfunctiondiff s' s'1).
  intuition; [apply disj_diff |apply join_diff; unfold s1; apply extends_eq |apply disj_diff 
    | apply join_diff; unfold s'1; apply extends_eq | idtac | idtac | idtac].
  assert (H0m := H0 m).
  assert (H1m' := H1 m').
  destruct H0m as [l (Hl,Htrue)]; try omega.
  destruct H1m' as [l' (Hl',Htrue')]; try omega.
  unfold s1, s'1.
  replace (beq_nat m m) with true; [idtac | apply beq_nat_refl].
  replace (beq_nat m' m') with true; [idtac | apply beq_nat_refl].
  exists l; exists l'; intuition; assumption.
  assert (H3 := le_trans m (S m) n (le_n_Sn m) H2).
  specialize (H0 n H3).
  destruct H0 as [l (Hl,Htrue)].
  unfold Pfunctiondiff,s1.
  replace (beq_nat n m) with false; [idtac | apply beq_nat_neq;omega].
  exists l; intuition; assumption.
  assert (H3 := le_trans m' (S m') n' (le_n_Sn m') H2).
  specialize (H1 n' H3).
  destruct H1 as (l'&Hl'&Htrue).
  unfold Pfunctiondiff,s'1.
  replace (beq_nat n' m') with false; [idtac | apply beq_nat_neq;omega].
  exists l'; intuition; assumption.
  destruct H as (f1 & f2 & f'1 & f'2 & Hdisj & Hjoin & Hdisj' & Hjoin' & foo & Hf2 & Hf2').
  destruct foo as (l & l' & Hf1 & Hf1' & htrue).  
  compare m n.
  intro; subst.
  unfold joinPfunction.
  rewrite Hf1; exists l; intuition;trivial.
  intro; assert (n >= S m); try omega.
  hreplace Hf2 with (Hf2 n H).
  destruct Hf2 as (l2 & Hf2 & Htrue).
  subst; unfold joinPfunction.
  hreplace Hdisj with (Hdisj n).
  assert (domPfunction f2 n <-> True).
  firstorder.
  rewrite Hf2.
  rewrite H1 in Hdisj.
  setoid_replace (domPfunction f1 n /\ True) with (domPfunction f1 n) using relation iff in Hdisj by intuition.
  exists l2.
  assert ( f1 n = None).
  rewrite <- (notdom f1 n).
  apply Hdisj.
  rewrite H2;intuition;trivial.
  destruct H as (f1 & f2 & f'1 & f'2 & Hdisj & Hjoin & Hdisj' & Hjoin' & foo & Hf2 & Hf2').
  destruct foo as (l & l' & Hf1 & Hf1' & htrue).  
  compare m' n'.
  intro; subst.
  unfold joinPfunction.
  rewrite Hf1'; exists l'; intuition;trivial.
  intro; assert (n' >= S m'); try omega.
  hreplace Hf2' with (Hf2' n' H).
  destruct Hf2' as (l2' & Hf2' & Htrue).
  subst; unfold joinPfunction.
  hreplace Hdisj' with (Hdisj' n').
  assert (domPfunction f'2 n' <-> True).
  firstorder.
  rewrite Hf2'.
  rewrite H1 in Hdisj'.
  setoid_replace (domPfunction f'1 n' /\ True) with (domPfunction f'1 n') using relation iff  in Hdisj' by intuition.
  exists l2'.
  assert (f'1 n' = None).
  rewrite <- (notdom f'1 n').
  apply Hdisj'.
  rewrite H2;intuition;trivial.
Qed.

Lemma existsBlank : forall n n', stateRelEq ({n,n'}|->-) (Ex x,Ex x', ({n,n'}|-> {x,x'})).
Proof.
  unfold stateRelEq, ptsto, stateExists; simpl.
  firstorder.
  exists x; exists x0;exists x; exists x0.
  auto.
Qed.


Lemma ptsto_later : forall m m' P, stateRelLeq ({m,m'} |=> Perp P) ({m,m'} |=> Later(Perp P)).
Proof.
  unfold stateRelLeq.
  intros until k; intro Hlater.
  unfold ptsto in *; simpl in *.
  destruct Hlater as (l & l' & Hs & Hs'& Hperp).
  exists l; exists l'; intuition.
  unfold perpOf in *; simpl in *.
  intros until j0; intros Hj0 HP.
  apply Hperp.
  omega.
  assumption.
Qed.

Lemma ptsto_later_c : forall m m' P, cloudRelLeq ({m,m'} ||=> Perp P) ({m,m'} ||=> Later(Perp P)).
Proof.
  unfold cloudRelLeq.
  intros until k; intro Hlater.
  unfold ptsto in *; simpl in *.
  destruct Hlater as (l & l' & Hs & Hs'& Hperp).
  exists l; exists l'; intuition.
  unfold perpOf in *; simpl in *.
  intros until j0; intros Hj0 HP.
  apply Hperp.
  omega.
  assumption.
Qed.
