(* programs.v

   Basic machine model.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research

   This is the revised nnn version with a concrete instruction 
   Programs represented as maps from positive numbers to instruction 
   (using the library FMapPositive)
   This enables to use (find i prog) to get the ith instruction of prog   
*)


Require Export states.

Require Export FMapPositive.

(* lvalues *)
Inductive Dest : Set :=
| d_imm : nat -> Dest          (* store immediate, e.g. [4] <- blah *)
| d_ind : nat -> Dest          (* indirect, e.g. [[7]] <- blah *)
| d_indo : nat -> nat -> Dest.  (* indirect + offset, e.g. [[4]+2] <- blah *)

(* rvalues *)
Inductive src : Set :=
  | s_cst : nat -> src           (* nat literal, e.g. d <- 7 *)
  | s_imm : nat -> src           (* contents of a location, e.g. d <- [7] *)
  | s_ind : nat -> src           (* contents of contents of location d <- [[7]] *)
  | s_indo : nat -> nat -> src.   (* contents of (contents of loc n + offset)
                                    e.g. d <- [[4]+2] offset is first arg *)

Inductive instruction : Set :=
  | i_halt : instruction
  | i_move : Dest -> src -> instruction
  | i_op : (nat -> nat -> nat) -> Dest -> src -> src -> instruction
  | i_branch : src -> instruction
  | i_brz : src -> src -> instruction
  | i_brnz : src -> src -> instruction
  | i_eqtst : src -> src -> src -> instruction.

(* Semantics of instructions *)

Definition sem_Dest (de:Dest) (s:state) : nat :=
match de with
 | d_imm n => n
 | d_ind n => s n
 | d_indo ofs n => s n + ofs
end.

Definition sem_src (sr:src) (s:state) :=
match sr with
 | s_cst n => n
 | s_imm n => s n
 | s_ind n => s (s n)
 | s_indo ofs n => s (s n + ofs)
end.

Definition sem_instr (ins:instruction) (s:state) (pc:nat) : option (state * nat) :=
match ins with
 | i_halt => None
 | i_move de sr => Some (update s (sem_Dest de s) (sem_src sr s), S pc)
 | i_op op de sr1 sr2 => Some (update s (sem_Dest de s) (op (sem_src sr1 s) (sem_src sr2 s)), S pc)
 | i_branch sr => Some (s, sem_src sr s)
 | i_brz srscrut srtarg => 
     Some (s, match sem_src srscrut s with 0 => sem_src srtarg s | S _ => S pc end)
 | i_brnz srscrut srtarg => 
     Some (s, match sem_src srscrut s with 0 => S pc | S _ => sem_src srtarg s end)
 | i_eqtst sr1 sr2 srtarg => Some (s, if beq_nat (sem_src sr1 s) (sem_src sr2 s) then sem_src srtarg s else S pc)
end.

(* Programs and Program Fragments *)

Definition program : Set := nat -> instruction.

Definition program_fragment := Pfunction nat instruction.

Definition joinfragment (p1 p2: program_fragment) := joinPfunction p1 p2.

Definition disjfragment (p1 p2: program_fragment) := disjointPfunction p1 p2.

Definition subfragment (p1 p2: program_fragment) := subPfunction p1 p2.

Notation "p1 ';;' p2" := (joinfragment p1 p2) (at level 65, right associativity).

Notation "p1 '<#>' p2" := (disjfragment p1 p2) (at level 65, right associativity).

Notation "p1 '<<' p2" := (subfragment p1 p2) (at level 65, right associativity).

Lemma subfrag_refl : forall p:program_fragment, p << p.
Proof.
  intros; unfold subfragment.
  apply subPfunction_refl.
Qed.

Lemma subfrag_trans : forall (p1 p2 p3:program_fragment), 
                    p1 << p2 -> p2 << p3 -> p1 << p3.
Proof.
  unfold subfragment; intros until p3.
  apply subPfunction_trans.
Qed.

Add Relation program_fragment (subfragment)
   reflexivity proved by subfrag_refl
   transitivity proved by subfrag_trans
 as program_fragment_Rel.

Lemma extends_join : forall p1 p2, p1 <#> p2 -> p1 << p1 ;; p2.
Proof.
  intros; unfold subfragment, subPfunction.
  exists p2.
  auto.
Qed.

Hint Rewrite disj_join_left : prog_frag.
Hint Rewrite disj_join_right : prog_frag.

Ltac unfold_frag :=  unfold disjfragment, joinfragment, subfragment.

Tactic Notation "resolve_frag" := let rec foo _ := 
                     unfold_frag; try first [ tauto | intuition; rewrite disjointPfunction_comm ; tauto |
                                            intuition ; rewrite disj_join_left ; foo 0 |
                                            intuition ; rewrite disj_join_right ; foo 0] in foo 0.

Ltac normal_form_frag := try repeat (rewrite disj_join_assoc; resolve_frag).

(* Actual behaviour of programs *)

Definition halting_extension (pf : program_fragment) : program :=
 fun l => match pf l with Some i => i | None => i_halt end.

Definition diverging_extension (pf : program_fragment) : program :=
 fun l => match pf l with Some i => i | None => i_branch (s_cst l) end.

(* This is "terminates in fewer than k steps" *)

Fixpoint kstepsafe (k:nat) (p:program) (s:state) (l:nat) {struct k} : Prop :=
match k with
 | O => True
 | (S j) => match sem_instr (p l) s l with
             | None => False
             | Some (s', l') => kstepsafe j p s' l'
            end
end.

Fixpoint kstepterm (k:nat) (p:program) (s:state) (l:nat) {struct k} : Prop :=
match k with
 | O => False
 | (S j) => match sem_instr (p l) s l with
             | None => True
             | Some (s', l') => kstepterm j p s' l'
            end
end.

Fixpoint kstepreduce (k:nat) (p:program) (s s':state) (l l':nat) 
{struct k} : Prop :=
match k with
 | O => (s = s') /\ (l = l')
 | (S j) => match sem_instr (p l) s l with
             | None => False
             | Some (s'', l'') => kstepreduce j p s'' s' l'' l'
            end
end.
 
Definition terminates p s l := exists k, kstepterm k p s l.


Lemma terminate_reduce : forall p s s' l l' k,
  terminates p s' l' -> kstepreduce k p s s' l l' -> terminates p s l.
Proof.
  unfold terminates; intros.
  destruct H as (k',H).
  exists (k+k').
  generalize dependent l; generalize dependent  s; induction k.
  intros; simpl in *; destruct H0; subst; assumption.
  simpl in *; intros; destruct (sem_instr (p l) s l); trivial;  destruct p0.
  apply (IHk s0 n H0).
Qed.

Lemma kstep_induction : forall j k p s l, k <= j -> kstepterm k p s l -> kstepterm j p s l.
Proof.
  induction j.
  intros; assert (k=0); try omega; subst; simpl; inversion H0.
  destruct k.
  intros ; inversion H0.
  unfold kstepterm; simpl; fold kstepterm; intros.
  destruct (sem_instr (p l) s l); trivial.
  destruct p0; simpl.
  eapply IHj; try eassumption.
  omega.
Qed.

Lemma stepterm_reduce : forall k1 k k2 p s s' l l',
  k1 + k2 = k -> kstepreduce k1 p s s' l l' -> 
  ( kstepterm k2 p s' l' <-> kstepterm k p s l).
Proof.
  induction k1.
  simpl; intuition; destruct H1; subst;
  eapply kstep_induction; try eassumption; omega.
  intuition.
  destruct k.
  assert (k2 = 0) ; try omega.
  subst ; inversion H1.
  unfold kstepterm; fold kstepterm;
  unfold kstepreduce in H0; fold kstepreduce in H0;
  destruct (sem_instr (p l) s l); trivial.
  destruct p0; simpl;
  specialize (IHk1 k k2 p s0 s' n l').
  destruct IHk1 as (Hs0,Hs'); try assumption.
  omega.
  apply (Hs0 H1).
  destruct k.
  assert (k2 = 0) ; try omega.
  subst ; inversion H1.
  unfold kstepterm in H1; fold kstepterm in H1;
  unfold kstepreduce in H0; fold kstepreduce in H0.
  destruct (sem_instr (p l) s l); try inversion H0.
  destruct p0; simpl;
  specialize (IHk1 k k2 p s0 s' n l').
  destruct IHk1 as (Hs0,Hs'); try assumption.
  omega.
  apply (Hs' H1).  
Qed.

Lemma stepterm_reduce2 : forall k k1 p s s' l l',
  kstepreduce k1 p s s' l l' -> kstepterm k p s l -> k1 <= k.
Proof.
  induction k.
  intuition; destruct H0.
  destruct k1; try solve [intuition].
  intros until l'; simpl in *.
  case (sem_instr (p l) s l); intuition.
  rewrite (surjective_pairing p0) in *.
  set (s'' := fst p0) in * ; set (l'' := snd p0) in *.
  specialize (IHk k1 p s'' s' l'' l' H H0).
  intuition.
Qed.


(* This is just for simulation *)
Fixpoint execute_bounded (k:nat) (p:program) (s:state) (l:nat) {struct k} : state * nat :=
match k with
 | 0 => (s,l)
 | S j => match sem_instr (p l) s l with
            | None => (s,l)
            | Some (s',l') => execute_bounded j p s' l'
          end
end.

(* Programs represented as maps from positive numbers to instruction
This enables to use (find i prog) to get the ith instruction of prog *)

Set Implicit Arguments.

Definition prog := PositiveMap.tree instruction.

Definition empty_prog := PositiveMap.empty instruction.

Definition Prog_add := PositiveMap.add.

Definition Prog_find := PositiveMap.find.

Definition fragfromprog  (p:prog) (startaddress codeptr : nat) : option instruction :=
   if bleq_nat startaddress codeptr then
	 Prog_find (P_of_succ_nat (codeptr - startaddress)) p
  else
	None.

Fixpoint progfromlist_rec (n:nat) (list_instr : list instruction) {struct list_instr} : prog :=
  match list_instr with
     | nil => empty_prog
     | h :: t => Prog_add (P_of_succ_nat n) h (progfromlist_rec (S n) t)
  end.

Definition progfromlist := progfromlist_rec 0.

Lemma P_of_succ_nat_injective : forall m n, m <> n ->  (P_of_succ_nat m) <> (P_of_succ_nat n).
Proof.
  double induction n m.
  tauto.
  intros;  apply sym_not_eq; simpl.
  apply Psucc_not_one.
  intros;  simpl; apply Psucc_not_one.
  intros;simpl.
  intuition.
  apply Psucc_inj in H2.
  assert (n1 = n -> False); try omega.
  apply (H0 n H3 H2).
Qed.

Lemma partialy_def_prog_rec_find : forall l n l' m, n < (length l + m) ->
  Prog_find (P_of_succ_nat n) (progfromlist_rec m (l++l')) = Prog_find (P_of_succ_nat n) (progfromlist_rec m l).
Proof.
  induction l.
  simpl.
  induction l'; trivial.
  simpl.
  intros.
  rewrite PositiveMap.gso .
  assert (n < S m); try omega.
  hreplace IHl' with (IHl' (S m) H0); trivial.
  apply P_of_succ_nat_injective; omega.
  simpl; intros.
  compare m n.
  intro; rewrite e.
  rewrite PositiveMap.gss.
  rewrite PositiveMap.gss.
  trivial.
  intro.
  apply sym_not_eq in n0.
  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective;assumption).
  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective;assumption).
  replace (S (length l +m)) with (length l + S m) in H; try omega.
  apply (IHl n l' (S m) H).
Qed.

Lemma partialy_def_prog_find : forall l n l' , n < length l ->
             Prog_find (P_of_succ_nat n) (progfromlist (l++l')) = Prog_find (P_of_succ_nat n) (progfromlist l).
Proof.
  unfold progfromlist.
  intros.
  apply partialy_def_prog_rec_find.
  omega.
Qed.

Ltac instrextract startaddr offset :=
  unfold fragfromprog;
  match offset with
    | 0 => replace (startaddr - startaddr) with offset
    | _ => replace (startaddr + offset- startaddr) with offset
  end;
  [idtac | omega];
  (*unfold_plus;*)
  first
  [
     rewrite <- bleq_nat_false ; [idtac | omega];
     first [simpl; reflexivity |
            rewrite partialy_def_prog_find; [simpl; reflexivity | simpl;omega]]
     |
     rewrite <- bleq_nat_true ; [idtac | omega];
     first [simpl; reflexivity |
            rewrite partialy_def_prog_find; [simpl; reflexivity | simpl;omega]]
  ].

Lemma fragfromprog_empty : forall start , fragfromprog empty_prog start  = nullPfunction nat instruction.
Proof.
  intros; rewrite <- Extensionality; intro.
  unfold fragfromprog,empty_prog, Prog_find.
  case  (bleq_nat start a); [ rewrite PositiveMap.gempty | idtac]; reflexivity.
Qed.

Lemma impossible_find : forall l m n , n < m -> Prog_find (P_of_succ_nat n) (progfromlist_rec m l) = None.
Proof.
  induction l.
  simpl.
  intros; unfold empty_prog, Prog_find; rewrite PositiveMap.gempty; trivial.
  unfold progfromlist_rec; fold progfromlist_rec; intros.
  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective; omega).
  assert (n < S m); try omega.
  apply (IHl (S m) n H0).
Qed.

Lemma suffix_code_rec_single : forall l a start m , fragfromprog (progfromlist_rec m l) (start+1)  << fragfromprog (progfromlist_rec m (a::l)) start.
Proof.
  induction l.
  simpl;intros.
  rewrite fragfromprog_empty.
  unfold subfragment; apply subPfunction_unit.
  unfold progfromlist_rec; fold progfromlist_rec.
  intros.
  unfold subfragment; rewrite sub_is_extends; unfold extendsPfunction.
  intros until y.
  unfold fragfromprog.
  assert ((start + 1) <= x  \/ ~ (start+1) <= x); try apply dec_le.
  destruct H.
  assert (start <= x);try omega.
  rewrite <- bleq_nat_true ; [idtac | omega].
  rewrite <- bleq_nat_true ; [idtac | omega].
  assert ((x- (start+1)) <= m  \/ ~ ((x- (start+1))<=m)); try apply dec_le.
  destruct H1.
  compare (x- (start+1)) m.
  intro.
  replace (x- start) with (1+(x-(start+1))) ; try omega.
  rewrite e.
  rewrite PositiveMap.gss.
  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective; omega).
  rewrite PositiveMap.gss.
  trivial.
  intro.
  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective; omega).
  rewrite impossible_find; try omega.
  intro Habs; inversion Habs.

  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective; omega).
  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective; omega).
  hreplace IHl with (IHl a start (S m)).
  unfold subfragment in IHl ; rewrite sub_is_extends in IHl; unfold extendsPfunction in IHl.
  hreplace IHl with (IHl x y).
  unfold fragfromprog in IHl .
  rewrite <- bleq_nat_true in IHl; [idtac | omega].
  rewrite <- bleq_nat_true in IHl ; [idtac | omega].
  apply IHl.
  rewrite <- bleq_nat_false; [idtac | omega].
  intro Habs ; inversion Habs.
Qed.

Lemma suffix_code_rec : forall l l' start m , fragfromprog (progfromlist_rec m l) (start+ length l')  << fragfromprog (progfromlist_rec m (l'++l)) start.
Proof.
  induction l'.
  intros;simpl; replace (start+0) with start ; try omega;apply subfrag_refl.
  rewrite <- app_comm_cons.
  intros.
  eapply subfrag_trans; try apply suffix_code_rec_single.
  simpl.
  hreplace IHl' with (IHl' (S start) m).
  replace (start + 1 ) with (S start) ; try omega.
  replace (start + S (length l')) with (S start + length l'); try omega.
  apply IHl'.
Qed.

Lemma suffix_code : forall l l' start, fragfromprog (progfromlist l) (start+length l')  << fragfromprog (progfromlist (l'++l)) start.
Proof.
  unfold progfromlist.
  intros.
  apply suffix_code_rec.
Qed.

Lemma prefix_code_rec : forall l l' m start, fragfromprog (progfromlist_rec m l) (start)  << fragfromprog (progfromlist_rec m (l++l')) start.
Proof.
  induction l.
  simpl; intros.
  rewrite fragfromprog_empty.
  unfold subfragment; apply subPfunction_unit.
  simpl;intros.
  unfold subfragment; rewrite sub_is_extends; unfold extendsPfunction.
  intros until y.
  unfold fragfromprog.
  assert (start <= x  \/ ~ start <= x); try apply dec_le.
  destruct H.
  rewrite <- bleq_nat_true ; [idtac | omega].
  compare (x-start) m.
  unfold Prog_find, Prog_add.
  intro.
  rewrite e.
  rewrite PositiveMap.gss.
  rewrite PositiveMap.gss.
  trivial.
  intro.
  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective); try omega.
  rewrite PositiveMap.gso; try  (apply P_of_succ_nat_injective; omega).
  hreplace IHl with (IHl l'  (S m) start).
  unfold subfragment in IHl ; rewrite sub_is_extends in IHl; unfold extendsPfunction in IHl.
  hreplace IHl with (IHl x y).
  unfold fragfromprog in IHl .
  rewrite <- bleq_nat_true in IHl; [idtac | omega].
  apply IHl.
  rewrite <- bleq_nat_false; [idtac | omega].
  intro Habs ; inversion Habs.
Qed.

Lemma prefix_code : forall l l' start, fragfromprog (progfromlist l) (start)  << fragfromprog (progfromlist (l++l')) start.
Proof.
  unfold progfromlist; intros.
  apply prefix_code_rec.
Qed.

Lemma sub_code : forall l l' l'' start, fragfromprog (progfromlist l) (start + length l'')  << fragfromprog (progfromlist (l''++l++l')) start.
Proof.
  intros.
  eapply subfrag_trans ; try apply suffix_code.
  rewrite <- app_ass ; apply prefix_code.
Qed.

Lemma dismember_program : forall p l l' start,
    PfunctionExtendsFunction p (fragfromprog (progfromlist (l++l')) start) ->
   (PfunctionExtendsFunction p (fragfromprog (progfromlist (l)) start) /\
    PfunctionExtendsFunction p (fragfromprog (progfromlist (l')) (start + length l))).
Proof.
  intuition.
  eapply subPfunction_extends; try apply H.
  rewrite <- sub_is_extends; apply prefix_code.
  eapply subPfunction_extends; try apply H.
  rewrite <- sub_is_extends; apply suffix_code.
Qed.

Definition extract_code_from_globalcode (global_code : (nat -> list instruction * nat * nat) * list instruction * nat) start := (fst (fst global_code)) start.

Definition extract_stacksize_from_code (code : list instruction * nat * nat) :=
  snd (fst code).

Definition instruct_of (code :list instruction * nat * nat) := fst (fst code).

Definition prog_extends_code (global_code : (nat -> list instruction * nat * nat) * list instruction * nat) p start := PfunctionExtendsFunction p (fragfromprog (progfromlist (fst (fst (extract_code_from_globalcode global_code start)))) start). 

Definition prog_extends_auxcode (global_code : (nat -> list instruction * nat * nat) * list instruction * nat) p start := PfunctionExtendsFunction p (fragfromprog (progfromlist (snd (fst global_code))) start).

Definition prog_extends_maincode (closed_code : nat * list instruction) p start := 
  PfunctionExtendsFunction p (fragfromprog (progfromlist (snd closed_code)) start).

(* some definitions to test code *)

Definition program_step (p:program) s pc :=
  match sem_instr (p pc) s pc with
    | None => (s, pc)
    | Some (s', pc') => (s', pc')
  end.

Fixpoint program_trace (p:program) (s:state) (pc:nat) (steps:nat) {struct steps} : list nat :=
  match steps with
    | 0 => nil
    | S n => let (s', pc') := program_step p s pc in
             pc :: program_trace p s' pc' n
  end.