(* state_relations.v

   Define logical relation on states using Setoids
   Define the splitting tensor, the quantifications 
   and the modalities Later and Necessity.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export programs.

Implicit Arguments nullPfunction [].
Implicit Arguments join_disj_comm [A B].

Definition Astate := nat.
Definition Bstate := nat.

Record stateRel : Type := mkStateRel {
  staterel :> (Pfunction Astate Bstate) -> (Pfunction Astate Bstate)-> 
              program -> program -> nat ->
              Prop;
  stateRel_cond : forall s1 s2 s1' s2' p1 p2 k1 k2, 
    staterel s1 s2 p1 p2 k1 -> 
    subPfunction s1 s1' -> 
    subPfunction s2 s2' -> 
    (k2 <= k1) -> 
    staterel s1' s2' p1 p2 k2}.

Implicit Arguments stateRel_cond [].

Definition stateRelEq (P Q: stateRel) := forall s s' p p' k , P s s' p p' k <-> Q s s' p p' k.

Lemma stateRelEq_refl : forall (P : stateRel), stateRelEq P P.
Proof.
  firstorder.
Qed.

Lemma stateRelEq_comm : forall (P Q : stateRel), stateRelEq P Q -> stateRelEq Q P.
Proof.
  unfold stateRelEq; intros.
  hreplace H with (H s s' p p' k).
  intuition.
Qed.


Lemma stateRelEq_trans : forall (P Q R : stateRel), stateRelEq P Q -> stateRelEq Q R -> stateRelEq P R.
Proof.
  unfold stateRelEq; intros; simpl.
  hreplace H0 with (H0 s s' p p' k); hreplace H with (H s s' p p' k).
  firstorder.
Qed.


Hint Resolve stateRelEq_refl stateRelEq_trans.

Add Relation stateRel stateRelEq
   reflexivity proved by stateRelEq_refl
   symmetry proved by stateRelEq_comm
   transitivity proved by stateRelEq_trans
 as stateRelEqRel.

Add Morphism staterel with signature stateRelEq ==> (@eq (Pfunction Astate Bstate)) ==>
  (@eq (Pfunction Astate Bstate)) ==> (@eq program) ==> (@eq program) ==> (@eq nat) ==> iff  
  as staterelcompat.
Proof.
  auto.
Qed.

Definition stateRelLeq (P Q : stateRel) := forall s s' p p' k , P s s' p p' k -> Q s s' p p' k.

Lemma stateRelLeq_refl : forall P, stateRelLeq P P.
Proof.
  unfold stateRelLeq; intuition.
Qed.

Lemma stateRelLeq_trans : forall P Q R,  stateRelLeq P Q -> stateRelLeq Q R -> stateRelLeq P R.
Proof.
  unfold stateRelLeq; intuition.
Qed.

Hint Resolve stateRelLeq_refl stateRelLeq_trans.

Add Relation stateRel stateRelLeq
   reflexivity proved by stateRelLeq_refl
   transitivity proved by stateRelLeq_trans
 as stateRelLeqRel.

Add Morphism stateRelLeq with signature stateRelEq ==> stateRelEq ==> iff
as ppredleqcompat.
Proof.
  unfold stateRelEq, stateRelLeq; intuition.
  assert (foo := H s s' p p' k ); assert (foo2 := H0 s s' p p' k).
  intuition.
  assert (foo := H s s' p p' k ); assert (foo2 := H1 s s' p p' k); assert (foo3 := H0 s s' p p' k).
  intuition.
Qed.

Lemma stateRelEq_Leq: forall P Q, stateRelEq P Q -> stateRelLeq P Q.
Proof.
  unfold stateRelLeq,stateRelEq; intros.
  erewrite <- H; assumption.
Qed.

(* Lifting pure predicates, true and false *)

Definition stateRel_lift (P: Prop) : stateRel.
Proof.
  intro P.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun  k => P) _);
  (intros; assumption).
Defined.

Add Morphism stateRel_lift with signature iff ==> stateRelEq
as ppredliftcompat.
Proof.
  unfold stateRelEq, stateRel_lift; simpl; intuition. 
Qed.

Definition statePtrue := stateRel_lift (True).

Definition statePfalse := stateRel_lift (False).

Lemma stateRellift_Leq : forall (P Q : Prop), (P -> Q) -> stateRelLeq (stateRel_lift P) (stateRel_lift Q).
Proof.
  unfold stateRel_lift, stateRelLeq; intuition.
Qed.

Lemma stateRelLeq_false_min : forall P , stateRelLeq statePfalse P.
Proof.
  unfold stateRelLeq, statePfalse, stateRel_lift; simpl; intros.
  inversion H.
Qed.
  
(*-------------*)
(* stateRelTensor *)
(*-------------*)

Definition stateRelTensor (P1 P2 : stateRel) : stateRel.
Proof. 
  intros P1 P2.
  refine (mkStateRel (fun f => fun f' => fun p => fun p' => fun k => 
    exists f1 : (Pfunction Astate Bstate), exists f2 : (Pfunction Astate Bstate), 
    exists f'1 : (Pfunction Astate Bstate), exists f'2 : (Pfunction Astate Bstate), 
      disjointPfunction f1 f2 /\  f = (joinPfunction f1 f2) /\  
      disjointPfunction f'1 f'2 /\  f' = (joinPfunction f'1 f'2) /\
      P1 f1 f'1 p p' k /\ P2 f2 f'2 p p' k)
                _).
  intros.
  witness f1 in H; witness f2 in H; witness f'1 in H; witness f'2 in H.
  deconstruct H as [Hdisj12 [Hjoin112 [Hdisj34 [Hjoin134 [HP13 HP24]]]]].
  unfold subPfunction in H0; witness fdiff in H0.
  unfold subPfunction in H1; witness fdiff' in H1.
  exists (joinPfunction f1 fdiff); exists f2.
  exists (joinPfunction f'1 fdiff'); exists f'2.
  rewrite Hjoin112 in H0; destruct H0.
  rewrite Hjoin134 in H1; destruct H1.
  assert (Hdisj := disj_join_left f1 f2 fdiff). 
  split.
  rewrite  (disj_join_left f1 fdiff f2);  split; [intuition | idtac] . 
  rewrite (disjointPfunction_comm fdiff f2); intuition.
  split.
  rewrite  <- (disj_join_assoc); [idtac | intuition | rewrite (disjointPfunction_comm fdiff f2); intuition | intuition].
  rewrite (join_disj_comm fdiff f2); [idtac | rewrite (disjointPfunction_comm fdiff f2); intuition ].
  rewrite  (disj_join_assoc); [idtac | intuition | intuition | intuition].
  assumption.
  assert (Hdisj' := disj_join_left f'1 f'2 fdiff'). 
  split.
  rewrite  (disj_join_left f'1 fdiff' f'2);  split; [intuition | idtac] . 
  rewrite (disjointPfunction_comm fdiff' f'2); intuition.
  split.
  rewrite  <- (disj_join_assoc); [idtac | intuition | rewrite (disjointPfunction_comm fdiff' f'2); intuition | intuition].
  rewrite (join_disj_comm fdiff' f'2); [idtac | rewrite (disjointPfunction_comm fdiff' f'2); intuition ].
  rewrite  (disj_join_assoc); [idtac | intuition | intuition | intuition].
  assumption.
  split.  
  eapply (stateRel_cond P1).   
  apply HP13.
  apply Pfunction_ext; intuition.
  unfold subPfunction; exists fdiff'; intuition. 
  assumption.
  apply (stateRel_cond P2 f2 f'2 f2 f'2 p1 p2 k1 k2);
  [assumption | apply subPfunction_refl | apply subPfunction_refl | assumption].
Defined.


Add Morphism stateRelTensor with signature stateRelEq ==> stateRelEq ==> stateRelEq
as ppredtensorcompat.
Proof.
  unfold stateRelEq, stateRelTensor; simpl; firstorder;
  (exists x1; exists x2; exists x3; exists x4;
  hreplace H0 with (H0 x2 x4 p p' k); hreplace H with (H x1 x3 p p' k);
  intuition).
Qed.

Add Morphism stateRelTensor with signature stateRelLeq ++> stateRelLeq ++> stateRelLeq
as ppredtensorcompat2.
Proof.
  unfold stateRelLeq,stateRelTensor; simpl; intros.
  destruct H1 as [f1 [f2 [f1' [f2' H1]]]].
  exists f1; exists f2; exists f1'; exists f2'.
  intuition.
Qed.


(******************)
(*    Notation     *)
(*******************)

Notation "P1 ## P2" := (stateRelTensor P1 P2) (at level 75, right associativity).

(* RelTensor Properties in a setoid (or semantic consequence properties) *)

Lemma stateRelTensor_unit : forall P, stateRelEq (statePtrue ## P) P.
Proof.
  unfold stateRelEq; unfold stateRelTensor; simpl. 
  intros ; split; intros.
  witness f1 in H; witness f2 in H.
  witness f'1 in H; witness f'2 in H.
  deconstruct H as [Hdisj12 [Hjoins12 [Hdisj12' [Hjoins12' [HP1 Htrue]]]]].
  rewrite Hjoins12; rewrite Hjoins12'.
  eapply stateRel_cond.
  apply Htrue.
  rewrite join_disj_comm.
  apply Pfunction_ext; intuition.
  rewrite disjointPfunction_comm; assumption.
  assumption.
  rewrite join_disj_comm.
  apply Pfunction_ext; intuition.
  rewrite disjointPfunction_comm; assumption.
  assumption.
  trivial.
  
  exists (nullPfunction Astate Bstate); exists s;  
  exists (nullPfunction Astate Bstate); exists s'. 
  intuition.
  rewrite disjointPfunction_comm; apply disjointPfunction_unit.
  rewrite join_disj_comm; [
  apply join_unit | rewrite disjointPfunction_comm; apply disjointPfunction_unit].
  rewrite disjointPfunction_comm; apply disjointPfunction_unit.
  rewrite join_disj_comm; [
  apply join_unit | rewrite disjointPfunction_comm; apply disjointPfunction_unit].
Qed.


Lemma stateRelTensor_zero : forall P, stateRelEq (statePfalse ## P) statePfalse.
Proof.
  unfold stateRelEq; unfold stateRelTensor; unfold statePfalse; simpl.
  firstorder.
Qed.


Lemma stateRelTensor_comm : forall P Q, stateRelEq (P ## Q) (Q ## P).
Proof.
  intros; unfold stateRelEq, stateRelTensor; simpl; intros.  
  split; intro H;  witness f1 in H;  witness f2 in H; witness f'1 in H;  witness f'2 in H.
  exists f2;  exists f1; exists f'2;  exists f'1.  
  firstorder. 
  rewrite H0;  apply join_disj_comm; assumption.
  rewrite H2;  apply join_disj_comm; assumption.
  exists f2;  exists f1; exists f'2;  exists f'1.  
  firstorder. 
  rewrite H0;  apply join_disj_comm; assumption.
  rewrite H2;  apply join_disj_comm; assumption.
Qed.
  
Lemma stateRelTensor_assoc : forall P Q R, stateRelEq (P ## (Q ## R)) ((P ## Q) ## R).
Proof.
  intros;  unfold stateRelEq, stateRelTensor; simpl; intros.
  split; intros; witness f1 in H; witness f2 in H; witness f'1 in H;  witness f'2 in H.
  deconstruct H as [Hdisj12 [Hjoin12 [Hdisj12' [Hjoin12' [HP1 Hex]]]]].
  witness f3 in Hex; witness f4 in Hex; witness f'3 in Hex; witness f'4 in Hex.  
  deconstruct Hex as [Hdij34 [Hjoin234 [Hdij34' [Hjoin234' [HQ3 HR4]]]]].
  exists (joinPfunction f1 f3);  exists f4; exists (joinPfunction f'1 f'3);  exists f'4.
  split.

  rewrite Hjoin234 in Hdisj12.
  
  apply disjPfunction_joinl; [assumption | assumption].

  rewrite Hjoin234 in Hjoin12.
  rewrite Hjoin234 in Hdisj12.
  assert (H0:= disj_join_right f1 f3 f4).
 
  split.
  rewrite Hjoin12.
   
  rewrite <- disj_join_assoc; [reflexivity | intuition | intuition | intuition].
  
  split.
  rewrite Hjoin234' in Hdisj12'.
  apply disjPfunction_joinl; [assumption | assumption].

  rewrite Hjoin234' in Hjoin12'.
  rewrite Hjoin234' in Hdisj12'.
  assert (H0':= disj_join_right f'1 f'3 f'4).

  split.
  rewrite Hjoin12'.
   
  rewrite <- disj_join_assoc; [reflexivity | intuition | intuition | intuition].

  split ; [idtac|assumption].

  exists f1; exists f3;   exists f'1; exists f'3.
  intuition.

  deconstruct H as [Hdisj12 [Hjoin12 [Hdisj12' [Hjoin12' [Hex HR2]]]]].
  witness f3 in Hex;  witness f4 in Hex;witness f'3 in Hex;  witness f'4 in Hex.
  deconstruct Hex as [Hdisj34 [Hjoin134 [Hdisj34' [Hjoin134' [HP3 HQ4]]]]].
  exists f3;  exists (joinPfunction f4 f2); exists f'3;  exists (joinPfunction f'4 f'2).

  rewrite Hjoin134 in Hdisj12.
  assert (Hdisj42 := disj_join_left f3 f4 f2).
  split.

  apply disjPfunction_joinr; [assumption | assumption].
  split.
  rewrite Hjoin134 in Hjoin12.
  rewrite Hjoin12.
  apply sym_eq.
  apply disj_join_assoc; [assumption | intuition | intuition].
  
  rewrite Hjoin134' in Hdisj12'.
  assert (Hdisj42' := disj_join_left f'3 f'4 f'2).
  split.

  apply disjPfunction_joinr; [assumption | assumption].
  split.
  rewrite Hjoin134' in Hjoin12'.
  rewrite Hjoin12'.
  apply sym_eq.
  apply disj_join_assoc; [assumption | intuition | intuition].
 
  split.
  assumption.  
   
  exists f4; exists f2; exists f'4; exists f'2.
  split; [intuition | split; [reflexivity | split; [intuition | intuition]]].
Qed.

(*-------------------------*)
(* Disj *)
(*-------------------------*)

Definition stateRelDisj (P Q: stateRel) : stateRel.
Proof. 
  intros.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => P s s' p p' k \/ Q s s' p p' k)
                  _).
  intros.
  destruct H.
  left; apply (stateRel_cond P s1 s2 s1' s2' p1 p2 k1 k2);assumption.
  right; apply (stateRel_cond Q s1 s2 s1' s2' p1 p2 k1 k2); assumption.
Defined.


Add Morphism stateRelDisj with signature stateRelEq ==> stateRelEq ==> stateRelEq
as ppreddisjcompat.
Proof.
  unfold stateRelDisj in *; unfold stateRelEq in *; simpl; intros.
  hreplace H with (H s s' p p' k); hreplace H0 with (H0 s s' p p' k).
  intuition.
Qed.

Add Morphism stateRelDisj with signature stateRelLeq ++> stateRelLeq ++> stateRelLeq
as ppreddisjcompat2.
Proof.
  unfold stateRelDisj; unfold stateRelLeq; simpl; intros.
  intuition.
Qed.


(******************)
(*     Notation    *)
(*******************)

Notation "P1 @@ P2" := (stateRelDisj P1 P2) (at level 75, right associativity).

(*----------------------------*)
(* Disj properties *)
(*----------------------------*)

Lemma stateRelDisj_unit : forall P: stateRel, stateRelEq (statePfalse @@ P) P.
Proof.
  unfold stateRelEq, stateRelDisj; simpl;intros.
  intuition.
Qed.

Lemma stateRelDisj_comm : forall P Q, stateRelEq (P @@ Q) (Q @@ P).
Proof.
  unfold stateRelEq, stateRelDisj; simpl; intros.
  tauto.
Qed.

Lemma stateRelDisj_assoc : forall P Q R, stateRelEq (P @@ (Q @@ R)) ((P @@ Q) @@ R).
Proof.
  unfold stateRelEq, stateRelDisj; simpl; intros.
  tauto.
Qed.

(*----------------------------*)
(* Tensor vs Disj properties *)
(*----------------------------*)

Lemma stateRelDistributivity : forall P Q R, stateRelEq ( ( P @@ Q) ## R) ((P ## R) @@ (Q ## R)).
Proof.
  unfold stateRelEq, stateRelDisj, stateRelTensor; simpl; intuition.
  firstorder.
  left;exists x; exists x0; exists x1; exists x2; firstorder.
  right;exists x; exists x0; exists x1; exists x2; firstorder.
  firstorder; exists x; exists x0; exists x1; exists x2; firstorder.
  firstorder; exists x; exists x0; exists x1; exists x2; firstorder.
Qed.

(*----------------------------*)
(* Definition of a ring on stateRel *)
(*---------------------------*)


Definition ppred_ring : semi_ring_theory statePfalse statePtrue stateRelDisj stateRelTensor stateRelEq. 
Proof.
apply mk_srt.
apply stateRelDisj_unit. apply stateRelDisj_comm. apply stateRelDisj_assoc.
apply stateRelTensor_unit. apply stateRelTensor_zero. apply stateRelTensor_comm.
apply stateRelTensor_assoc. apply stateRelDistributivity.
Qed.

Definition my_setoid : Setoid_Theory stateRel stateRelEq.
Proof.
  constructor.
  unfold Reflexive; apply stateRelEq_refl.
  unfold Symmetric; apply stateRelEq_comm.
  unfold Transitive; apply stateRelEq_trans.
Defined.

Definition my_ring : sring_eq_ext stateRelDisj stateRelTensor stateRelEq.
  apply mk_seqe.
  apply ppreddisjcompat.
  apply ppredtensorcompat.
Defined.
  
Add Ring stateRel_ring : ppred_ring (setoid my_setoid my_ring).

(* tensor_predlift *)

Lemma stateRelTensor_weakening : forall P Q, stateRelLeq (P ## Q) P.
Proof.
  unfold stateRelTensor, stateRelLeq; simpl;  intros.
  witness f1 in H; witness f2 in H; witness f'1 in H; witness f'2 in H.
  deconstruct  H as [H1 [H2 [H3 [H4 [H5 H6]]]]].
  eapply (stateRel_cond P);
  [ apply H5 | idtac | idtac | trivial].
  firstorder.
  firstorder.
Qed.

Lemma stateRelTensor_weakening_general : forall P Q R, stateRelLeq P R -> 
	 			stateRelLeq (P ## Q) R.
Proof.
  unfold stateRelTensor, stateRelLeq; simpl;  intros.
  witness f1 in H0; witness f2 in H0; witness f'1 in H0; witness f'2 in H0. 
  eapply (stateRel_cond R).
  deconstruct  H0 as [H1 [H2 [H3 [H4 [H5 H6]]]]].
  apply (H f1 f'1 p p' k H5).
  firstorder.
  firstorder.
  trivial.  
Qed.

Lemma statetensor_predlift : forall (P : stateRel) (Q:Prop)  s s' p p' k , 
  P s s' p p' k /\ (stateRel_lift Q) s s' p p' k
  -> (P ## (stateRel_lift Q)) s s' p p' k.
Proof.
  unfold stateRelTensor; simpl; intros.
  exists s; exists (nullPfunction Astate Bstate); exists s'; exists (nullPfunction Astate Bstate).
  intuition.
Qed.


(************)
(* Extensionality to rewrite under Exists and Forall  *)
(************)


Definition stateRelExtensionalityLeq (X:Type) (P Q : X -> stateRel) := forall n, stateRelLeq (P n) (Q n).

Add Relation (nat -> stateRel) (@stateRelExtensionalityLeq nat) as stateRelExt2.

Add Relation ((list nat) -> stateRel) (@stateRelExtensionalityLeq (list nat)) as stateRelExtList2.


Definition stateRelExtensionality (X:Type) (P Q : X -> stateRel) := forall n, stateRelEq (P n) (Q n).

Add Relation (nat -> stateRel) (@stateRelExtensionality nat) as stateRelExt.

Add Relation ((list nat) -> stateRel) (@stateRelExtensionality (list nat)) as stateRelExtList.


(************)
(* ExistP  *)
(************)

Definition stateExists (X : Type) (h : X -> stateRel ) : stateRel.
  intros.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => exists a: X, h a s s' p p' k) _).
  intros.
  witness a in H.
  exists a.
  apply (stateRel_cond (h a) s1 s2 s1' s2' p1 p2 k1 k2); assumption.
Defined.

Implicit Arguments stateExists [X].

Notation "'Exists' h" := (stateExists  h ) (at level 70, right associativity).

Notation "'Ex' x , P" := (Exists (fun x=> P) ) (at level 70, x at next level,right associativity).

Add Morphism (@stateExists nat) with signature (@stateRelExtensionality nat) ==> stateRelEq
as existsstatenat.
Proof.
    unfold stateRelEq, stateRelExtensionality,stateExists; simpl; firstorder.
    exists x0; rewrite <- H; intuition.
    exists x0; rewrite H; intuition.
Qed.

Add Morphism (@stateExists (list nat)) with signature (@stateRelExtensionality (list nat)) ==> stateRelEq
as existsstatellistnat.
Proof.
  unfold stateRelEq, stateRelExtensionality,stateExists; simpl; firstorder. 
   exists x0; rewrite <- H; intuition.
   exists x0; rewrite H; intuition.
Qed.


Add Morphism (@stateExists nat) with signature (@stateRelExtensionalityLeq nat) ++> stateRelLeq
as existsstatenat2.
Proof.
    unfold stateRelExtensionalityLeq,stateExists,stateRelLeq; simpl; firstorder.
Qed.

Add Morphism (@stateExists (list nat)) with signature (@stateRelExtensionalityLeq (list nat)) ==> stateRelLeq
as existsstatellistnat2.
Proof.
    unfold stateRelExtensionalityLeq,stateExists,stateRelLeq; simpl; firstorder.
Qed.

(*****************************)
(* some lemmas about Exists *)
(*****************************)

Lemma stateExP_mono : forall (X: Type) (P Q : X -> stateRel), 
  (forall x, stateRelLeq (P x) (Q x)) -> (stateRelLeq (Exists P) (Exists Q)).
Proof.
  unfold stateRelLeq; simpl;  intros.
  witness a in H0.
  exists a.
  firstorder.
Qed.

Lemma stateExP_rewrite : forall (X:Type) (P Q: X->stateRel), 
  (forall x, stateRelEq (P x) (Q x)) ->  (stateRelEq (Exists P) (Exists Q)).
Proof.
  unfold stateRelEq; intros; simpl.
  firstorder; exists x; assert (foo:= H x s s' p p' k);  firstorder.
Qed.

Lemma stateExP_tensor : forall (X: Type) (h:X -> stateRel) P,
  stateRelEq (Ex x, (h x ## P)) (Ex x, (h x) ## P).
Proof.
  split; unfold stateRelLeq; simpl; intros.
  firstorder;  exists x0; exists x1; exists x2; exists x3;  firstorder.
  firstorder.
  exists x3; exists x; exists x0; exists x1; exists x2; firstorder.
Qed.

Lemma stateExists_weakening:  forall (A : Type) (P Q : A -> stateRel), 
	stateRelLeq (Ex x, (P x ## Q x)) (Ex x, P x).
  intros.
  unfold stateRelLeq.
  apply stateExP_mono.
  intros.
  apply stateRelTensor_weakening.
Qed.

Lemma stateExistsLeq : forall (A:Type) (h:A->stateRel) P Q,
 (exists x : A, stateRelLeq Q (h x ## P))
 -> stateRelLeq Q ((Exists h) ## P).
  intros until Q; unfold stateRelLeq, stateExists.
  simpl.
  firstorder.
  hreplace H with (H s s' p p' k H0).
  firstorder.
  exists x0; exists x1; exists x2; exists x3.
  firstorder.
Qed.

Lemma stateForallLeq : forall (A:Type) (h:A->stateRel) P Q,
 (forall x : A, stateRelLeq ( h x ## P) Q)
 -> stateRelLeq ((Exists h)##P) Q .
Proof. 
  intros until Q; unfold stateRelLeq, stateExists; simpl;  firstorder.
  apply (H x3 s s' p p' k).
  exists x; exists x0; exists x1; exists x2.
  intuition.
Qed.

Lemma stateExistsLeq_single : forall (A:Type) (h:A->stateRel) Q,
 (exists x : A, stateRelLeq Q (h x))
 -> stateRelLeq Q ((Exists h)).
  intros until Q; unfold stateRelLeq, stateExists.
  simpl.
  firstorder.
Qed.

Lemma stateForallLeq_single : forall (A:Type) (h:A->stateRel) Q,
 (forall x : A, stateRelLeq (h x) Q)
 -> stateRelLeq ((Exists h)) Q .
Proof. 
  intros until Q; unfold stateRelLeq, stateExists.
  simpl.
  firstorder.
Qed.

Lemma stateRellift_val : forall  P Q s s' p p' k, ((stateRel_lift Q) ## P) s s' p p' k -> Q.
Proof.
  unfold stateRel_lift; simpl; intros.
  firstorder.
Qed.

Lemma stateExists_tensorsplit : forall (X: Type) (h g:X -> stateRel),
  stateRelLeq (Ex x, (h x ## g x)) (Ex x, Ex y , (h x ## g y)).
Proof.
  unfold stateRelLeq; simpl; intros.
  firstorder.
  exists x; exists x; exists x0; exists x1; exists x2; exists x3.
  firstorder.
Qed.

Lemma stateExP_comm : forall (X Y: Type) (h: X -> Y -> stateRel),
  stateRelEq (Ex x,Ex y, (h x y)) (Ex y, Ex x, (h x y)).
Proof.
  split; unfold stateRelLeq; simpl; intros;  firstorder.
Qed.



(************)
(* ForAllP  *)
(************)

Definition stateForall (X : Type) (h : X -> stateRel ) : stateRel.
  intros.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => forall a: X, h a s s' p p' k) _).
  intros.
  hreplace H with (H a).
  apply (stateRel_cond (h a) s1 s2 s1' s2' p1 p2 k1 k2);  assumption.
Defined.

Implicit Arguments stateForall [X].

Notation "'ForAll' h" := (stateForall  h ) (at level 70, right associativity).

Notation "'Forall' x , P" := (ForAll (fun x=> P) ) (at level 70, x at next level,  right associativity).

Add Morphism (@stateForall nat) with signature (@stateRelExtensionality nat) ==> stateRelEq
as forallPstatenat.
Proof.
    unfold stateRelEq, stateRelExtensionality,stateForall; simpl; intuition.
    rewrite <- H; intuition.
    rewrite H; intuition.
Qed.

Add Morphism (@stateForall (list nat)) with signature (@stateRelExtensionality (list nat)) ==> stateRelEq
as foralstatellistnat.
Proof.
  unfold stateRelEq, stateRelExtensionality,stateForall; simpl; intuition. 
  rewrite <- H; intuition.
  rewrite H; intuition.
Qed.


Add Morphism (@stateForall nat) with signature (@stateRelExtensionalityLeq nat) ++> stateRelLeq
as forallPstatenat2.
Proof.
    unfold stateRelExtensionalityLeq,stateForall,stateRelLeq; simpl; intuition.
Qed.

Add Morphism (@stateForall (list nat)) with signature (@stateRelExtensionalityLeq (list nat)) ==> stateRelLeq
as foralstatellistnat2.
Proof.
    unfold stateRelExtensionalityLeq,stateForall,stateRelLeq; simpl; intuition.
Qed.

Lemma stateForall_ForallLeq : forall (X:Type) (h:X->stateRel) P Q,
  stateRelLeq Q ((Forall x, h x) ## P)
  -> (forall x : X, stateRelLeq Q ((h x) ## P)).
Proof.
  unfold stateRelLeq, stateForall; simpl; firstorder.
  hreplace H with  (H s s' p p' k H0);
  firstorder.
  exists x0; exists x1; exists x2; exists x3; firstorder.
Qed.

Lemma stateForall_ExistsLeq : forall (X:Type) (h:X->stateRel) P Q,
 (exists x : X, stateRelLeq ((h x)##P) Q)
 -> stateRelLeq ((Forall x, h x)##P) Q .
Proof. 
  unfold stateRelLeq, stateExists; simpl;  firstorder.
  apply (H s s' p p' k).
  exists x0; exists x1; exists x2; exists x3; firstorder.
Qed.

Lemma stateForall_ForallLeq_single : forall (X:Type) (h:X->stateRel) Q,
 (forall x : X, stateRelLeq Q (h x)) <-> stateRelLeq Q ((Forall x, h x)).
Proof.
  unfold stateRelLeq, stateExists;simpl; firstorder.
Qed.

Lemma stateForall_ExistsLeq_single : forall (X:Type) (h:X->stateRel) Q,
 (exists x : X, stateRelLeq (h x) Q) -> stateRelLeq ((Forall x, h x)) Q .
Proof. 
  unfold stateRelLeq, stateExists; simpl; firstorder.
Qed.

Lemma stateForall_tensor : forall (X: Type) (h:X -> stateRel) P,
  stateRelLeq ((Forall x, (h x)) ## P) (Forall x, (h x) ## P).
Proof.
  unfold stateRelLeq; simpl; intros;  firstorder.
Qed.

Lemma statelP_mono : forall (X: Type) (P Q : X -> stateRel), 
  (forall x, stateRelLeq (P x) (Q x)) -> (stateRelLeq (ForAll P) (ForAll Q)).
Proof.
  unfold stateRelLeq; simpl;  intros.
  apply (H a); apply (H0 a).
Qed.

Lemma stateForall_rewrite : forall (X:Type) (P Q: X->stateRel), 
  (forall x, stateRelEq (P x) (Q x)) ->  (stateRelEq (ForAll P) (ForAll Q)).
Proof.
  unfold stateRelEq; simpl; firstorder;  hreplace H with (H a s s' p p' k); intuition.
Qed.


(* *******************  Modality ********************* *)

Definition Later_s (P:stateRel) : stateRel.
  intros P.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => 
    forall j, j < k -> P s s' p p' j) _).
  intuition.
  specialize (H j).
  eapply stateRel_cond; intuition.
Defined.

Add Morphism Later_s with signature 
  stateRelEq ==> stateRelEq
as laterScompat.
Proof.
  intros R R'; subst;  unfold Later_s,stateRelEq; intros; subst; simpl; intuition;
  eapply H; eapply H0; assumption. 
Qed.

Add Morphism Later_s with signature 
  stateRelLeq ++> stateRelLeq
as laterScompat2.
Proof.
  intros R R'; subst;  unfold Later_s,stateRelLeq; intros; subst; simpl; intuition;
  eapply H; eapply H0; assumption. 
Qed.

Definition Necessity_s (P:stateRel) : stateRel.
  intros P.
  refine (mkStateRel (fun s => fun s' => fun p => fun p' => fun k => 
    forall j, j <= k -> P s s' p p' j) _).
  intuition.
  specialize (H j).
  eapply stateRel_cond; intuition.
Defined.

Add Morphism Necessity_s with signature 
  stateRelEq ==> stateRelEq
as necessityScompat.
Proof.
  intros R R'; subst;  unfold Necessity_s,stateRelEq; intros; subst; simpl; intuition;
  eapply H; eapply H0; assumption. 
Qed.

Add Morphism Necessity_s with signature 
  stateRelLeq ++> stateRelLeq
as necessityScompat2.
Proof.
  intros R R'; subst;  unfold Necessity_s,stateRelLeq; intros; subst; simpl; intuition;
  eapply H; eapply H0; assumption. 
Qed.