(* theory_of_types.v

   Define the relational semantics of types. 
   The interesting case is for the arrow type.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import compiler.
Require Import judgment.
Require Import cloud_relations.

Lemma triple_destruction : forall (A B:Type) (p1 p2:A * B),
  (fst p1 = fst p2 /\ snd p1= snd p2) <-> p1 = p2.
Proof.
  destruct p1; destruct p2.
  simpl;  split;
  [ intros; destruct H as (Hfst,Hsnd); rewrite Hfst ; rewrite Hsnd; reflexivity |
  intuition; inversion H; trivial].
Qed.

Definition storing_space n n2 n3 n' n2' n3' := 
  ({n+2,n'+2}|->{n2,n2'}) ## ({n+3,n'+3}|->{n3,n3'}) ## ({n+4,n'+4}|->-).

Definition unused_space := (7|->-) ## (8|->-) ## (9|->-).

Definition stack_list_type := list ((nat -> nat -> cloudRel) * (nat * nat)).

Notation "n |-> { m , - }" := (left_ptsto n (fun l => l=m )) (at level 75, m at next level).

Notation "n |-> { - , m }" := (right_ptsto n (fun l => l=m )) (at level 75, m at next level).


Fixpoint stack_description (stack_list:stack_list_type) ptr ptr' cloud {struct stack_list} :=
  match stack_list with 
  | nil => cloudToState cloud
  | pair h (pair ptr_h ptr_h') :: t => 
    ({ptr,ptr'} |-> {ptr_h,ptr_h'}) ## 
    stack_description t (ptr-1) (ptr'-1) ((h ptr_h ptr_h') //\\ cloud)
  end.

Definition Post_arrow b (Ra Rc: stateRel) Rc_cloud (n n' stack_ptr stack_ptr': nat):=
  Ex ptr_result, Ex ptr_result', (
    ({stack_ptr,stack_ptr'} |-> {ptr_result,ptr_result'}) ## 
    ({stack_ptr+1,stack_ptr'+1}|->-)
    ## cloudToState ((b Ra ptr_result ptr_result') //\\ Rc_cloud)
    ## (workreg |-> -) ## (argreg |-> -) ## (retreg |-> -) ## Ra ## Rc
    ## (3|->-) ## (4 |->-)  ## (spreg|-> {stack_ptr,stack_ptr'}) ##
    (envreg|-> {n,n'})) ## unused_space.   

Definition Pre_arrow R_private ptr_function ptr_function' Ra a b:= Ex Rc, Ex Rc_cloud, Ex n, Ex n', Ex ptr_arg, Ex ptr_arg', Ex stack_ptr, Ex stack_ptr',
    (({stack_ptr,stack_ptr'}|-> {ptr_arg,ptr_arg'}) ## 
      ({stack_ptr+1,stack_ptr'+1}|-> {ptr_function,ptr_function'}) ##  
    cloudToState (R_private //\\ a Ra ptr_arg ptr_arg' //\\ Rc_cloud) ##
    cloudToState (({n+4,n'+4}||=> Later (Perp (Post_arrow b Ra Rc Rc_cloud n n' stack_ptr stack_ptr'))) //\\ stateToCloud Rc)
    ## (workreg |-> -) ## (argreg |-> -) ## (retreg |-> -) ## Ra 
    ## (3|->-) ## (4|->-) ## (spreg|-> {stack_ptr+1,stack_ptr'+1}) ##
    (envreg|-> {n,n'})) ## unused_space.

Reserved Notation "'[|' t '|]'" (at level  75).

Reserved Notation "'[||' env '||]'" (at level  75).

Definition bool_of_nat n := match n with 
  | 0 => false
  | S n => true
end.

Fixpoint semantics_of_types (t:ExpType) (Ra:stateRel) ptr ptr' {struct t} := 
  match t with 
  | Int Pred => cloudRel_lift (Pred ptr /\ Pred ptr' /\ (ptr=ptr'))
  | Bool Pred => cloudRel_lift (Pred (bool_of_nat ptr) /\  Pred (bool_of_nat ptr') /\ (bool_of_nat ptr = bool_of_nat ptr'))
  | a ** b  => Ex_c value, Ex_c value2, Ex_c value', Ex_c value2', 
      (({ptr,ptr'}||->{value,value'}) //\\ ({ptr+1,ptr'+1}||->{value2,value2'})
        //\\ [|b|] Ra value value' //\\ [|a|] Ra value2 value2')
  | a ---> b  => Ex_c Rprivate, ({ptr,ptr'} ||=> 
       Later ( Perp (Pre_arrow Rprivate ptr ptr' Ra ([|a|]) ([|b|]))) //\\ Rprivate)
  end
  where "'[|' t '|]'" := (semantics_of_types t ).

Fixpoint semantics_of_env (env:EnvType) Ra current current' {struct env} :=
  match env with 
  | nil => cloudPtrue
  | h :: t => Ex_c ptr, Ex_c ptr', Ex_c next, Ex_c next', 
      (({current,current'} ||-> {ptr,ptr'}) //\\ ({current+1,current'+1} ||-> {next,next'})
       //\\ [|h|] Ra ptr ptr' //\\ [|| t ||] Ra next next')
  end
where  "'[||' env '||]'" := (semantics_of_env env).


Definition memory_specification Ra Rc Rc_cloud env stack_list stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3':= 
          (spreg|->{stack_ptr,stack_ptr'}) ##
          Block (stack_ptr+1) (stack_ptr'+1) stack_free stack_free' ##
          (envreg|-> {n,n'}) ##  storing_space n n2 n3 n' n2' n3' ##
	   stack_description stack_list stack_ptr stack_ptr' 
             ([||env||] Ra n n' //\\ Rc_cloud)
          ## (workreg |-> -) ## (argreg |-> -) ## (retreg |-> -) ## Ra ## Rc
	  ## (3|->-) ## (4|->-) ## unused_space.

Add Morphism stack_description with signature 
  (@eq (list ((nat -> nat -> cloudRel) * (nat * nat)))) ==> (@eq nat) ==> (@eq nat) ==> 
  cloudRelLeq ++> stateRelLeq
as stackdescriptioncompat2.
Proof.
  intro list; induction list.
  simpl ; trivial.
  simpl; intros; rewrite (surjective_pairing a); rewrite (surjective_pairing (snd a)).
  eapply ppredtensorcompat2; try trivial.
  eapply IHlist; rewrite H; trivial.
Qed.

Add Morphism stack_description with signature 
  (@eq (list ((nat -> nat -> cloudRel) * (nat * nat)))) ==> 
  (@eq nat) ==> (@eq nat) ==> cloudRelEq ==> stateRelEq
as stackdescriptioncompat.
Proof.
  intro list; induction list.
  simpl ; trivial.
  simpl; intros; rewrite (surjective_pairing a); rewrite (surjective_pairing (snd a)).
  eapply ppredtensorcompat; try trivial.
  eapply IHlist; rewrite H; trivial.
Qed.

Lemma stack_description_explicit : forall list ptr ptr', exists P, exists Q , 
  forall cloud, stateRelEq (stack_description list ptr ptr' cloud) (cloudToState (Q//\\ cloud) ## P).
Proof.
  induction list.
  simpl; intros.
  exists statePtrue; exists cloudPtrue; intros.
  rewrite cloudRelConj_unit; ring.

  simpl; intros.
  rewrite (surjective_pairing a); rewrite (surjective_pairing (snd a)).
  set (h := fst a); set (prearg :=snd a); set (arg := fst prearg); set (arg' :=snd prearg).
  hreplace IHlist with (IHlist (ptr-1) (ptr'-1)).
  destruct IHlist as [P [ Q IHlist]].
  exists (P ## {ptr,ptr'}|->{arg,arg'});  exists (Q //\\ h arg arg').
  intro cloud; rewrite (IHlist (h arg arg' //\\ cloud)).
  rewrite stateRelTensor_comm; rewrite <- stateRelTensor_assoc.
  repeat rewrite <- cloudRelConj_assoc; reflexivity.
Qed.  

Ltac explicit_stack list ptr ptr' :=   
  assert (Hstack := stack_description_explicit list ptr ptr');
  destruct Hstack as [Pstack [Qstack Hstack]].

Lemma stack_description_weakening : forall list ptr ptr' P,
	stateRelLeq (stack_description list ptr ptr' P) (cloudToState P).
Proof.
  intros.
  explicit_stack list ptr ptr';  rewrite (Hstack P).
  rewrite stateRelTensor_weakening.
  rewrite cloudRelConj_comm; rewrite cloudRelConj_weakening; reflexivity.
Qed.

Lemma stack_description_strength : forall list ptr ptr' P Q,
	stateRelLeq (P ## stack_description list ptr ptr' Q) 
         (stack_description list ptr ptr' (stateToCloud P //\\ Q)).
Proof.
  intros; explicit_stack list ptr ptr'.
  rewrite (Hstack Q); rewrite (Hstack (stateToCloud P //\\ Q)).
  rewrite stateRelTensor_assoc; rewrite TensorConj_strength.
  eapply ppredtensorcompat2; try reflexivity.
  eapply cloudtostatecompat2; apply cloudRelEq_Leq; ring.
Qed.

Lemma stateExP_stack : forall (X:Type) list ptr ptr' (h:X -> cloudRel),
  stateRelEq (Ex x, stack_description list ptr ptr' (h x)) 
     (stack_description list ptr ptr' (Ex_c x, (h x))).
Proof.
  intros; explicit_stack list ptr ptr'.
  rewrite (Hstack (Ex_c x , h x)).
  rewrite cloudRelConj_comm; rewrite <- cloudExP_conj; try exact 5.
  rewrite <- cloudExP_extract; rewrite <- stateExP_tensor.
  apply stateExP_rewrite; intro arg.
  rewrite (Hstack (h arg)).
  eapply ppredtensorcompat; try reflexivity.
  eapply cloudtostatecompat; ring.
Qed.

(* some lemmas about the structure of env *)

Fixpoint unfold_env env current current' finish finish' Ra {struct env} : cloudRel :=
  match env with
    | nil => cloudRel_lift (current=finish /\ current'=finish')
    | h :: t => Ex_c ptr, Ex_c ptr', Ex_c next, Ex_c next', 
      (({current,current'} ||-> {ptr,ptr'}) //\\ ({current+1,current'+1}||->{next,next'}) //\\
                           [|h|] Ra ptr ptr' //\\ unfold_env t next next' finish finish' Ra)
  end.

Lemma unfold_env_semantic : forall headenv tailenv Ra start start' x0 x0',  
   cloudRelLeq (unfold_env headenv start start' x0 x0' Ra //\\ [||tailenv||] Ra x0 x0')
                             ([||headenv++tailenv||] Ra start start') .
Proof.
  induction headenv.
  simpl; intros.
  compare start x0.
  intro e; rewrite e.
  setoid_replace (x0=x0) with True using relation iff; try (intuition; omega).
  compare start' x0'.
  intro e'; rewrite e'.
  setoid_replace (x0'=x0') with True using relation iff; try (intuition; omega).
  setoid_replace (True /\ True) with True using relation iff; [idtac | intuition].
  fold statePtrue; setoid_rewrite cloudRelConj_unit; apply cloudRelEq_Leq; ring.
  intro.
  setoid_replace (start'=x0') with False using relation iff; try omega.
  setoid_replace (True /\ False) with False using relation iff; [idtac | intuition].
  setoid_rewrite cloudRelConj_zero.
  apply cloudRelLeq_false_min.
  
  intro.
  setoid_replace (start=x0) with False using relation iff; try omega.
  setoid_replace (False /\ start' = x0') with False using relation iff; [idtac | intuition].
  setoid_rewrite cloudRelConj_zero.
  apply cloudRelLeq_false_min.

  simpl; intros.
  repeat (setoid_rewrite <- cloudExP_conj ; trivial; apply cloudExP_mono; intro).
  hreplace IHheadenv with (IHheadenv tailenv Ra x2 x3 x0 x0').
  repeat rewrite <- cloudRelConj_assoc.
  setoid_rewrite <- IHheadenv; reflexivity.
Qed.

Lemma tail_unfold_env_semantic : forall env a Ra start start' x0 x0' x1 x1' x2 x2',  
   cloudRelLeq (({x0+1,x0'+1}||-> {x2,x2'}) //\\ ({x0,x0'}||->{x1,x1'})
     //\\ [|a|] Ra x1 x1'//\\ unfold_env env start start' x0 x0' Ra)
               (unfold_env (env++[a]) start start' x2 x2' Ra).
Proof.
  induction env.
  simpl;intros.
  apply cloudExistsLeq_single; exists x1.
  apply cloudExistsLeq_single; exists x1'.
  apply cloudExistsLeq_single; exists x2.
  apply cloudExistsLeq_single; exists x2'.
  repeat rewrite cloudRelConj_assoc.
  compare start x0.
  intro e; setoid_replace (start=x0) with True using relation iff; try (intuition; omega).
  compare start' x0'.
  intro e'; setoid_replace (start'=x0') with True using relation iff; try (intuition; omega).
  setoid_replace (x2=x2) with True using relation iff; try (intuition; omega).
  setoid_replace (x2'=x2') with True using relation iff; try (intuition; omega).
  rewrite e; rewrite e'.
  rewrite (cloudRelConj_comm ({x0 + 1, x0' + 1}||-> {x2, x2'})).
  apply cloudRelEq_Leq;  ring.
  intro.
  setoid_replace (start'=x0') with False using relation iff; try omega.
  rewrite cloudRelConj_comm.
  setoid_replace (True /\ False) with False using relation iff; [idtac | intuition].
  rewrite cloudRelConj_zero.
  apply cloudRelLeq_false_min.

  intro.
  setoid_replace (start=x0) with False using relation iff; try omega.
  rewrite cloudRelConj_comm.
  setoid_replace (False /\ start' = x0') with False using relation iff; [idtac | intuition].
  rewrite cloudRelConj_zero.
  apply cloudRelLeq_false_min.

  simpl;intros.
  repeat rewrite cloudRelConj_assoc.
  rewrite cloudRelConj_comm.
  repeat (setoid_rewrite <- cloudExP_conj; trivial; apply cloudExP_mono; intro).
  repeat rewrite <- cloudRelConj_assoc. 
  do 3 (apply ppredconjcompat2; [apply cloudRelEq_Leq;  ring | idtac]).
  rewrite cloudRelConj_comm.
  repeat rewrite <- cloudRelConj_assoc.
  eapply IHenv.
Qed.
