(* tools.v

   A bench of tactics and lemmas that will be use during the proof 
   of the correcteness of the compiler.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export focusing_efficient.
Require Export lambda.
Require Export Basics.
Typeclasses eauto := dfs 10000.

Tactic Notation (at level 0) "isolate_conj" :=
  let Conj_Block := fresh "Conj_Block" in
    let Block := fresh "Block" in
  try match goal with
  | |- judgment (?P##?Q) ?p ?p' ?l ?l' =>  set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | |- stateRelEq (?P##?Q') ?Q => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | |- stateRelLeq (?P##?Q') ?Q => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | |- stateRelEq ?P ?Q =>  set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | |- stateRelLeq ?P ?Q => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  end;
  unfold Conj_Block at 1 in Block.

Tactic Notation (at level 0) "isolate_conj'" :=
  let Conj_Block := fresh "Conj_Block" in
    let Block := fresh "Block" in
  try match goal with
  | |- stateRelEq ?Q (?P##?Q') => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | |- stateRelLeq ?Q (?P##?Q') => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | |- stateRelEq ?Q ?P =>  set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | |- stateRelLeq ?Q ?P => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  end;
  unfold Conj_Block at 1 in Block.


Tactic Notation (at level 0) "isolate_conj" "in" ident(H) :=
  let Conj_Block := fresh "Conj_Block" in
    let Block := fresh "Block" in
  try match goal with
  | H:judgment (?P##?Q) ?p ?p' ?l ?l' |- _ => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | H:stateRelEq (?P##?Q') ?Q |- _ => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | H:stateRelLeq (?P##?Q') ?Q |- _ => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | H:stateRelEq ?P ?Q |- _ => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  | H:stateRelLeq ?P ?Q |- _ => set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  end;
  unfold Conj_Block at 1 in Block.

Tactic Notation (at level 0) "rebuilt_conj" ident(Conj_Block) ident(Block) :=
  rewrite <- Block; clear Conj_Block Block.

Tactic Notation (at level 0) "rebuilt_conj" "in" ident(H) ident(Conj_Block) ident(Block) :=
  rewrite <- Block in H; clear Conj_Block Block.


Tactic Notation (at level 0) "ring_simplify" "in" "goal":=
  try match goal with
  | |- judgment ?P ?p ?p' ?l ?l' => ring_simplify P
  | |- derivation ?P ?p ?p' ?l ?l' ?k => ring_simplify P
  | |- stateRelEq ?P ?Q => ring_simplify P Q
  | |- stateRelLeq ?P ?Q => ring_simplify P Q
  end.

Tactic Notation (at level 0) "ring_simplify" "in" ident(H) :=
  try match goal with
  | H:judgment ?P ?p ?p' ?l ?l' |- _ => ring_simplify P in H
  | H:derivation ?P ?p ?p' ?l ?l' ?k |- _ => ring_simplify P in H
  | H:stateRelEq ?P ?Q |- _ => ring_simplify P Q in H
  | H:stateRelLeq ?P ?Q |- _ => ring_simplify P Q in H
  end.

Ltac normal_form :=
  try (focus 0); try (focus' 0).

Tactic Notation (at level 0) "normal_form" "in" ident(H) := 
  focus_hyp 0.

Ltac ineq_ring := apply stateRelEq_Leq; ring.

Ltac normal_form_conj := 
  focus_conj 0.

Tactic Notation (at level 0) "normal_form_conj" "in" ident(H) := 
  focus_conj_hyp 0.

Ltac ineq_ring_conj := apply cloudRelEq_Leq; ring.

Ltac sub_heap :=
  try repeat (rewrite stateRelTensor_assoc);
  first [apply stateRelTensor_weakening | try repeat (
	(rewrite <- stateRelTensor_assoc);
	try (apply stateRelTensor_weakening)
	)].

Tactic Notation (at level 0) "destruct_tensor" ident(H) := 
  let f1 := fresh "f1" in let f2 := fresh "f2" in 
  let f'1 := fresh "f'1" in let f'2 := fresh "f'2" in 
  let Hdisj := fresh "Hdisj" in let Hjoin := fresh "Hjoin" in 
  let Hdisj' := fresh "Hdisj'" in let Hjoin' := fresh "Hjoin'" in 
  destruct H as (f1 & f2 & f'1 & f'2 & Hdisj & Hjoin & Hdisj' & Hjoin' & H).

Ltac elim_head := setoid_rewrite stateRelTensor_comm;
  eapply stateRelLeq_trans;
  [ apply stateRelTensor_weakening | idtac].

Ltac hide_tail :=  let tail := fresh "tail" in
  match goal with
    | |- judgment (?P##?Q) ?p ?p' ?l ?l' => set (tail := Q)
    | |- stateRelLeq (?P##?Q) ?Q' => set (tail := Q)
    | |- stateRelEq (?P##?Q) ?Q' => set (tail := Q)
  end.

Ltac hide_tail' :=  let tail := fresh "tail" in
  match goal with
    | |- (stateRelLeq ?Q' (?P##?Q)) => set (tail := Q)
    | |- (stateRelEq ?Q' (?P##?Q))  => set (tail := Q)
  end.

Lemma topfromsplit : forall (n hwm : nat), 
               stateRelEq (Topfrom hwm hwm) ((Block hwm hwm n n) ## (Topfrom (hwm+n) (hwm+n))).
Proof.
  intros.
  
  induction n.  
  simpl; rewrite plus_0_r; symmetry; apply stateRelTensor_unit.

  rewrite IHn; unfold Block; fold Block.
  omega_replace (hwm+S n) with (S (hwm+n)). 
  normal_form. 
  focus' 2; focus' 1.
  rewrite stateRelTensor_assoc; rewrite <- topfromBlock.
  focus 1; reflexivity.
Qed.

Lemma judgment_lift : forall P (Q:Prop) p p' l l', (Q -> judgment P p p' l l') ->
                                  judgment (stateRel_lift Q##P) p p' l l'.
Proof.
  unfold judgment, derivation,Perp; simpl.
  unfold stateRelTensor, perpOf; simpl;intros.
  destruct_tensor H1; destruct H1 as (HQ,HP).
  apply (H HQ k s s' j) ; try assumption.
  unfold Interp; eapply stateRel_cond; try eassumption; trivial.
  rewrite join_disj_comm in Hjoin; try assumption.
  eapply join_extends; try eassumption.
  rewrite join_disj_comm in Hjoin'; try assumption.
  eapply join_extends; try eassumption.
Qed.

Lemma judgment_lift_cloud : forall P R (Q:Prop) p p' l l', (Q -> judgment (cloudToState R ## P) p p' l l') ->
                                  judgment (cloudToState(cloudRel_lift Q //\\ R) ## P) p p' l l'.
Proof.
  unfold judgment, derivation,Perp; simpl.
  unfold stateRelTensor, cloudRelConj, perpOf; simpl;intros.
  destruct_tensor H1; destruct H1 as ((HQ,HR),HP).
  apply (H HQ k s s' j) ; try assumption.
  exists f1; exists f2; exists f'1; exists f'2; intuition.
Qed.

Lemma judgment_lift_stack : forall P R (Q:Prop) p p' l l' list ptr ptr',
  (Q -> judgment (stack_description list ptr ptr' R ## P) p p' l l') ->
        judgment (stack_description list ptr ptr' (cloudRel_lift Q //\\ R) ## P) p p' l l'.
Proof.
  intros.
  explicit_stack list ptr ptr'.
  rewrite (Hstack (cloudRel_lift Q //\\ R)).
  rewrite (Hstack R) in H.
  rewrite cloudRelConj_comm.
  normal_form.
  normal_form_conj.
  apply judgment_lift_cloud.  
  rewrite cloudRelConj_comm.
  setoid_rewrite <- stateRelTensor_assoc in H.
  assumption.
Qed.

Lemma existsDerivation : forall (A:Type) (h:A->stateRel) P p p' l l' k,
             (forall (x : A), derivation (h x ## P) p p' l l' k) <->
             derivation ((Exists h) ## P ) p p' l l' k.
Proof.
  intros until l;unfold derivation,Perp,perpOf;simpl;split.
  intros.  
  destruct_tensor H1 ;destruct H1 as ([x Hf1],Hf2). 
  hreplace H with (H x s s' j H0).
  apply H.
  exists f1;exists f2;exists f'1;exists f'2;intuition.
  intros.
  destruct_tensor H1;destruct H1 as (Hf1,Hf2).
  hreplace H with (H s s' j H0).
  apply H.
  exists f1;exists f2; exists f'1;exists f'2; intuition.
  exists x; intuition.
Qed.  

Lemma existsJudgment : forall (A:Type) (h:A->stateRel) P p p' l l',
             (forall (x : A), judgment ((h x) ## P) p p' l l') <->
             judgment ((Exists h) ## P ) p p' l l'.
Proof.
  unfold judgment; intuition.
  rewrite <- existsDerivation; intro; apply (H x k).
  generalize x; rewrite existsDerivation; intro; apply (H k).
Qed.  

Lemma existsJudgment_total : forall (A:Type) (h:A->stateRel) R P p p' l1 l1' l2 l2',
             (forall (x : A), explicit_judgment ((h x) ## P) R p p' l1 l1' l2 l2') <->
             explicit_judgment ((Exists h) ## P ) R p p' l1 l1' l2 l2'.
Proof.
  unfold explicit_judgment; intuition.
  unfold Interp, stateExists  in H0; simpl in H0.
  destruct_tensor H0; destruct H0 as (Hh & HP).
  destruct Hh as (x & Hh).
  specialize (H x k s1 s1').
  destruct H as (j1 & j1' & s2 & s2' & Hstep & Hstep' & HR).
  exists f1; exists f2; exists f'1; exists f'2; intuition.
  exists j1; exists j1'; exists s2; exists s2'; intuition.
  specialize (H k s1 s1').
  destruct H as (j1 & j1' & s2 & s2' & Hstep & Hstep' & HR).
  destruct_tensor H0; exists f1; exists f2; exists f'1; exists f'2; intuition.
  exists x; assumption.
  exists j1; exists j1'; exists s2; exists s2'; intuition.
Qed.      

Lemma existsDerivation_single : forall (A:Type) (h:A->stateRel) p p' l l' k,
             (forall (x : A), derivation (h x) p p' l l' k) <->
             derivation (Exists h) p p' l l' k.
Proof.
  unfold derivation,Perp,perpOf;simpl;intuition.
  destruct H1 as [f Hf];  hreplace H with (H f s s' j H0);  intuition.
  destruct H1 as [f Hf];  hreplace H with (H f s s' j H0);  intuition.
  hreplace H with (H s s' j H0); firstorder.
  hreplace H with (H s s' j H0); firstorder.
Qed.  

Lemma existsJudgmentSingle_total : forall (A:Type) (h:A->stateRel) R p p' l1 l1' l2 l2',
             (forall (x : A), explicit_judgment (h x)R p p' l1 l1' l2 l2') <->
             explicit_judgment (Exists h) R p p' l1 l1' l2 l2'.
Proof.
  unfold explicit_judgment; intuition.
  unfold Interp, stateExists  in H0; simpl in H0.
  destruct H0 as (x & Hh).
  specialize (H x k s1 s1').
  destruct H as (j1 & j1' & s2 & s2' & Hstep & Hstep' & HR); trivial.
  exists j1; exists j1'; exists s2; exists s2'; intuition.
  specialize (H k s1 s1').
  destruct H as (j1 & j1' & s2 & s2' & Hstep & Hstep' & HR).
  exists x; assumption.
  exists j1; exists j1'; exists s2; exists s2'; intuition.
Qed.   

Lemma existsJudgment_single : forall (A:Type) (h:A->stateRel) p p' l l',
             (forall (x : A), judgment (h x) p p' l l') <->
             judgment (Exists h) p p' l l'.
Proof.
  unfold judgment; intuition.
  rewrite <- existsDerivation_single; intro; apply (H x k).
  generalize x; rewrite existsDerivation_single; intro; apply (H k).
Qed.  
            
Lemma forallDerivation_single : forall (A:Type) (h:A->stateRel) p p' l l' k,
             (exists x, derivation (h x) p p' l l' k) ->
             derivation (ForAll h) p p' l l' k.
Proof.
  unfold derivation,Perp,perpOf;simpl; intuition.
  destruct H as [f H]; hreplace H with (H s s' j H0); firstorder.
  destruct H as [f H]; hreplace H with (H s s' j H0); firstorder.
Qed.  

Lemma forallJudgment_single : forall (A:Type) (h:A->stateRel) p p' l l' ,
             (exists x, judgment (h x) p p' l l' ) ->
             judgment (ForAll h) p p' l l'.
Proof.
  unfold judgment; intuition.
  apply forallDerivation_single.
  destruct H; exists x;  apply (H k).
Qed.  
  
Add Morphism Pre_arrow with signature cloudRelEq ==> (@eq nat) ==> (@eq nat) ==> (@eq stateRel) ==>
(@eq (stateRel-> nat-> nat-> cloudRel)) ==> (@eq (stateRel-> nat-> nat->cloudRel)) ==> stateRelEq
as prearrowcompat.
Proof.
  intros; unfold Pre_arrow.
  repeat(
    setoid_rewrite <- stateExP_tensor;
    apply stateExP_rewrite; intro).
  setoid_rewrite H; reflexivity.
Qed.

Add Morphism Pre_arrow with signature cloudRelLeq ++> (@eq nat) ==> (@eq nat) ==> (@eq stateRel) ==>
(@eq (stateRel->nat->nat->cloudRel)) ==> (@eq (stateRel->nat->nat->cloudRel)) ==> stateRelLeq
as prearrowcompat2.
Proof.
  intros; unfold Pre_arrow.
  repeat(
    setoid_rewrite <- stateExP_tensor;
    apply stateExP_mono; intro).
  setoid_rewrite H; reflexivity.
Qed.

Ltac state_antimono s H k :=
  unfold Interp in H;
  unfold Interp;
  apply (stateRel_cond _ _ (functionToPfunction s) _ _ k H);
  [apply subPfunction_refl | omega ].

Ltac state_antimono_simpl H k :=
  unfold Interp in H;
  match goal with
	| |- Interp ?P ?s ?p ?j =>
         	apply (stateRel_cond _ _ (functionToPfunction s) _  _ k H)
  end;
  [apply subPfunction_refl | omega ].

Ltac state_antimono_hyp j Hpre :=
  let foo := fresh in
  match goal with
	| H:Interp ?P ?s ?p ?k |- _ =>
		assert (foo : Interp P s p j)
  end;
  try (match goal with
  	  | |- Interp ?P ?s ?p ?k =>
    	  	state_antimono s Hpre k
        end).

Ltac Interp_weakening :=
  match goal with
	| H:Interp ?p ?s  ?p?k |- _ =>
		apply Interp_leq with (P1:= p)
  end;
  try rewrite ptsto_blank_leq;
  try  ineq_ring.

Ltac conclusion Hpost:=
  apply Hpost ;
  [ assumption | idtac |assumption].

Lemma judgment_internalization : forall P p p' l l',
  judgment P p p' l l' <-> judgment (P ## intern(p,p')) p p' l l'.
Proof.
  intuition.
  eapply judgment_leq; try apply H.
  apply stateRelTensor_weakening.
  unfold judgment, derivation , Perp in *; simpl; simpl in H.
  unfold perpOf in * ; simpl; simpl in H.
  intuition; hreplace H with (H k s s' j H0);
  rewrite split_conj in H; destruct H as (H,H');
  try apply H; try apply H'; try assumption;
  exists (functionToPfunction s); exists (nullPfunction Astate Bstate);
  exists (functionToPfunction s'); exists (nullPfunction Astate Bstate);
  intuition;try apply disjointPfunction_unit; try apply Extends_refl.
Qed.

Lemma judgment_internalization_ptsto : forall P Q Post m m' p p' l l', judgment (cloudToState ({m,m'} ||=> Later (Perp Post) //\\ Q) ## P) p p' l l'
                 <-> judgment (cloudToState ({m,m'}||=> Later (Perp (Post ## intern(p,p'))) //\\ Q) ## P) p p' l l' .
Proof.
  intuition.
  unfold cloudToState, judgment, derivation, Later , Perp,Perp in *; simpl; simpl in H.
  unfold perpOf,perpOf in * ; simpl; simpl in H.
  intros;  hreplace H with (H k s s' j H0).
  apply H;clear H.
  destruct_tensor H1 ;destruct H1 as (([ret[ret' (Hm,(Hm',Hptsto))]],HQ),HP).
  exists f1; exists f2;exists f'1; exists f'2; intuition.
  exists ret; exists ret';intuition;
  hreplace Hptsto with (Hptsto j0 H s0 s'0 j1 H1);
  rewrite split_conj in Hptsto; destruct Hptsto as (Hptsto,Hptsto');
  try apply Hptsto;try apply Hptsto' ; try assumption;
  exists (functionToPfunction s0); exists (nullPfunction Astate Bstate);
  exists (functionToPfunction s'0); exists (nullPfunction Astate Bstate);
  intuition; try apply disjointPfunction_unit; try apply Extends_refl.
  eapply judgment_leq; try apply H.
  eapply ppredtensorcompat2; try reflexivity.
  apply cloudtostatecompat2; eapply ppredconjcompat2; try reflexivity.
  apply ptstoCovariant_c; trivial.
  apply latercompat2; apply perpContravariant.
  unfold flip; apply stateRelTensor_weakening.
Qed.

Lemma judgment_internalization_ptsto_stack : forall p p' l l' m m' Post list ptr ptr' P Q,
        judgment (stack_description list ptr ptr' ({m,m'}||=>Later (Perp Post) //\\ Q) ## P) p p' l l'
        <-> judgment (stack_description list ptr ptr' ({m,m'}||=>Later (Perp (Post ## intern(p,p'))) //\\ Q) ## P) p p' l l' .
Proof.
  intros; explicit_stack list ptr ptr'.
  setoid_rewrite (Hstack ({m,m'}||=>Later (Perp Post) //\\ Q)); setoid_rewrite (Hstack ({m,m'}||=> Later(Perp (Post ## intern(p,p'))) //\\ Q)).
  setoid_rewrite (cloudRelConj_comm Qstack) at 1.
  setoid_rewrite (cloudRelConj_comm Qstack) at 1.
  repeat (setoid_rewrite <- cloudRelConj_assoc at 1).
  repeat (setoid_rewrite <- stateRelTensor_assoc at 1).
  apply judgment_internalization_ptsto.
Qed.

Lemma continuation_simplification : forall m m' P Q , stateRelLeq (({m,m'} |=> Later(Perp P))) P ->
              stateRelLeq (({m,m'} |=> Later(Perp P)) ## Q) ({m,m'} |=> Later(Perp (({m,m'} |=> Later(Perp P)) ## Q))).
Proof.
  intros.
  setoid_rewrite stateRelTensor_weakening at 1.
  assert (Hm: m = m); trivial.
  assert (Hm': m' = m'); trivial.
  rewrite <- (ptstoCovariant m m Hm m' m' Hm' (Later(Perp ({m,m'} |=> Later(Perp P)))) (Later(Perp (({m,m'} |=> Later(Perp P))##Q))));
  [ idtac | apply latercompat2;apply perpContravariant; unfold flip; apply stateRelTensor_weakening].
  apply ptstoCovariant; trivial.
  apply latercompat2;apply perpContravariant.
  assumption.
Qed.

Lemma continuation_simplification_cloud : forall m m' Post P1 P2 P3,
  stateRelLeq (P1 ## cloudToState (({m,m'} ||=> Later(Perp Post))//\\P2) ## P3)
                          Post ->
  stateRelLeq (P1 ## cloudToState (({m,m'} ||=> Later(Perp Post))//\\P2) ## P3)
                           ({m,m'}|=> Later(Perp (P1 ## cloudToState (({m,m'} ||=> Later(Perp Post))//\\P2) ## P3))).
Proof.
  intros.
  assert (Hm:m=m); trivial.
  assert (Hm': m' = m'); trivial.
  rewrite <- (ptstoCovariant m m Hm m' m' Hm' (Later(Perp Post)) (Later(Perp (P1 ## cloudToState (({m,m'} ||=> Later(Perp Post))//\\P2) ## P3))));
    [idtac | apply latercompat2;apply perpContravariant; unfold flip; assumption].
  focus 2; rewrite stateRelTensor_assoc;  rewrite stateRelTensor_comm; apply stateRelTensor_weakening_general.
  rewrite cloudRelConj_weakening.
  rewrite cloudStatePtsot; reflexivity.
Qed.

Lemma continuation_simplification_stack : forall m m' Post list ptr ptr' P1 P2 P3,
  stateRelLeq (P1 ## stack_description list ptr ptr' (({m,m'} ||=> Later(Perp Post))//\\P2) ## P3)
                          Post ->
  stateRelLeq (P1 ## stack_description list ptr ptr' (({m,m'} ||=> Later(Perp Post))//\\P2) ## P3)
                           (({m,m'}|=> Later(Perp (P1 ## stack_description list ptr ptr' (({m,m'} ||=> Later(Perp Post))//\\P2) ## P3)))).
Proof.
  intros.
  set (block1 := (P1 ## stack_description list ptr ptr' (({m,m'} ||=> Later(Perp Post))//\\P2) ## P3)).
  assert (Hm:m=m); trivial.
  assert (Hm': m' = m'); trivial.
  rewrite <- (ptstoCovariant m m Hm m' m' Hm' (Later(Perp Post)) (Later(Perp block1))).
  unfold block1; clear block1.
  focus 2; rewrite stateRelTensor_assoc;   rewrite stateRelTensor_comm.
  apply stateRelTensor_weakening_general.
  rewrite cloudRelConj_weakening.
  rewrite <- cloudStatePtsot;  rewrite stack_description_weakening;  reflexivity.
  apply latercompat2;apply perpContravariant; unfold flip; assumption.
Qed.

Lemma continuation_enrichment : forall P Post ptr ptr' p p' l l' start start',
         judgment Post p p' l l' ->
         judgment (({ptr,ptr'}|=> Perp Post) ## P) p p' start start'->
         judgment (({ptr,ptr'}|-> {l,l'}) ## P) p p' start start'.
Proof.
  unfold judgment, derivation;simpl;unfold perpOf ; simpl; intuition;
  hreplace H0 with (H0 k s s' j H1);
  destruct_tensor H2 ;destruct H2 as ([l0 [l0' (hl0,(hl0',(h2,h2')))]],HP);
  rewrite split_conj in H0; destruct H0 as (H0,H0'); subst;
  try apply H0; try apply H0'; try assumption;
  exists f1;exists f2; exists f'1;exists f'2; intuition;
  exists l; exists l'; intuition;
  unfold perpOf,perpOf;simpl; intuition;
  specialize (H j s0 s'0 j0 H2); rewrite split_conj in H; destruct H as (H,H');
  try apply H ;try apply H' ; try  assumption.
Qed.

Lemma continuation_enrichment_cloud : forall Post ptr ptr' l l' p p' start start' P1 P2,
         judgment Post p p' l l' ->
         judgment ( cloudToState ({ptr,ptr'}||=> (Perp Post) //\\ P1) ## P2) p p' start start' ->
         judgment ( cloudToState (({ptr,ptr'}||-> {l ,l'}) //\\ P1) ## P2) p p' start start'.
Proof.
  unfold judgment, derivation;simpl;unfold perpOf ; simpl; intuition;
  hreplace H0 with (H0 k s s' j H1);
  destruct_tensor H2;destruct H2 as (([l0 [l0' (hl0,(hl0',(h2,h2')))]],HP1),Hp2);
  rewrite split_conj in H0; destruct H0 as (H0,H0'); subst;
  try apply H0; try apply H0'; try assumption;
  exists f1;exists f2; exists f'1;exists f'2; intuition;
  exists l; exists l'; intuition;
  unfold perpOf,perpOf;simpl; intuition;
  hreplace H with (H j s0 s'0 j0 H2); rewrite split_conj in H; destruct H as (H,H');
  try apply H ;try apply H' ; try  assumption.
Qed.

Lemma continuation_enrichment_stack : forall Post ptr0 ptr0' l l' p p' start start' list ptr ptr' P1 P2,
         judgment Post p p' l l' ->
         judgment (stack_description list ptr ptr' (({ptr0,ptr0'}||=> Perp Post) //\\ P1) ## P2)
                  p p' start start'->
         judgment (stack_description list ptr ptr' (({ptr0,ptr0'}||-> {l,l'}) //\\ P1) ## P2)
                  p p' start start'.
Proof.
  intros; explicit_stack list ptr ptr'.
  rewrite (Hstack ({ptr0,ptr0'}||=> Perp Post //\\ P1)) in H0; rewrite (Hstack (({ptr0,ptr0'} ||->{l,l'}) //\\ P1)).
  setoid_rewrite cloudRelConj_comm at 1.
  normal_form; normal_form_conj.
  setoid_rewrite cloudRelConj_comm at 1 in H0; 
  normal_form in H0; normal_form_conj in H0.
  eapply continuation_enrichment_cloud; try apply H; assumption.
Qed.

Lemma continuation_deduction : forall Post P reg reg' return_add return_add' p p',
         judgment Post p p' return_add return_add' ->
         stateRelLeq (({reg,reg'} |-> {return_add,return_add'}) ## intern(p,p') ## cloudToState P)
                                  (({reg,reg'} |=> Later (Perp Post)) ## intern(p,p') ## cloudToState P).
Proof.
  unfold stateRelLeq, judgment; simpl;intros.
  destruct_tensor H0;destruct H0 as ( [l[l'(hl,(hl',(h1,h1')))]],foo).
  destruct_tensor foo;destruct foo as ((Hp,Hp'),HP).
  exists f1; exists f2; exists f'1; exists f'2; intuition.
  exists l; exists l'; intuition.
  unfold derivation in H.
  rewrite (Extends_eq Hp);rewrite (Extends_eq Hp');subst; apply (H j).
  exists f0; exists f3 ;exists f'0; exists f'3 ; intuition.
Qed.

Lemma continuation_enrichment_counter : forall Post ptr ptr' l l' p p' start start' P k,
         derivation Post p p' l l' k ->
         judgment ({ptr,ptr'}|=> (Later (Perp Post)) ## P) p p' start start' ->
         derivation (({ptr,ptr'}|-> {l,l'}) ## P) p p' start start' (S k).
Proof.
  unfold judgment, derivation,Later;simpl;unfold perpOf ; simpl; unfold perpOf; simpl.
  intuition;  hreplace H0 with (H0 (S k) s s' j H1);
  rewrite split_conj in H0 ; destruct H0 as (H0,H'0);
  try apply H0; try apply H'0; try assumption; clear H0 H'0;
  destruct_tensor H2;destruct H2 as ([l0 [l'0 (hl,(hl',(hl0,hl'0)))]],HP);subst;
  exists f1;exists f2; exists f'1;exists f'2; intuition;
  exists l; exists l'; intuition;
  hreplace H with (H s0 s'0 j1);
  rewrite split_conj in H ; destruct H as (H,H'); try omega;
  intuition.
Qed.

Lemma continuation_enrichment_cloud_counter : forall Post ptr ptr' l l' p p' start start' P1 P2 k,
         derivation Post p p' l l' k ->
         judgment ( cloudToState ({ptr,ptr'}||=> (Later (Perp Post)) //\\ P1) ## P2) p p' start start' ->
         derivation ( cloudToState (({ptr,ptr'}||-> {l,l'}) //\\ P1) ## P2) p p' start start' (S k).
Proof.
  unfold judgment, derivation,Later;simpl;unfold perpOf ; simpl; unfold perpOf; simpl.
  intuition;  hreplace H0 with (H0 (S k) s s' j H1);
  rewrite split_conj in H0 ; destruct H0 as (H0,H'0);
  try apply H0; try apply H'0; try assumption;
  destruct_tensor H2 ;destruct H2 as (([l0 [l'0 (hl,(hl',(hl0,hl'0)))]],HP1),HP2);subst;
  exists f1;exists f2; exists f'1;exists f'2; intuition;
  exists l; exists l'; intuition;
  hreplace H with (H s0 s'0 j1);
  rewrite split_conj in H ; destruct H as (H,H'); try omega;
  intuition.
Qed.

Lemma continuation_enrichment_stack_counter : forall Post ptr0 ptr0' l l' p p' start start' list ptr ptr' P1 P2 k ,
         derivation Post p p' l l' k ->
         judgment (stack_description list ptr ptr' (({ptr0,ptr0'}||=> Later (Perp Post)) //\\ P1) ## P2) p p' start start'->
         derivation (stack_description list ptr ptr' (({ptr0,ptr0'}||-> {l,l'}) //\\ P1) ## P2) p p' start start' (S k).
Proof.
  intros; explicit_stack list ptr ptr'.
  rewrite (Hstack ({ptr0,ptr0'}||=>Later (Perp Post) //\\ P1)) in H0; rewrite (Hstack ({ptr0,ptr0'} ||-> {l,l'} //\\ P1)).
  normal_form.
  rewrite cloudRelConj_assoc; rewrite (cloudRelConj_comm Qstack). 
  normal_form_conj.
  eapply continuation_enrichment_cloud_counter; try apply H.
  focus_conj 1.
  normal_form in H0; assumption.
Qed.

Lemma continuation_lob : forall ptr ptr' l l' p p' P,
         judgment ({ptr,ptr'}|=> (Later (Perp (({ptr,ptr'}|-> {l,l'}) ## P))) ## P) p p' l l' ->
         judgment (({ptr,ptr'}|-> {l,l'}) ## P) p p' l l'.
Proof.
  intros.
  unfold judgment; induction k.
  unfold derivation; simpl; unfold perpOf,perpOf; intros;  inversion H0.
  simpl; intuition.
  eapply continuation_enrichment_counter; eassumption.
Qed.

Lemma recursion_continuation : forall ptr ptr' fcode fcode' Ra a b P p p',
  judgment (Pre_arrow (([|a--->b|]) Ra ptr ptr' //\\ P) ptr ptr' Ra ([|a|]) ([|b|])) p p' fcode fcode'->
  judgment (Pre_arrow (({ptr,ptr'}||-> {fcode, fcode'}) //\\ P) ptr ptr' Ra  ([|a|]) ([|b|])) p p' fcode fcode'.
Proof.
  intros; unfold judgment; induction k.
  unfold derivation; simpl; unfold perpOf,perpOf; intros;  inversion H0.
  simpl;intuition.
  unfold Pre_arrow in *.
  repeat (setoid_rewrite <- existsDerivation; intro).
  setoid_rewrite <- existsJudgment in H ; hreplace H with (H x).
  setoid_rewrite <- existsJudgment in H ; hreplace H with (H x0).
  setoid_rewrite <- existsJudgment in H ; hreplace H with (H x1).
  setoid_rewrite <- existsJudgment in H ; hreplace H with (H x2).
  setoid_rewrite <- existsJudgment in H ; hreplace H with (H x3).
  setoid_rewrite <- existsJudgment in H ; hreplace H with (H x4).
  setoid_rewrite <- existsJudgment in H ; hreplace H with (H x5).
  setoid_rewrite <- existsJudgment in H ; hreplace H with (H x6).
  normal_form in H;  rewrite stateRelTensor_assoc in H; rewrite stateRelTensor_comm in H.
  normal_form in H;  simpl in H;  repeat rewrite <- cloudExP_conj in H.
  rewrite <- cloudExP_extract in H; rewrite <- existsJudgment in H.
  hreplace H with (H ({ptr, ptr'}||->{fcode,fcode'} //\\ P)).
  normal_form_conj in H.
  rewrite (cloudRelConj_assoc P P) in H.
  rewrite <- (cloudConjContraction P P) in H; try reflexivity.
  rewrite (cloudConjContraction ({ptr, ptr'}||->{fcode,fcode'}) ({ptr, ptr'}||->{fcode,fcode'})); try reflexivity.
  normal_form.
  setoid_rewrite stateRelTensor_assoc; setoid_rewrite stateRelTensor_comm.
  normal_form; normal_form_conj.
  eapply continuation_enrichment_cloud_counter.
  apply IHk.
  unfold Pre_arrow in H; normal_form; eassumption.
Qed.

